/***************************
GTK Pattern Creation Lab
Copyright (C) 1999  Peter Stuart

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*****************************/

#include <gtk/gtk.h>
#include "../include/math3d.h"
#include "../include/matrix.h"
#include "types.h"
#include "brick.h"
#include "gpcl.h"
#include "gpcldlgs.h"

extern GtkWidget *mainwindow;
extern tpreferences gpreferences;

static void initialize_brick(pattern_class *pat);

static void brick_shade(pattern_class *pat, FLT s, FLT t, pvec3 rvec);
static void brick_configure(pattern_class *pat, GtkWidget *widget);
static void brick_update(pattern_class *pat);
static void brick_delete(pattern_class *pat);
static char* brick_name(void);

pattern_procs brick_procs =
   {brick_shade, brick_configure, brick_update,
    brick_delete, brick_name};

int brick_new(pattern_class *pat)
{
   pat->data = malloc(sizeof(brick_data));
   if (pat->data == NULL)
      return FALSE;

   pat->procs = &brick_procs;

   initialize_brick(pat);

   return TRUE;
}

void initialize_brick(pattern_class *pat)
{
   pbrick_data dat;
   dat = (pbrick_data)(pat->data);
   if (dat == NULL)
      return;

   dat->brick_width = 0.25;
   dat->brick_height = 0.08;
   dat->mortar_thickness = 0.01;

   dat->Cbrick[0] = 0.5;
   dat->Cbrick[1] = 0.15;
   dat->Cbrick[2] = 0.14;

   dat->Cmortar[0] = 0.5;
   dat->Cmortar[1] = 0.5;
   dat->Cmortar[2] = 0.5;

   dat->bmwidth = dat->brick_width + dat->mortar_thickness;
   dat->bmheight = dat->brick_height + dat->mortar_thickness;
   dat->mwf = (dat->mortar_thickness*0.5)/dat->bmwidth;
   dat->mhf = (dat->mortar_thickness*0.5)/dat->bmheight;

   pat->controls = NULL;
}

void brick_shade(pattern_class *pat, FLT s, FLT t, FLT* rvec)
{
   FLT ss, tt, sbrick, tbrick, w, h;
   FLT scoord = s;
   FLT tcoord = t;
   pbrick_data dat;

   dat = (pbrick_data)(pat->data);
   if (dat == NULL)
      return;

   ss = scoord / dat->bmwidth;
   tt = tcoord / dat->bmheight;

   if (mod(tt*0.5, 1.0) > 0.5)
      ss += 0.5;

   sbrick = floor(ss);
   tbrick = floor(tt);
   ss -= sbrick;
   tt -= tbrick;

   w = STEP(dat->mwf, ss) - STEP(1.0 - dat->mwf, ss); 
   h = STEP(dat->mhf, tt) - STEP(1.0 - dat->mhf, tt);

   cmix(dat->Cmortar, dat->Cbrick, w*h, rvec);
}

void brick_configure(pattern_class *pat, GtkWidget *container)
{
  GtkWidget *brick_box;
  GtkWidget *brick_width_box;
  GtkObject *brick_width_spin_adj;
  GtkWidget *brick_width_spin;
  GtkWidget *label1;
  GtkWidget *brick_height_box;
  GtkObject *brick_height_spin_adj;
  GtkWidget *brick_height_spin;
  GtkWidget *label2;
  GtkWidget *brick_thickness_box;
  GtkObject *brick_thickness_spin_adj;
  GtkWidget *brick_thickness_spin;
  GtkWidget *label3;
  GtkWidget *brick_color_button;
  GtkWidget *mortar_color_button;

   pbrick_data dat;
   dat = (pbrick_data)(pat->data);
   if (dat == NULL)
      return;

  brick_box = gtk_vbox_new (TRUE, 3);
  gtk_widget_show (brick_box);
  gtk_container_add (GTK_CONTAINER (container), brick_box);
  gtk_container_border_width (GTK_CONTAINER (brick_box), 2);

  brick_width_box = gtk_hbox_new (TRUE, 4);
  gtk_widget_show (brick_width_box);
  gtk_box_pack_start (GTK_BOX (brick_box), brick_width_box, TRUE, FALSE, 0);

  brick_width_spin_adj = gtk_adjustment_new (0.0, 0.0, 1.0, 0.01, 0.1, 0.0);
  brick_width_spin = gtk_spin_button_new (GTK_ADJUSTMENT (brick_width_spin_adj), 0.01, gpreferences.digits);
  gtk_spin_button_set_wrap(GTK_SPIN_BUTTON(brick_width_spin), TRUE);
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(brick_width_spin), dat->brick_width);
  gtk_object_set_data (GTK_OBJECT (mainwindow), "brick_width_spin", brick_width_spin);
  gtk_widget_show (brick_width_spin);
  gtk_box_pack_start (GTK_BOX (brick_width_box), brick_width_spin, FALSE, TRUE, 0);

  label1 = gtk_label_new ("Brick Width");
  gtk_widget_show (label1);
  gtk_box_pack_start (GTK_BOX (brick_width_box), label1, TRUE, TRUE, 0);

  brick_height_box = gtk_hbox_new (TRUE, 4);
  gtk_widget_show (brick_height_box);
  gtk_box_pack_start (GTK_BOX (brick_box), brick_height_box, TRUE, TRUE, 0);

  brick_height_spin_adj = gtk_adjustment_new (0.0, 0.0, 1.0, 0.01, 0.1, 0.0);
  brick_height_spin = gtk_spin_button_new (GTK_ADJUSTMENT (brick_height_spin_adj), 0.1, gpreferences.digits);
  gtk_spin_button_set_wrap(GTK_SPIN_BUTTON(brick_height_spin), TRUE);
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(brick_height_spin), dat->brick_height);
  gtk_object_set_data (GTK_OBJECT (mainwindow), "brick_height_spin", brick_height_spin);
  gtk_widget_show (brick_height_spin);
  gtk_box_pack_start (GTK_BOX (brick_height_box), brick_height_spin, FALSE, TRUE, 0);

  label2 = gtk_label_new ("Brick Height");
  gtk_widget_show (label2);
  gtk_box_pack_start (GTK_BOX (brick_height_box), label2, TRUE, TRUE, 0);

  brick_thickness_box = gtk_hbox_new (TRUE, 4);
  gtk_object_set_data (GTK_OBJECT (mainwindow), "brick_thickness_box", brick_thickness_box);
  gtk_widget_show (brick_thickness_box);
  gtk_box_pack_start (GTK_BOX (brick_box), brick_thickness_box, TRUE, FALSE, 0);

  brick_thickness_spin_adj = gtk_adjustment_new (0.0, 0.0, 1.0, 0.01, 0.1, 0.0);
  brick_thickness_spin = gtk_spin_button_new (GTK_ADJUSTMENT (brick_thickness_spin_adj), 0.01, gpreferences.digits);
  gtk_spin_button_set_wrap(GTK_SPIN_BUTTON(brick_thickness_spin), TRUE);
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(brick_thickness_spin), dat->mortar_thickness);
  gtk_object_set_data (GTK_OBJECT (mainwindow), "brick_thickness_spin", brick_thickness_spin);
  gtk_widget_show (brick_thickness_spin);
  gtk_box_pack_start (GTK_BOX (brick_thickness_box), brick_thickness_spin, FALSE, TRUE, 0);

  label3 = gtk_label_new ("Mortar Thickness");
  gtk_widget_show (label3);
  gtk_box_pack_start (GTK_BOX (brick_thickness_box), label3, TRUE, TRUE, 0);

  brick_color_button = gtk_button_new_with_label ("Brick Color");
  gtk_signal_connect (GTK_OBJECT (brick_color_button), "clicked",
                      GTK_SIGNAL_FUNC (on_color_button_clicked),
                      dat->Cbrick);
  gtk_widget_show (brick_color_button);
  gtk_box_pack_start (GTK_BOX (brick_box), brick_color_button, FALSE, FALSE, 0);

  mortar_color_button = gtk_button_new_with_label ("Mortar Color");
  gtk_signal_connect (GTK_OBJECT (mortar_color_button), "clicked",
                      GTK_SIGNAL_FUNC (on_color_button_clicked),
                      dat->Cmortar);
  gtk_widget_show (mortar_color_button);
  gtk_box_pack_start (GTK_BOX (brick_box), mortar_color_button, FALSE, FALSE, 0);

   pat->controls = brick_box;
}

void brick_update(pattern_class *pat)
{
   GtkWidget *current;

   pbrick_data dat;

   dat = (pbrick_data)(pat->data);
   if (dat == NULL)
      return;

   current = get_widget(mainwindow, "brick_width_spin");
   dat->brick_width = gtk_spin_button_get_value_as_float(
      GTK_SPIN_BUTTON(current));

   current = get_widget(mainwindow, "brick_height_spin");
   dat->brick_height = gtk_spin_button_get_value_as_float(
      GTK_SPIN_BUTTON(current));

   current = get_widget(mainwindow, "brick_thickness_spin");
   dat->mortar_thickness = gtk_spin_button_get_value_as_float(
      GTK_SPIN_BUTTON(current));

   dat->bmwidth = dat->brick_width + dat->mortar_thickness;
   dat->bmheight = dat->brick_height + dat->mortar_thickness;
   dat->mwf = (dat->mortar_thickness*0.5)/dat->bmwidth;
   dat->mhf = (dat->mortar_thickness*0.5)/dat->bmheight;
}

void brick_delete(pattern_class *pat)
{
   pbrick_data dat;

   dat = (pbrick_data)(pat->data);
   if (dat == NULL)
      return;

   free(dat);
   dat = NULL;

   gtk_widget_destroy(pat->controls);
}

char* brick_name(void)
{
   return "Brick";
}
