/***************************
GTK Pattern Creation Lab
Copyright (C) 1999  Peter Stuart

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*****************************/


#include <gtk/gtk.h>
#include "../include/math3d.h"
#include "../include/matrix.h"
#include "types.h"
#include "checker.h"
#include "gpcl.h"
#include "gpcldlgs.h"

extern GtkWidget *mainwindow;

static void initialize_checker(pattern_class *pat);

static void checker_shade(pattern_class *pat, FLT s, FLT t, pvec3 rvec);
static void checker_configure(pattern_class *pat, GtkWidget *widget);
static void checker_update(pattern_class *pat);
static void checker_delete(pattern_class *pat);
static char* checker_name(void);

pattern_procs checker_procs =
   {checker_shade, checker_configure, checker_update,
    checker_delete, checker_name};

int  checker_new(pattern_class *pat)
{
   pat->data = malloc(sizeof(checker_data));
   if (pat->data == NULL)
      return FALSE;

   pat->procs = &checker_procs;

   initialize_checker(pat);

   return TRUE;
}

void initialize_checker(pattern_class *pat)
{
   pchecker_data dat;
   dat = (pchecker_data)(pat->data);
   if (dat == NULL)
      return;

   dat->checker_frequency = 10;
   dat->checker_color1[0] = 1.0;
   dat->checker_color1[1] = 1.0;
   dat->checker_color1[2] = 1.0;
   vec3init(dat->checker_color2);

   pat->controls = NULL;
}

void checker_shade(pattern_class *pat, FLT s, FLT t, FLT* rvec)
{
   FLT smod;
   FLT tmod;

   pchecker_data dat = NULL;

   dat = (pchecker_data)(pat->data);
   if (dat == NULL)
      return;

   smod = mod(s*dat->checker_frequency, 1);
   tmod = mod(t*dat->checker_frequency, 1);

   if (smod < (FLT)0.5)
   {
      if (tmod < (FLT)0.5)
      {
         vec3cpy(rvec, dat->checker_color1);
      }
      else
      {
         vec3cpy(rvec, dat->checker_color2);
      }
   }
   else
   {
      if (tmod < (FLT)0.5)
      {
         vec3cpy(rvec, dat->checker_color2);
      }
      else
      {
         vec3cpy(rvec, dat->checker_color1);
      }
   }
} 

void checker_configure(pattern_class *pat, GtkWidget* container)
{
  GtkWidget *checker_box;
  GtkWidget *checker_freq_box;
  GtkObject *checker_frequency_spin_adj;
  GtkWidget *checker_frequency_spin;
  GtkWidget *checker_freq_label;
  GtkWidget *checker_color1_button;
  GtkWidget *checker_color2_button;

  pchecker_data dat;
  dat = (pchecker_data)(pat->data);
  if (dat == NULL)
     return;

  checker_box = gtk_vbox_new (FALSE, 3);
  gtk_widget_show (checker_box);
  gtk_container_add (GTK_CONTAINER (container), checker_box);
  gtk_container_border_width (GTK_CONTAINER (checker_box), 2);

  checker_freq_box = gtk_hbox_new (FALSE, 4);
  gtk_widget_show (checker_freq_box);
  gtk_box_pack_start (GTK_BOX (checker_box), checker_freq_box, TRUE, TRUE, 0);

  checker_frequency_spin_adj = gtk_adjustment_new (dat->checker_frequency, 0, 1000, 1, 10, 10);
  checker_frequency_spin = gtk_spin_button_new (GTK_ADJUSTMENT (checker_frequency_spin_adj), 1, 0);
  gtk_object_set_data (GTK_OBJECT (mainwindow), "checker_frequency_spin", checker_frequency_spin);
  gtk_widget_show (checker_frequency_spin);
  gtk_box_pack_start (GTK_BOX (checker_freq_box), checker_frequency_spin, TRUE,
TRUE, 0);

  checker_freq_label = gtk_label_new ("Frequency");
  gtk_widget_show (checker_freq_label);
  gtk_box_pack_start (GTK_BOX (checker_freq_box), checker_freq_label, TRUE, TRUE, 0);

  checker_color1_button = gtk_button_new_with_label ("Color 1");
  gtk_widget_show (checker_color1_button);
  gtk_box_pack_start (GTK_BOX (checker_box), checker_color1_button, TRUE, TRUE, 0);
  gtk_signal_connect (GTK_OBJECT (checker_color1_button), "clicked",
                      GTK_SIGNAL_FUNC (on_color_button_clicked),
                      dat->checker_color1);

  checker_color2_button = gtk_button_new_with_label ("Color 2");
  gtk_widget_show (checker_color2_button);
  gtk_box_pack_start (GTK_BOX (checker_box), checker_color2_button, TRUE, TRUE, 0);
  gtk_signal_connect (GTK_OBJECT (checker_color2_button), "clicked",
                      GTK_SIGNAL_FUNC (on_color_button_clicked),
                      dat->checker_color2);

  pat->controls = checker_box;
}

void checker_update(pattern_class *pat)
{
   GtkWidget *current;

   pchecker_data dat;

   dat = (pchecker_data)(pat->data);
   if (dat == NULL)
      return;

   current = get_widget(mainwindow, "checker_frequency_spin");
   dat->checker_frequency = gtk_spin_button_get_value_as_float(
      GTK_SPIN_BUTTON(current));
}

void checker_delete(pattern_class *pat)
{
   pchecker_data dat;

   dat = (pchecker_data)(pat->data);
   if (dat == NULL)
      return;

   free(dat);
   dat = NULL;

   gtk_widget_destroy(pat->controls);
}

char* checker_name(void)
{
   return "Checker";
}
