/***************************
GTK Pattern Creation Lab
Copyright (C) 1999  Peter Stuart

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*****************************/


#include <gtk/gtk.h>
#include "types.h"
#include "patterns.h"
#include "gpcldlgs.h"
#include "gpcl.h"

static GtkWidget *preferences_dialog = NULL;
static GtkWidget *color_selection = NULL;
static GtkWidget *about_window = NULL;
static GtkWidget *progress_window = NULL;

static pattern_class current_pattern;

char filename[128];
char progress[20];
tpreferences gpreferences;
GtkWidget *configure_frame = NULL;
GtkWidget *generate_button = NULL;
GtkWidget *interrupt_button = NULL;
GtkWidget *progress_bar = NULL;
GtkWidget *progress_label = NULL;

static guchar *idle_buf;
static int generate_idle = 0;
static int idle_count = 0;

static void gpcl_initialize(void);
static void gpcl_clean(void);
static gint generate_idle_func(GtkWidget *preview);
static void end_generation(void);

int
main (int argc, char *argv[])
{
  gpcl_initialize();

  gtk_set_locale ();
  gtk_init (&argc, &argv);

  mainwindow = create_mainwindow ();

  gtk_widget_show (mainwindow);

  gtk_main ();
  return 0;
}

void
gpcl_initialize(void)
{
   current_pattern.data = NULL;
   interrupt = FALSE;
   gpreferences.width = 128;
   gpreferences.height = 128;
   gpreferences.digits = 2;
   gpreferences.increment = 0.01;
   gpreferences.is_stretched = FALSE;
   gpreferences.show_progress = FALSE;
}

void
gpcl_clean(void)
{
   if (current_pattern.data != NULL)
      current_pattern.procs->delete(&current_pattern);
}


void
on_file_new_activate                   (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
   g_print("File->New not done\n");
}


void
on_file_open_activate                  (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
   g_print("File->Open not done\n");
}


void
on_file_save_activate                  (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
   g_print("File->Save not done\n");
}


void
on_file_save_as_activate               (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
   g_print("File->Save As... not done\n");
}


void
on_file_preferences_activate           (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
   if (preferences_dialog == NULL)
   {
      preferences_dialog = create_preferences_dialog();
      gtk_widget_show(preferences_dialog);
   }
}


void
on_mainwindow_end_event                (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
   gpcl_clean();
   gtk_main_quit();
}


void
on_pattern_selection                   (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
   NEWFUNC create = (NEWFUNC)user_data;
   GtkWidget *configure = NULL;

   configure = get_widget(mainwindow, "configurebox");
   gtk_widget_hide(configure);

   if (current_pattern.data != NULL)
   {
      current_pattern.procs->delete(&current_pattern);
      gtk_widget_destroy(configure_frame);
   }

   if (!create(&current_pattern))
      g_print("Error in creation\n");

   configure_frame = gtk_frame_new(current_pattern.procs->name());
   gtk_box_pack_start(GTK_BOX(configure),
                      configure_frame, FALSE, FALSE, 0);
   gtk_widget_show(configure_frame);

   current_pattern.procs->configure(&current_pattern, configure_frame);

   gtk_widget_show(configure);

   gtk_widget_set_sensitive(generate_button, TRUE);
}


void
on_help_about_activate                 (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
   if (about_window == NULL)
   {
      about_window = create_about_window();
      gtk_widget_show(about_window);
   }
}


void
on_generate_button_clicked             (GtkButton       *button,
                                        gpointer         user_data)
{
   GtkWidget* window = NULL;
   GtkWidget *preview;

   if (current_pattern.data == NULL)
      return;

   if (progress_window != NULL)
      return;

   current_pattern.procs->update(&current_pattern);

   /* attempt to allocate memory for color buffer */
   if ((idle_buf = calloc(gpreferences.width*3, sizeof(guchar))) == NULL)
   {
      /* message_box("Could not allocate\n memory."); */
      return;
   }

   if (gpreferences.show_progress)
   {
      /* create progress bar */
      progress_window = create_progress_window();
   }

   /* create new window for display */
   window = gtk_window_new (GTK_WINDOW_TOPLEVEL);

   gtk_window_set_title (GTK_WINDOW (window), "Generated Texture");

   /* create new preview and insert into window */
   preview = gtk_preview_new (GTK_PREVIEW_COLOR);
   gtk_preview_size (GTK_PREVIEW (preview), 
         gpreferences.width, gpreferences.height);
   gtk_container_add (GTK_CONTAINER (window), preview);

   /* enable/disable various functions */
   gtk_widget_set_sensitive(generate_button, FALSE);
   gtk_widget_set_sensitive(interrupt_button, TRUE);

   /* create idle loop */
   generate_idle = gtk_idle_add((GtkFunction) generate_idle_func, preview);

   gtk_widget_show(preview);
   gtk_widget_show(window);
}

gint
generate_idle_func(GtkWidget *preview)
{
   static FLT pos_x = 0.0;
   static FLT pos_y = 0.0;
   static FLT delta_x = 0.0;
   static FLT delta_y = 0.0;
   static int width = 0;
   static int height = 0;

   int j, k;
   COLOR color;

   if (!GTK_IS_PREVIEW(preview) || (idle_count > height)
       || interrupt)
   {
      if (progress_window != NULL)
      {
         gtk_widget_destroy(progress_window);
         progress_window = NULL;
      }
      end_generation();
      return FALSE;
   }

   /* initialize static variables */
   if (idle_count == 0)
   {
      width = gpreferences.width;
      height = gpreferences.height;

      /* set delta_x and delta_y. delta_y will vary as delta_x if
         is_stretched is FALSE
      */
      delta_x = (FLT)1.0/width;
      if (gpreferences.is_stretched)
         delta_y = (FLT)1.0/height;
      else
         delta_y = delta_x;

      /* start at point 0,0 */
      pos_x = pos_y = (FLT)0.0;
   }

   /* reset pos_x */
   pos_x = 0.0;

   /* loop through each pixel in row */
   for (j = 0, k = 0; j < width; j++)
   {
      current_pattern.procs->shade(&current_pattern, pos_x, pos_y, color);
      idle_buf[k+0] = rgbtoint(color[0]);
      idle_buf[k+1] = rgbtoint(color[1]);
      idle_buf[k+2] = rgbtoint(color[2]);
      pos_x += delta_x;
      k += 3;
   }
   /* draw row of preview */
   gtk_preview_draw_row (GTK_PREVIEW (preview), idle_buf,
         0, idle_count, width);

   pos_y += delta_y;

   idle_count++;

   if ((progress_window != NULL) && (progress_bar != NULL))
   {
      gtk_progress_bar_update(GTK_PROGRESS_BAR(progress_bar), 
                           (float)idle_count/height);
      sprintf(progress, "Progress: %d%%", 
                 (int)(100*((float)idle_count/height)));
      gtk_label_set(GTK_LABEL(progress_label), progress);
   }
   gtk_widget_draw(preview, NULL);

   return TRUE;
}

void
end_generation(void)
{
   idle_count = 0;
   interrupt = FALSE;
   gtk_widget_set_sensitive(generate_button, TRUE);
   gtk_widget_set_sensitive(interrupt_button, FALSE);

   free(idle_buf);
}

void
on_interrupt_button_clicked            (GtkButton       *button,
                                        gpointer         user_data)
{
   gtk_idle_remove(generate_idle);
   generate_idle = 0;

   end_generation();

   if (progress_window != NULL)
   {
      gtk_widget_destroy(progress_window);
      progress_window = NULL;
   }

}


void
on_preferences_apply_button_clicked    (GtkButton       *button,
                                        gpointer         user_data)
{
   GtkWidget *current = NULL;

   current = get_widget(preferences_dialog, "preferences_is_stretched_button");
   gpreferences.is_stretched = GTK_TOGGLE_BUTTON(current)->active;

   current = get_widget(preferences_dialog, "preferences_show_progress_button");
   gpreferences.show_progress = GTK_TOGGLE_BUTTON(current)->active;

   current = get_widget(preferences_dialog, "preferences_width_spinner");
   gpreferences.width = gtk_spin_button_get_value_as_float(
      GTK_SPIN_BUTTON(current)); 

   current = get_widget(preferences_dialog, "preferences_height_spinner");
   gpreferences.height = gtk_spin_button_get_value_as_float(
      GTK_SPIN_BUTTON(current));

   current = get_widget(preferences_dialog, "preferences_digits_spinner");
   gpreferences.digits = (int)gtk_spin_button_get_value_as_float(
      GTK_SPIN_BUTTON(current));

   gpreferences.increment = 1.0 * pow(10, -gpreferences.digits);
}


void
on_preferences_close_button_clicked    (GtkButton       *button,
                                        gpointer         user_data)
{
   gtk_widget_destroy(preferences_dialog);
   preferences_dialog = NULL;
}

void
on_color_button_clicked                (GtkButton       *button,
                                        gpointer         user_data)
{
   if (color_selection == NULL)
   {
      color_selection = create_colorselection_dialog(user_data);
      gtk_widget_show(color_selection);
   }
}


void
on_color_selection_ok_button_clicked   (GtkButton       *button,
                                        gpointer         user_data)
{
   gtk_color_selection_get_color(GTK_COLOR_SELECTION(
                     GTK_COLOR_SELECTION_DIALOG(
                     color_selection)->colorsel), user_data);
   gtk_widget_destroy(color_selection);
   color_selection = NULL;
}


void
on_color_selection_cancel_button_clicked
                                        (GtkButton       *button,
                                        gpointer         user_data)
{
   gtk_widget_destroy(color_selection);
   color_selection = NULL;
}

void
on_color_selection_help_button_clicked
					(GtkButton	*button,
					gpointer	user_data)
{

}

void
on_about_ok_button_clicked
                                        (GtkButton      *button,
                                        gpointer        user_data)
{
   gtk_widget_destroy(about_window);
   about_window = NULL;
}
