/******************************
GTK Pattern Creation Lab
Copyright (C) 1999  Peter Stuart

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
******************************/

#include <gtk/gtk.h>
#include "../include/math3d.h"
#include "../include/matrix.h"
#include "../include/slpinc.h"
#include "rmansl.h"
#include "gpcl.h"
#include "gpcldlgs.h"

extern GtkWidget *mainwindow;
extern tpreferences gpreferences;

static GtkWidget *selection_dialog = NULL;
static GtkWidget *options_dialog = NULL;

static COLOR temp_color1;
static COLOR temp_color2;

static void initialize_rmansl(pattern_class *pat);
static void initialize_globals(pattern_class *pat);

static void rmansl_shade(pattern_class *pat, FLT s, FLT t, pvec3 rvec);
static void rmansl_configure(pattern_class *pat, GtkWidget *widget);
static void rmansl_update(pattern_class *pat);
static void rmansl_delete(pattern_class *pat);
static char* rmansl_name(void);

static GtkWidget* create_renderman_selection(pattern_class *pat);
static GtkWidget* create_options_window(pattern_class *pat);

static void
on_renderman_selection                 (GtkButton       *button,
                                        gpointer         user_data);
static void
on_renderman_selection_ok_button_clicked
                                        (GtkButton       *button,
                                        gpointer         user_data);
static void
on_renderman_selection_cancel_button_clicked
                                        (GtkButton       *button,
                                         gpointer         user_data);
static void
on_options_button_clicked
                                        (GtkButton       *button,
                                        gpointer         user_data);

static void
on_options_ok_button_clicked
                                        (GtkButton       *button,
                                        gpointer         user_data);

static void
on_options_cancel_button_clicked
                                        (GtkButton       *button,
                                        gpointer         user_data);

static int load_shader(pattern_class *pat);
static void delete_data(pattern_class *pat);

/* create a new control based on parameter type */
static void create_control(pattern_class *pat, TOKEN token, int pos,
                         PSHADER_DATA data, GtkWidget* box);

pattern_procs rmansl_procs =
   {rmansl_shade, rmansl_configure, rmansl_update,
    rmansl_delete, rmansl_name};

int rmansl_new(pattern_class *pat)
{
   pat->data = malloc(sizeof(rmansl_data));
   if (pat->data == NULL)
      return FALSE;

   pat->procs = &rmansl_procs;

   initialize_rmansl(pat);

   return TRUE;
}

void initialize_rmansl(pattern_class *pat)
{
   prmansl_data dat;

   dat = (prmansl_data)(pat->data);
   if (dat == NULL)
      return;

   dat->filename = NULL;
   dat->is_restricted = TRUE;
   dat->interpolate_spherical = TRUE;
   dat->list = NULL;
   dat->start = NULL;

   dat->info.shader_name = NULL;
   dat->info.float_num = 0;
   dat->info.vector_num = 0;
   dat->info.string_num = 0;
   dat->info.color_num = 0;

   /* set input color and opacity to white */
   dat->global_vars.Cs[0] = 1.0;
   dat->global_vars.Cs[1] = 1.0;
   dat->global_vars.Cs[2] = 1.0;

   dat->global_vars.Os[0] = 1.0;
   dat->global_vars.Os[1] = 1.0;
   dat->global_vars.Os[2] = 1.0;

   /* set light color to white */
   dat->global_vars.Cl[0] = 1.0;
   dat->global_vars.Cl[1] = 1.0;
   dat->global_vars.Cl[2] = 1.0;

   initialize_globals(pat);
}

void initialize_globals(pattern_class *pat)
{
   prmansl_data dat;

   dat = (prmansl_data)(pat->data);
   if (dat == NULL)
      return;

   vec3init(dat->global_vars.P);
   vec3init(dat->global_vars.dPdu);
   vec3init(dat->global_vars.dPdv);

   /* initialize texture coordinates and parameters to 0.0 */
   dat->global_vars.u = 0.0;
   dat->global_vars.v = 0.0;
   dat->global_vars.du = 0.0;
   dat->global_vars.dv = 0.0;
   dat->global_vars.s = 0.0;
   dat->global_vars.t = 0.0;
   dat->global_vars.dsdu = 0.0;
   dat->global_vars.dsdv = 0.0;
   dat->global_vars.dtdu = 0.0;
   dat->global_vars.dtdv = 0.0;

   /* initialize normal to point out of x-y plane */
   vec3init(dat->global_vars.N);
   dat->global_vars.N[2] = 1.0;
   vec3init(dat->global_vars.Ng);
   dat->global_vars.Ng[2] = 1.0;

   /* initialize light source */
   vec3init(dat->global_vars.L);
   dat->global_vars.L[0] = -0.4575;
   dat->global_vars.L[1] = -0.4575;
   dat->global_vars.L[2] = -0.7625;
   vec3norm(dat->global_vars.L);

   /* initialize incoming vector to point into plane of image */
   vec3init(dat->global_vars.I);
   dat->global_vars.I[2] = 1.0;

   /* initialize miscellaneous variables to zero */
   colorinit(dat->global_vars.Ci);
   colorinit(dat->global_vars.Oi);
   vec3init(dat->global_vars.E);
   dat->global_vars.E[2] = -10.0;
}

void rmansl_shade(pattern_class *pat, FLT s, FLT t, FLT *rvec)
{
   PDATA_STACK stack = NULL;

   prmansl_data dat;

   dat = (prmansl_data)(pat->data);
   if (dat == NULL)
   {
      interrupt = TRUE;
      return;
   }

   /* if start of execution does not exist do not render */
   if (dat->start == NULL)
   {
      interrupt = TRUE;
      return;
   }

   /* initialize data stack */
   datastack_init(&stack);

   /* reset input opacity global variable to white */
   dat->global_vars.Os[0] = 1.0;
   dat->global_vars.Os[1] = 1.0;
   dat->global_vars.Os[2] = 1.0;

   /* interpolate normal */
   if (dat->interpolate_spherical)
   {
      dat->global_vars.Ng[0] = s-0.5;
      dat->global_vars.Ng[1] = 0.5-t;
      dat->global_vars.Ng[2] = sqrt(-SQR(s)+s+t-SQR(t)-0.25);

      dat->global_vars.P[0] = dat->global_vars.Ng[0];
      dat->global_vars.P[1] = dat->global_vars.Ng[1];
      dat->global_vars.P[2] = dat->global_vars.Ng[2];

      dat->global_vars.I[0] = s-0.5;
      dat->global_vars.I[1] = 0.5-t;
      dat->global_vars.I[2] = sqrt(-SQR(s)+s+t-SQR(t)+0.25);

      /* normalize Ng (vec3norm doesn't work!) */
      vec3smult(2.0, dat->global_vars.Ng, dat->global_vars.Ng);
      vec3cpy(dat->global_vars.N, dat->global_vars.Ng);
   }

   /* set texture coordinates and parameters */
   dat->global_vars.v = dat->global_vars.t = t;
   dat->global_vars.u = dat->global_vars.s = s;

   /* attempt to execute RenderMan shader */
   if (!slparser_execute(dat->start, &stack,
                  &dat->global_vars, &dat->info))
   {
      interrupt = TRUE;
      return;
   }

   vec3cpy(rvec, dat->global_vars.Ci);
}

void rmansl_configure(pattern_class *pat, GtkWidget *container)
{
  GtkWidget *rmansl_box;
  GtkWidget *rmansl_shader_button;
  GtkWidget *rmansl_options_button;

   prmansl_data dat;
                                                                                   dat = (prmansl_data)(pat->data);
   if (dat == NULL)
      return;

  rmansl_box = gtk_vbox_new (FALSE, 3);
  gtk_widget_show (rmansl_box);
  gtk_container_add (GTK_CONTAINER (container), rmansl_box);
  gtk_container_border_width(GTK_CONTAINER(rmansl_box), 2);

  rmansl_shader_button = gtk_button_new_with_label ("Choose Shader...");
  gtk_widget_show (rmansl_shader_button);
  gtk_box_pack_start (GTK_BOX (rmansl_box), rmansl_shader_button, TRUE, TRUE, 0);
  gtk_signal_connect (GTK_OBJECT (rmansl_shader_button), "clicked",
                      GTK_SIGNAL_FUNC (on_renderman_selection),
                      pat);

   rmansl_options_button = gtk_button_new_with_label ("Options...");
   gtk_widget_show (rmansl_options_button);
   gtk_box_pack_start (GTK_BOX (rmansl_box), rmansl_options_button, 
                      TRUE, TRUE, 0);
   gtk_signal_connect (GTK_OBJECT (rmansl_options_button), "clicked",
                      GTK_SIGNAL_FUNC (on_options_button_clicked),
                      pat);

  pat->controls = rmansl_box;
}

void rmansl_update(pattern_class *pat)
{
   prmansl_data dat;
   int i;
                                                                                   dat = (prmansl_data)(pat->data);
   if (dat == NULL)
      return;

   for (i=0; i<dat->info.float_params; i++)
   {
      /* assign float value of spinner to the shader parameter */
      dat->info.float_data[i] =
          gtk_spin_button_get_value_as_float(
          GTK_SPIN_BUTTON(get_widget(mainwindow, dat->info.float_name[i])));
   }

   initialize_globals(pat);

   /* set texture derivatives */
   dat->global_vars.du = 1.0/gpreferences.width;
   dat->global_vars.dv = 1.0/gpreferences.height;
   dat->global_vars.dsdv = dat->global_vars.dsdu = 1.0;
   dat->global_vars.dtdv = dat->global_vars.dtdv = 1.0;
}

void rmansl_delete(pattern_class *pat)
{
   prmansl_data dat;

   dat = (prmansl_data)(pat->data);
   if (dat == NULL)
      return;

   delete_data(pat);
   free(dat->filename);
   free(dat);
   dat = NULL;

   gtk_widget_destroy(pat->controls);
}

char* rmansl_name(void)
{
   return "RenderMan Shader";
}

GtkWidget*
create_renderman_selection (pattern_class* pat)
{
  GtkWidget *renderman_selection;
  GtkWidget *renderman_selection_ok_button;
  GtkWidget *renderman_selection_cancel_button;

   prmansl_data dat;

   dat = (prmansl_data)(pat->data);
   if (dat == NULL)
      return NULL;

  renderman_selection = gtk_file_selection_new ("Load RenderMan Shader");
  gtk_object_set_data (GTK_OBJECT (renderman_selection), "renderman_selection",
renderman_selection);
  gtk_container_border_width (GTK_CONTAINER (renderman_selection), 10);
  gtk_signal_connect (GTK_OBJECT (renderman_selection), "delete_event",
             GTK_SIGNAL_FUNC (on_renderman_selection_cancel_button_clicked),
             NULL);
  gtk_signal_connect (GTK_OBJECT (renderman_selection), "destroy_event",
             GTK_SIGNAL_FUNC (on_renderman_selection_cancel_button_clicked),
             NULL);
  gtk_file_selection_hide_fileop_buttons (GTK_FILE_SELECTION 
       (renderman_selection));

  renderman_selection_ok_button = GTK_FILE_SELECTION 
       (renderman_selection)->ok_button;
  gtk_object_set_data (GTK_OBJECT (renderman_selection), "renderman_selection_ok
_button", renderman_selection_ok_button);
  gtk_widget_show (renderman_selection_ok_button);
  /* GTK_WIDGET_SET_FLAGS (renderman_selection_ok_button, GTK_CAN_DEFAULT); */
  gtk_signal_connect (GTK_OBJECT (renderman_selection_ok_button), "clicked",
             GTK_SIGNAL_FUNC (on_renderman_selection_ok_button_clicked),
             pat);

  renderman_selection_cancel_button = GTK_FILE_SELECTION (renderman_selection)->
cancel_button;
  gtk_object_set_data (GTK_OBJECT (renderman_selection), "renderman_selection_ca
ncel_button", renderman_selection_cancel_button);
  gtk_widget_show (renderman_selection_cancel_button);
  /* GTK_WIDGET_SET_FLAGS (renderman_selection_cancel_button, GTK_CAN_DEFAULT); */
  gtk_signal_connect (GTK_OBJECT (renderman_selection_cancel_button), "clicked",
              GTK_SIGNAL_FUNC (on_renderman_selection_cancel_button_clicked),
              NULL);

  return renderman_selection;
}

void
on_renderman_selection                 (GtkButton       *button,
                                        gpointer         user_data)
{
   if (selection_dialog == NULL)
   {
      selection_dialog = create_renderman_selection((pattern_class*)user_data);
      gtk_widget_show(selection_dialog);
   }
}

void
on_renderman_selection_ok_button_clicked
                                        (GtkButton       *button,
                                        gpointer         user_data)
{
   prmansl_data dat;
   pattern_class *pat;
   char *filen;

   pat = (pattern_class *)user_data;
   dat = (prmansl_data)(pat->data);
   if (dat == NULL)
      return;

   filen = gtk_file_selection_get_filename(
           GTK_FILE_SELECTION(selection_dialog));


   if (dat->filename != NULL)
   {
      free(dat->filename);
      dat->filename = NULL;
   }

   if (!slparser_calloc(strlen(filen)+1, sizeof(char), (void**)&dat->filename))
      return;

   strcpy(dat->filename, filen);

   gtk_widget_destroy(selection_dialog);
   selection_dialog = NULL;

   if (!load_shader(pat))
   {
      g_print("Error occurred loading %s\n", dat->filename);
      free(dat->filename);
      dat->filename = NULL;
      return;
   }
}

void
on_renderman_selection_cancel_button_clicked
                                        (GtkButton       *button,
                                        gpointer         user_data)
{
   gtk_widget_destroy(selection_dialog);
   selection_dialog = NULL;
}

GtkWidget* create_options_window(pattern_class *pat)
{
   GtkWidget *dialog;
   GtkWidget *vbox;
   GtkWidget *globals_action_area;
   GtkWidget *ok_button;
   GtkWidget *cancel_button;
   GtkWidget *model_sphere_button;
   GtkWidget *rmansl_is_restricted_button;
   GtkWidget *start_color_button;
   GtkWidget *light_color_button;

   prmansl_data dat;

   dat = (prmansl_data)(pat->data);
   if (dat == NULL)
      return NULL;

   dialog = gtk_dialog_new();
   gtk_window_set_title(GTK_WINDOW(dialog), "Options");
   gtk_window_set_policy(GTK_WINDOW(dialog), FALSE, TRUE, TRUE);
   gtk_signal_connect(GTK_OBJECT(dialog), "delete_event",
                      GTK_SIGNAL_FUNC(on_options_cancel_button_clicked),
                      NULL);

   gtk_signal_connect(GTK_OBJECT(dialog), "destroy",
                      GTK_SIGNAL_FUNC(on_options_cancel_button_clicked),
                      NULL);

   vbox = GTK_DIALOG(dialog)->vbox;
   gtk_widget_show(vbox);

   model_sphere_button = gtk_check_button_new_with_label ("Model Sphere");
   gtk_object_set_data (GTK_OBJECT (dialog),
                       "model_sphere_button",
                        model_sphere_button);
   gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(model_sphere_button), 
                       dat->interpolate_spherical);
   gtk_widget_show (model_sphere_button);
   gtk_box_pack_start (GTK_BOX (vbox), model_sphere_button,
                      TRUE, TRUE, 0);

   rmansl_is_restricted_button =
            gtk_check_button_new_with_label("Restrict Spinners");
   gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(rmansl_is_restricted_button),
                              dat->is_restricted);
   gtk_object_set_data (GTK_OBJECT (dialog),
                       "rmansl_is_restricted_button",
                        rmansl_is_restricted_button);
   gtk_widget_show (rmansl_is_restricted_button);
   gtk_box_pack_start (GTK_BOX (vbox), rmansl_is_restricted_button, TRUE,
                      TRUE, 0);

   /* use temporary color */
   colorcpy(temp_color1, dat->global_vars.Cs);

   start_color_button = gtk_button_new_with_label("Start Color");
   gtk_widget_show(start_color_button);
   gtk_box_pack_start(GTK_BOX(vbox), start_color_button,
                      TRUE, TRUE, 3);
   gtk_signal_connect(GTK_OBJECT(start_color_button), "clicked",
                      GTK_SIGNAL_FUNC(on_color_button_clicked),
                      temp_color1);

   /* use temporary color */
   colorcpy(temp_color2, dat->global_vars.Cl);

   light_color_button = gtk_button_new_with_label("Light Color");
   gtk_widget_show(light_color_button);
   gtk_box_pack_start(GTK_BOX(vbox), light_color_button,
                      TRUE, TRUE, 3);
   gtk_signal_connect(GTK_OBJECT(light_color_button), "clicked",
                      GTK_SIGNAL_FUNC(on_color_button_clicked),
                      temp_color2);


   globals_action_area = GTK_DIALOG(dialog)->action_area;
   gtk_widget_show(globals_action_area);
   gtk_container_border_width(GTK_CONTAINER(globals_action_area), 10);

   ok_button = gtk_button_new_with_label("OK");
   gtk_widget_show(ok_button);
   gtk_box_pack_start(GTK_BOX(globals_action_area), ok_button,
                      TRUE, TRUE, 0);
   gtk_signal_connect(GTK_OBJECT (ok_button), "clicked",
                      GTK_SIGNAL_FUNC (on_options_ok_button_clicked),
                      pat);

   cancel_button = gtk_button_new_with_label("Cancel");
   gtk_widget_show(cancel_button);
   gtk_box_pack_start(GTK_BOX(globals_action_area), cancel_button,
                      TRUE, TRUE, 0);
   gtk_signal_connect(GTK_OBJECT (cancel_button), "clicked",
                      GTK_SIGNAL_FUNC (on_options_cancel_button_clicked),
                      NULL);

   return dialog;
}

void
on_options_button_clicked
                                        (GtkButton       *button,
                                        gpointer         user_data)
{
   if (options_dialog == NULL)
   {
      options_dialog = create_options_window((pattern_class*)user_data);
      gtk_widget_show(options_dialog);
   }
}

void
on_options_ok_button_clicked
                                        (GtkButton       *button,
                                        gpointer         user_data)
{
   GtkWidget *current;

   prmansl_data dat;
   pattern_class *pat;

   pat = (pattern_class *)user_data;
   dat = (prmansl_data)(pat->data);
   if (dat == NULL)
      return;

   current = get_widget(options_dialog, "model_sphere_button");
   dat->interpolate_spherical = GTK_TOGGLE_BUTTON(current)->active;

   current = get_widget(options_dialog, "rmansl_is_restricted_button");
   dat->is_restricted = GTK_TOGGLE_BUTTON(current)->active;

   colorcpy(dat->global_vars.Cs, temp_color1);
   colorcpy(dat->global_vars.Cl, temp_color2);

   gtk_widget_destroy(options_dialog);
   options_dialog = NULL;
}

void
on_options_cancel_button_clicked
                                        (GtkButton       *button,
                                        gpointer         user_data)
{
   gtk_widget_destroy(options_dialog);
   options_dialog = NULL;
}

int
load_shader(pattern_class *pat)
{
   int i;
   GtkWidget *control_box = NULL;
   GtkWidget *hseparator = NULL;
   GtkWidget *label = NULL;
   GtkWidget *hbox = NULL;

   prmansl_data dat;

   dat = (prmansl_data)(pat->data);
   if (dat == NULL)
      return FALSE;

   /* if there is a shader currently loaded, delete it */

   if (dat->list != NULL)
      delete_data(pat);

   /* initialize the token list */
   tokenlist_init(&dat->list);

   /* get tokens from file */
   if (!tokenizer_get_tokens(&dat->list, dat->filename))
   {
      /* message_box("Could not open file"); */
      return FALSE;
   }

   /* attempt to initialize the shader data structure */
   if (!slparser_init(dat->list, &dat->info))
   {
      /* message_box("Error in parser\n initialization"); */
      g_print("Error in parser initialization\n");
      delete_data(pat);
      return FALSE;
   }

   /* set position to start */
   dat->start = dat->list;
   slparser_start(&dat->start);

   /* create controls */

   gtk_widget_hide(pat->controls);

   control_box = gtk_vbox_new (FALSE, 5);
   gtk_object_set_data (GTK_OBJECT (mainwindow), "control_box", control_box);
   gtk_widget_show (control_box);
   gtk_box_pack_start (GTK_BOX (pat->controls), control_box, TRUE, TRUE, 0);

   hseparator = gtk_hseparator_new ();
   gtk_widget_show (hseparator);
   gtk_box_pack_start (GTK_BOX (control_box), hseparator, TRUE, TRUE, 0);

   label = gtk_label_new(dat->info.shader_name);
   gtk_widget_show (label);
   gtk_box_pack_start (GTK_BOX (control_box), label, TRUE, TRUE, 0);

   /* add generator informaiton */
   if (dat->info.generator_name != NULL)
   {
      hbox = gtk_hbox_new(FALSE, 5);
      gtk_container_add(GTK_CONTAINER(control_box), hbox);
      gtk_widget_show(hbox);

      label = gtk_label_new("Compiler:");
      gtk_widget_show(label);
      gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, TRUE, 0);

      label = gtk_label_new(dat->info.generator_name);
      gtk_widget_show(label);
      gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, TRUE, 0);

      label = gtk_label_new(dat->info.version_name);
      gtk_widget_show(label);
      gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, TRUE, 0);
   }

   /* create float controls */
   for (i=0; i< dat->info.float_params; i++)
      create_control(pat, FLOAT_TOKEN, i, &dat->info, control_box);

   /* create color controls */
   for (i=0; i< dat->info.color_params; i++)
      create_control(pat, COLOR_TOKEN, i, &dat->info, control_box);

   gtk_widget_show(pat->controls);

   return TRUE;
}

void delete_data(pattern_class *pat)
{
   prmansl_data dat;
   GtkWidget *box = NULL;

   dat = (prmansl_data)(pat->data);
   if (dat == NULL)
      return;

   slparser_delete(&dat->info);
   tokenlist_delete(&dat->list);

   if((box = get_widget(mainwindow, "control_box")) != NULL)
   {
      gtk_widget_destroy(box);
      gtk_object_set_data(GTK_OBJECT(mainwindow), "control_box", NULL);
   }
}

void create_control(pattern_class *pat, TOKEN id, 
                         int pos, PSHADER_DATA data, GtkWidget* box)
{
   prmansl_data dat;

   dat = (prmansl_data)(pat->data);
   if (dat == NULL)                                                                   return;

   switch(id)
   {
      case FLOAT_TOKEN:
      {

         /* Variables:
            GtkWidget* horzbox - box to hold spinner and label
            GtkWidget** spinner - pointer spinner control
            GtkWidget* label - label of spinner control will be name of parameter
            GtkAdjustment* adjust - adjustment for spinner control
         */

         GtkWidget* horzbox = NULL;
         GtkWidget* spinner = NULL;
         GtkWidget* label = NULL;
         GtkAdjustment* adjust = NULL;

         /* create new container for label and spinner */

         horzbox = gtk_hbox_new(TRUE, 5);
         gtk_container_add(GTK_CONTAINER(box), horzbox);
         gtk_widget_show(horzbox);

         /* create new spin button. If is_restricted is set, set adjustment
            to range between 0 and 1, otherwise set range to 0 to 1000
         */

         if ((data->float_data[pos] <= 1.0) && (data->float_data[pos] >= 0.0)
           && dat->is_restricted)
         {
            adjust = (GtkAdjustment*)gtk_adjustment_new(0.0,
                               0.0, 1.0, gpreferences.increment, 0.1, 0.0);
            spinner = gtk_spin_button_new(adjust, gpreferences.increment, 
                               gpreferences.digits);
         }
         else
         {
            adjust = (GtkAdjustment*)gtk_adjustment_new(0.0,
                               0.0, 1000.0, gpreferences.increment, 10.0, 0.0);
            spinner = gtk_spin_button_new(adjust, gpreferences.increment, 
                               gpreferences.digits);
         }

         gtk_spin_button_set_wrap(GTK_SPIN_BUTTON(spinner), TRUE);
         gtk_spin_button_set_value(GTK_SPIN_BUTTON(spinner),
                    data->float_data[pos]);
         gtk_object_set_data (GTK_OBJECT (mainwindow), 
             data->float_name[pos], spinner);

         gtk_box_pack_start(GTK_BOX(horzbox), spinner, TRUE, TRUE, 0);
         gtk_widget_show(spinner);

         /* create new label for parameter name */

         label = gtk_label_new(data->float_name[pos]);
         gtk_box_pack_start(GTK_BOX(horzbox), label, FALSE, FALSE, 0);
         gtk_widget_show(label);

         break;
      }
      case COLOR_TOKEN:
      {

         /* Variables:
            GtkWidget* button - button control
         */

         GtkWidget* button = NULL;

         /* create new button */

         button = gtk_button_new_with_label(data->color_name[pos]);
         gtk_box_pack_start(GTK_BOX(box), button, TRUE, TRUE, 0);
         gtk_signal_connect (GTK_OBJECT(button), "clicked", 
              GTK_SIGNAL_FUNC (on_color_button_clicked),
              data->color_data[pos]); 
         gtk_widget_show(button);

         break;
      }
   }
}
