/**************************************
Matrix and 3D Math
Copyright (C) 1999  Peter Stuart

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
******************************************/

/**********MODULE MATH3D***************
Filename: math3d.h
Author: ophiuchus42@hotmail.com
Description: This module implements many 3d mathematical
   functions, as well as general math functions.
Modifications: 
  12/28/98   commented code
*************************/

#ifndef __MATH3D_H__
#define __MATH3D_H__

#include "matrix.h"
#include <math.h>


/****** DEFINES ******/

/* small number */
#define VERY_SMALL      0.0000000000000000001f

/* accuracy limit */
#define CORRECT         0.00001
#define CORRECT2        0.01
#define CORRECT3        0.001
#define CORRECT4        0.0001
#define ICORRECT        0.99999

/* 2*PI to 2 decimal places */
#define TWOPI2D    6.28

/* PI to many decimal places */
#ifndef PI
#define PI         3.14159265358979323846f
#endif /* PI */

/* 2*PI to many decimal places */
#define TWOPI      6.28318530717958647692f

/* 1/2 * PI to many decimal places */
#define HALFPI     1.57079632679489661923f

/* natural log of e base 2 */
#define LOG2E      1.4426950408889634074f

/* natural log to base 2 */
#define LOG2(x)    (LOG2E * log(x))

#define LOG05  -0.693147180559945  /* log(0.5) */

/* close to epsilon */
#define NOT_ZERO(x) (abs(x) > 0.0001)

/******* MATH MACROS *******/

/* square */
#define SQR(x) ((x)*(x))

#undef MIN
#define MIN(a, b) ( (a) < (b) ? (a) : (b))

#undef MAX
#define MAX(a, b) ( (a) > (b) ? (a) : (b))

#undef SGN
#define SGN(x)   ( ((x)<0) ? -1 : ((x) ? 1:0) )

#undef ABS
#define ABS(x) ((x) < 0 ? -(x) : (x))

#define s_curve(t) ( SQR(t) * (3. - ((t) + (t))) )
#define SMOOTHSTEP(t, a, b)  ( (t) <= 0 ? (a) : ((t) >= 1 ? (b) : ((a) + s_curve(t)*((b)-(a))) ) )
#define STEP(a, x) ((FLT)((x) >= (a)))

/* floor and ceiling functions returning int */
#define FLOOR(x) ((int)(x) - ((x) < 0 && (x) != (int)(x)))
#define CEIL(x) ((int)(x) + ((x) > 0 && (x) != (int)(x)))

#undef CLAMP
#define CLAMP(x,a,b) ((x) <= (a) ? (a) : ((x) >= (b) ? (b) : (x)))

#define LERP(t,x0,x1)  ((x0) + (t)*((x1)-(x0)))
#define PULSE(a,b,x) (STEP((a),(x)) - STEP((b),(x)))
#define BOXSTEP(a,b,x) (CLAMP(((x)-(a))/((b)-(a)),0,1))

#define BIAS(a,x) (pow((x),log((b))/LOG05)
#define GAIN(a,x) ((x) < 0.5 ? BIAS(1-(a),2*(x))/2 : BIAS(1-(a),2-2*(x))/2)
#define GAMMACORRECT(a,x) (pow((x), 1/(a)))

/* FUNCTIONS */

void cmix(PCOLOR c0, PCOLOR c1, FLT f, PCOLOR cr);
FLT mod(FLT a, FLT b);


/*
 * Declarations and preprocessor definitions for procedural texture function
 * library.
 * Darwyn Peachey, June, 1994.
 */
extern float spline(float x, int nknots, float *knot);

extern float gnoise(float x, float y, float z);
extern float gvnoise(float x, float y, float z);
extern float scnoise(float x, float y, float z);
extern float vcnoise(float x, float y, float z);
extern float vnoise(float x, float y, float z);

/* Perlin functions */
float noise1(float arg);
float noise2(float vec[]);
float noise3(float vec[]);

#endif /* __MATH3D_H_ */
