/**************************************
Matrix and 3D Math
Copyright (C) 1999  Peter Stuart

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
******************************************/


/**********MODULE MATRIX***************
Filename: matrix.h
Author: ophiuchus42@hotmail.com
Description: This module implements all of the matrix 
   and vector operations necessary for 3d graphics. The vector
   operations are all implemented as macros.
Modifications: 
  12/28/98   commented code
*************************/

#ifndef __MATRIX_H__
#define __MATRIX_H__

#include <string.h>

/****** TYPE DEFINITIONS ******/

/* type FLT is used where floating point is needed. It can be double
   or float, or some other type.
*/
/* typedef float FLT;
   #define MATSIZE  64
   #define VEC2SIZE 8
   #define VEC3SIZE 12
   #define VEC4SIZE 16
*/

typedef double FLT;
#define MATSIZE  128
#define VEC2SIZE 16
#define VEC3SIZE 24
#define VEC4SIZE 32

/****** VECTOR DEFINITIONS ******/
typedef FLT vec2[2];
typedef FLT *pvec2;
typedef FLT vec3[3];
typedef FLT *pvec3;
typedef FLT vec4[4];
typedef FLT *pvec4;

/* define color to be 3 element a vector */
typedef vec3 COLOR;
typedef FLT *PCOLOR;

/****** VECTOR MACROS ******/

/* copy two vectors */
#define vec2cpy(a, b) {(a)[0]=(b)[0]; (a)[1]=(b)[1];};
#define vec3cpy(a, b) {vec2cpy((a), (b)); (a)[2]=(b)[2];};
#define vec4cpy(a, b) {vec3cpy((a), (b)); (a)[3]=(b)[3];};
#define colorcpy vec3cpy

/* set elements in vector to 0.0 */
#define vec2init(a) {(a)[0]=(FLT)0.0; (a)[1]=(FLT)0.0;}; 
#define vec3init(a) {vec2init(a); (a)[2]=(FLT)0.0;};
#define vec4init(a) {vec3init(a); (a)[3]=(FLT)0.0;};
#define colorinit vec3init

/* print out vector elements */
#define vec2print(a) (printf("%f %f ", (a)[0], (a)[1]))
#define vec3print(a) (printf("%f %f %f ", (a)[0], (a)[1], (a)[2]))
#define vec4print(a) (printf("%f %f %f %f ", (a)[0], (a)[1], (a)[2], (a)[3]))
#define colorprint vec3print

/* vector dot product */
#define vec2dot(a, b) ((a)[0]*(b)[0] + (a)[1]*(b)[1])
#define vec3dot(a, b) (vec2dot(a, b) + (a)[2]*(b)[2])
#define vec4dot(a, b) (vec3dot(a, b) + (a)[3]*(b)[3])

/* vector addition */
#define vec2add(a, b, c) {(c)[0] = (a)[0]+(b)[0]; (c)[1] = (a)[1]+(b)[1];};
#define vec3add(a, b, c) {vec2add(a, b, c); (c)[2] = (a)[2]+(b)[2];};
#define vec4add(a, b, c) {vec3add(a, b, c); (c)[3] = (a)[3]+(b)[3];};
#define coloradd vec3add

/* vector subtraction */
#define vec2sub(a, b, c) {(c)[0] = (a)[0]-(b)[0]; (c)[1] = (a)[1]-(b)[1];};
#define vec3sub(a, b, c) {vec2sub(a, b, c); (c)[2] = (a)[2]-(b)[2];};
#define vec4sub(a, b, c) {vec3sub(a, b, c); (c)[3] = (a)[3]-(b)[3];};
#define colorsub vec3sub

/* vector cross product for 3-vectors */
#define vec3cross(a, b, c) {(c)[0] = (a)[1]*(b)[2] - (a)[2]*(b)[1]; \
                            (c)[1] = (a)[2]*(b)[0] - (a)[0]*(b)[2]; \
                            (c)[2] = (a)[0]*(b)[1] - (a)[1]*(b)[0];};

/* multiply vector by scalar */
#define vec2smult(x, a, c) {(c)[0]=(FLT)(x)*(a)[0]; (c)[1]=(FLT)(x)*(a)[1];};
#define vec3smult(x, a, c) {vec2smult(x, a, c); (c)[2]=(FLT)(x)*(a)[2];};
#define vec4smult(x, a, c) {vec3smult(x, a, c); (c)[3]=(FLT)(x)*(a)[3];};
#define colorsmult vec3smult

/* negative vector */
#define vec2neg(a) {(a)[0] = -(a)[0]; (a)[1] = -(a)[1];};
#define vec3neg(a) {vec2neg(a); (a)[2] = -(a)[2];};
#define vec4neg(a) {vec3neg(a); (a)[3] = -(a)[3];};

/* length of vector */
#define vec2mag(a) (FLT)(sqrt(vec2dot(a, a)))
#define vec3mag(a) (FLT)(sqrt(vec3dot(a, a)))
#define vec4mag(a) (FLT)(sqrt(vec4dot(a, a)))

/* normalize vector */
#define vec2norm(a) {vec2smult(1.0/vec2mag(a), (a), (a));};
#define vec3norm(a) {vec3smult(1.0/vec3mag(a), (a), (a));};
#define vec4norm(a) {vec4smult(1.0/vec4mag(a), (a), (a));};


/* component multiplication for color */
#define colormult(a, b, c) {(c)[0]=(a)[0]*(b)[0]; \
                            (c)[1]=(a)[1]*(b)[1]; \
                            (c)[2]=(a)[2]*(b)[2];};

/******** VECTOR FUNCTIONS *******/

/* turn vector to face second vector */
void faceforward(vec3 a, vec3 b, vec3 c);

/****** 4x4 MATRIX MACROS ******/

/* copy 4x4 matrix into another */
#define mat4cpy(a, b) memcpy((a), (b), MATSIZE);

/* initialize matrix to identity */
#define mat4init(a) {memset((a), 0, MATSIZE); \
        (a)[0][0] = (a)[1][1] = (a)[2][2] = (a)[3][3] = (FLT)1.0;};

/****** 3x3 MATRIX FUNCTIONS *******/

/* determinant of 3x3 matrix (should be static) */
FLT  mat3deter  (FLT mat[][3]);

/****** 4x4 MATRIX FUNCTIONS ******/

/* print out matrix */
void mat4print  (FLT mat[][4]);

/* matrix cross product */
void mat4cross  (FLT a[][4], FLT b[][4], FLT out[][4]);

/* matrix minor */
FLT  mat4minor  (FLT mat[][4], int i, int j);

/* matrix determinant */
FLT  mat4deter  (FLT mat[][4]);

/* element cofactor */
void mat4cofac  (FLT mat[][4], FLT out[][4]);

/* matrix transposition */
void mat4trans  (FLT mat[][4], FLT out[][4]);

/* matrix inverse */
void mat4inver  (FLT mat[][4], FLT out[][4]);

/* matrix addition */
void mat4add    (FLT a[][4], FLT b[][4], FLT out[][4]);

/* matrix subtraction */
void mat4sub    (FLT a[][4], FLT b[][4], FLT out[][4]);

/* matrix and vector homogeneous cross product */
void m4v4cross  (FLT a[][4], vec4 b, vec4 out);

/* vector and matrix homogeneous cross product */
void v4m4cross  (vec4 a, FLT b[][4], vec4 out);

/* matrix and vector non-homogeneous cross product */
void m4v4crossn (FLT a[][4], vec4 b, vec4 out);

/* vector and matrix non-homogeneous cross product */
void v4m4crossn (vec4 a, FLT b[][4], vec4 out);

/* multiply matrix by scalar */
void mat4smult  (FLT mat[][4], FLT val, FLT out[][4]);

#endif /* __MATRIX_H__ */
