/**************************************
Matrix and 3D Math
Copyright (C) 1999  Peter Stuart

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
******************************************/


/**********MODULE MATRIX***************
Filename: matrix.c
Author: ophiuchus42@hotmail.com
Description: This module implements all of the matrix 
   and vector operations necessary for 3d graphics. The vector
   operations are all implemented as macros.
Modifications: 
  12/28/98   commented code
*************************/

#include <stdio.h>
#include "../include/matrix.h"

/***** VECTOR FUNCTIONS ******/

/* FUNCTION faceforward
Parameters:
   vec3 a - IN - first input vector
   vec3 b - IN - second input vector
   vec3 c - OUT - output vector
Return Value:
   void
Local Variables:
   none
Description: This function compares the directions of the
   two input vectors. If a is facing away from b (angle > 90)
   the c becomes -a. If they are facing opposite directions,
   c becomes a. This is determined by looking at the dot product.
   Since a.b = a*b*cos(angle), and cos(angle) > 0 as angle
   goes from 0 to 90, the vectors are facing the same
   direction if a.b > 0.
*/
void faceforward (vec3 a, vec3 b, vec3 c)
{
   if (vec3dot(a, b) < 0.0)
   {
      c[0] = a[0];
      c[1] = a[1];
      c[2] = a[2];
   }
   else
   {
      c[0] = -a[0];
      c[1] = -a[1];
      c[2] = -a[2];
   }
}

/****** 3x3 MATRIX FUNCTIONS ******/

/* FUNCTION mat3deter
Parameters:
   FLT mat[][3] - IN - 3x3 matrix
Return Value:
   Returns the determinant of the 3x3 matrix as FLT
Local Variables:
   FLT answer - stores calculated determinant
Description:
   This function calculates the determinant of the input 3x3
   matrix. This is done using the common "trick" of multiplying
   diagonals usable only for 3x3 matrices. 
   (This should be static, and answer can be removed)
*/
FLT mat3deter (FLT mat[][3])
{
   FLT answer;
   answer = (FLT)((mat[0][0]*mat[1][1]*mat[2][2])
          + (mat[0][1]*mat[1][2]*mat[2][0])
          + (mat[0][2]*mat[1][0]*mat[2][1])
          - (mat[2][0]*mat[1][1]*mat[0][2])
          - (mat[2][1]*mat[1][2]*mat[0][0])
          - (mat[2][2]*mat[1][0]*mat[0][1]));
   return answer;
}

/****** 4x4 MATRIX FUNCTIONS *******/

/* FUNCTION mat4print
Parameters:
   FLT mat[][4] - IN - input 4x4 matrix
Return Value:
   void
Local Variables:
   int i - index variable
Description:
   This function prints out all of the elements of a
   4x4 matrix.
*/
void mat4print (FLT mat[][4])
{
   int i;
   for(i=0; i<4; i++)
      printf("%f %f %f %f\n", mat[i][0], mat[i][1], mat[i][2], mat[i][3]);
   printf("\n");
}

/* FUNCTION mat4cross
Parameters:
   FLT a[][4] - IN - first input 4x4 matrix
   FLT b[][4] - IN - second input 4x4 matrix
   FLT out[][4] - OUT - output 4x4 matrix
Return Value:
   void
Local Variables:
   none
Description:
   This function calculates the cross product of two
   4x4 matrices. This is done by multiplying each element
   of the row of the first matrix by the corresponding element
   of the column of the second matrix. The loops have been
   unrolled for efficiency.
*/
void mat4cross (FLT a[][4], FLT b[][4], FLT out[][4])
{
   out[0][0] = (a[0][0] * b[0][0])
             + (a[0][1] * b[1][0])
             + (a[0][2] * b[2][0])
             + (a[0][3] * b[3][0]);

   out[0][1] = (a[0][0] * b[0][1])
             + (a[0][1] * b[1][1])
             + (a[0][2] * b[2][1])
             + (a[0][3] * b[3][1]);

   out[0][2] = (a[0][0] * b[0][2])
             + (a[0][1] * b[1][2])
             + (a[0][2] * b[2][2])
             + (a[0][3] * b[3][2]);

   out[0][3] = (a[0][0] * b[0][3])
             + (a[0][1] * b[1][3])
             + (a[0][2] * b[2][3])
             + (a[0][3] * b[3][3]);


   out[1][0] = (a[1][0] * b[0][0])
             + (a[1][1] * b[1][0])
             + (a[1][2] * b[2][0])
             + (a[1][3] * b[3][0]);

   out[1][1] = (a[1][0] * b[0][1])
             + (a[1][1] * b[1][1])
             + (a[1][2] * b[2][1])
             + (a[1][3] * b[3][1]);

   out[1][2] = (a[1][0] * b[0][2])
             + (a[1][1] * b[1][2])
             + (a[1][2] * b[2][2])
             + (a[1][3] * b[3][2]);

   out[1][3] = (a[1][0] * b[0][3])
             + (a[1][1] * b[1][3])
             + (a[1][2] * b[2][3])
             + (a[1][3] * b[3][3]);


   out[2][0] = (a[2][0] * b[0][0])
             + (a[2][1] * b[1][0])
             + (a[2][2] * b[2][0])
             + (a[2][3] * b[3][0]);

   out[2][1] = (a[2][0] * b[0][1])
             + (a[2][1] * b[1][1])
             + (a[2][2] * b[2][1])
             + (a[2][3] * b[3][1]);

   out[2][2] = (a[2][0] * b[0][2])
             + (a[2][1] * b[1][2])
             + (a[2][2] * b[2][2])
             + (a[2][3] * b[3][2]);

   out[2][3] = (a[2][0] * b[0][3])
             + (a[2][1] * b[1][3])
             + (a[2][2] * b[2][3])
             + (a[2][3] * b[3][3]);


   out[3][0] = (a[3][0] * b[0][0])
             + (a[3][1] * b[1][0])
             + (a[3][2] * b[2][0])
             + (a[3][3] * b[3][0]);

   out[3][1] = (a[3][0] * b[0][1])
             + (a[3][1] * b[1][1])
             + (a[3][2] * b[2][1])
             + (a[3][3] * b[3][1]);

   out[3][2] = (a[3][0] * b[0][2])
             + (a[3][1] * b[1][2])
             + (a[3][2] * b[2][2])
             + (a[3][3] * b[3][2]);

   out[3][3] = (a[3][0] * b[0][3])
             + (a[3][1] * b[1][3])
             + (a[3][2] * b[2][3])
             + (a[3][3] * b[3][3]);

}

/* FUNCTION mat4minor
Parameters:
   FLT mat[][4] - IN - input 4x4 matrix
   int row - IN - row of element to calculate minor
   int col - IN - column of element to calculate minor
Return Value:
   Returns the minor as FLT
Local Variables:
   FLT mat3x3[3][3] - 3x3 submatrix
   int rowPos - current row
   int colPos - current column
   int i - index variable
   int j - index variable
Description:
   This function calculates the minor of a position within the matrix.
   The minor is the determinant of the 3x3 matrix formed of all elements
   of the 4x4 matrix which are not part of the row or column of the
   input element. (This needs to be optimized, since it is called many
   times for the inverse operation).
*/
FLT  mat4minor (FLT mat[][4], int row, int col)
{
   FLT mat3x3[3][3];
   int rowPos=0, colPos=0;
   int i, j;

   for (i=0; i<4; i++)
   {
      for (j=0; j<4; j++)
      {
         if ((i!=row) && (j!=col))
         {
            mat3x3[rowPos][colPos] = mat[i][j];
            colPos++;
            if (colPos > 2)
            {
               colPos = 0;
               rowPos++;
            }
         }
      }
   }
   return mat3deter(mat3x3);
}

/* FUNCTION mat4deter
Parameters:
   FLT mat[][4] - IN - input 4x4 matrix
Return Value:
   Returns the calculated determinant
Local Variables:
   FLT val - value of determinant
Description:
   This function calculates the determinant of a given 4x4 matrix.
   This is done using Laplace development. (Can be optimized a little
   by removing val implementing minor as a macro)
*/
FLT  mat4deter (FLT mat[][4])
{
  FLT val = 0.0;

  val += mat[0][0] * mat4minor(mat, 0, 0);
  val -= mat[0][1] * mat4minor(mat, 0, 1);
  val += mat[0][2] * mat4minor(mat, 0, 2);
  val -= mat[0][3] * mat4minor(mat, 0, 3);

  return val;
}

/* FUNCTION mat4cofac
Parameters:
   FLT mat[][4] - IN - input 4x4 matrix
   FLT out[][4] - OUT - output cofactor matrix
Return Value:
   void
Local Variables:
   none
Description:
   This function creates the cofactor matrix of the input
   matrix. Each element of the cofactor matrix is the signed 
   minor of the element.
*/
void mat4cofac (FLT mat[][4], FLT out[][4])
{
   out[0][0] =  mat4minor(mat, 0, 0);
   out[0][1] = -mat4minor(mat, 0, 1);
   out[0][2] =  mat4minor(mat, 0, 2);
   out[0][3] = -mat4minor(mat, 0, 3);

   out[1][0] = -mat4minor(mat, 1, 0);
   out[1][1] =  mat4minor(mat, 1, 1);
   out[1][2] = -mat4minor(mat, 1, 2);
   out[1][3] =  mat4minor(mat, 1, 3);

   out[2][0] =  mat4minor(mat, 2, 0);
   out[2][1] = -mat4minor(mat, 2, 1);
   out[2][2] =  mat4minor(mat, 2, 2);
   out[2][3] = -mat4minor(mat, 2, 3);

   out[3][0] = -mat4minor(mat, 3, 0);
   out[3][1] =  mat4minor(mat, 3, 1);
   out[3][2] = -mat4minor(mat, 3, 2);
   out[3][3] =  mat4minor(mat, 3, 3);
}

/* FUNCTION mat4trans
Parameters:
   FLT mat[][4] - IN - input matrix
   FLT out[][4] - OUT - output matrix
Return Value:
   void
Local Variables:
   none
Description:
   This function constructs the transposition matrix of
   the input matrix. This is created by reversing the row
   and column values of each element.
*/
void mat4trans (FLT mat[][4], FLT out[][4])
{
   out[0][0] = mat[0][0];
   out[0][1] = mat[1][0];
   out[0][2] = mat[2][0];
   out[0][3] = mat[3][0];

   out[1][0] = mat[0][1];
   out[1][1] = mat[1][1];
   out[1][2] = mat[2][1];
   out[1][3] = mat[3][1];

   out[2][0] = mat[0][2];
   out[2][1] = mat[1][2];
   out[2][2] = mat[2][2];
   out[2][3] = mat[3][2];

   out[3][0] = mat[0][3];
   out[3][1] = mat[1][3];
   out[3][2] = mat[2][3];
   out[3][3] = mat[3][3];
}

/* FUNCTION mat4inver
Parameters:
   FLT mat[][4] - IN - input matrix
   FLT out[][4] - OUT - inverse matrix
Return Value:
   void
Local Variables:
   FLT comat[4][4] - cofactor matrix of input
   FLT trans[4][4] - transposition of cofactor matrix
Description:
   This function constructs the inverse matrix of the 
   input 4x4 matrix. The inverse is found by using:
   transpose(cofactor(matrix))/determinant(matrix)
   (This can be sped up by implementing mat4cofac,
   mat4trans, and mat4smult as macros)
*/
void mat4inver (FLT mat[][4], FLT out[][4])
{
   FLT comat[4][4];
   FLT trans[4][4];

   /* find cofactor matrix */
   mat4cofac(mat, comat);

   /* take transpose of cofactor */
   mat4trans(comat, trans);
   
   /* divide transpose of cofactor by the determinant of original matrix */
   mat4smult(trans, 1/mat4deter(mat), out);
}

/* FUNCTION mat4add
Parameters:
   FLT a[][4] - IN - first input matrix
   FLT b[][4] - IN - second input matrix
   FLT out[][4] - OUT - output matrix
Return Value:
   void
Local Variables:
   none
Description:
   This functions adds the elements of a to the
   elements of b and assigns the sum to out.
*/
void mat4add (FLT a[][4], FLT b[][4], FLT out[][4])
{
   out[0][0] = a[0][0] + b[0][0];
   out[0][1] = a[0][1] + b[0][1];
   out[0][2] = a[0][2] + b[0][2];
   out[0][3] = a[0][3] + b[0][3];

   out[1][0] = a[1][0] + b[1][0];
   out[1][1] = a[1][1] + b[1][1];
   out[1][2] = a[1][2] + b[1][2];
   out[1][3] = a[1][3] + b[1][3];

   out[2][0] = a[2][0] + b[2][0];
   out[2][1] = a[2][1] + b[2][1];
   out[2][2] = a[2][2] + b[2][2];
   out[2][3] = a[2][3] + b[2][3];

   out[3][0] = a[3][0] + b[3][0];
   out[3][1] = a[3][1] + b[3][1];
   out[3][2] = a[3][2] + b[3][2];
   out[3][3] = a[3][3] + b[3][3];
}

/* FUNCTION mat4sub
Parameters:
   Flt a[][4] - IN - first input matrix
   Flt b[][4] - IN - second input matrix
   FLT out[][4] - OUT - output matrix
Return Value:
   void
Local Variables:
   none
Description:
   This function subtracts each element of the first
   matrix by the element of the second matrix and
   assigns the difference to the output matrix.
*/
void mat4sub (FLT a[][4], FLT b[][4], FLT out[][4])
{
   out[0][0] = a[0][0] - b[0][0];
   out[0][1] = a[0][1] - b[0][1];
   out[0][2] = a[0][2] - b[0][2];
   out[0][3] = a[0][3] - b[0][3];

   out[1][0] = a[1][0] - b[1][0];
   out[1][1] = a[1][1] - b[1][1];
   out[1][2] = a[1][2] - b[1][2];
   out[1][3] = a[1][3] - b[1][3];

   out[2][0] = a[2][0] - b[2][0];
   out[2][1] = a[2][1] - b[2][1];
   out[2][2] = a[2][2] - b[2][2];
   out[2][3] = a[2][3] - b[2][3];

   out[3][0] = a[3][0] - b[3][0];
   out[3][1] = a[3][1] - b[3][1];
   out[3][2] = a[3][2] - b[3][2];
   out[3][3] = a[3][3] - b[3][3];
}

/* FUNCTION m4v4cross
Parameters:
   FLT a[][4] - IN - input matrix
   vec4 b - IN - input 4-vector
   vec4 out - OUT - output 4-vector
Return Value:
   void
Local Variables:
   none
Description:
   This function performs the cross multiplication of the
   input 4-vector with the input matrix. The vector is assumed
   to be a column vector.
*/
void m4v4cross (FLT a[][4], vec4 b, vec4 out)
{
   out[0] = a[0][0]*b[0] + a[0][1]*b[1]
          + a[0][2]*b[2] + a[0][3]*b[3];
   out[1] = a[1][0]*b[0] + a[1][1]*b[1]
          + a[1][2]*b[2] + a[1][3]*b[3];
   out[2] = a[2][0]*b[0] + a[2][1]*b[1]
          + a[2][2]*b[2] + a[2][3]*b[3];
   out[3] = a[3][0]*b[0] + a[3][1]*b[1]
          + a[3][2]*b[2] + a[3][3]*b[3];
}

/* FUNCTION m4v4cross
Parameters:
   vec4 a - IN - input 4-vector
   FLT b[][4] - IN - input matrix
   vec4 out - OUT - output 4-vector
Return Value:
   void
Local Variables:
   none
Description:
   This function performs the cross multiplication of the
   input 4-vector with the input matrix. The vector is assumed to
   be a row vector.
*/
void v4m4cross (vec4 a, FLT b[][4], vec4 out)
{
   out[0] = a[0]*b[0][0] + a[1]*b[1][0]
          + a[2]*b[2][0] + a[3]*b[3][0];
   out[1] = a[0]*b[0][1] + a[1]*b[1][1]
          + a[2]*b[2][1] + a[3]*b[3][1];
   out[2] = a[0]*b[0][2] + a[1]*b[1][2]
          + a[2]*b[2][2] + a[3]*b[3][2];
   out[3] = a[0]*b[0][3] + a[1]*b[1][3]
          + a[2]*b[2][3] + a[3]*b[3][3];
}

/* FUNCTION m4v4cross
Parameters:
   FLT a[][4] - IN - input matrix
   vec4 b - IN - input 4-vector
   vec4 out - OUT - output 4-vector
Return Value:
   void
Local Variables:
   none
Description:
   This function performs the cross multiplication of the
   input 4-vector with the input matrix. This cross multiplication
   is homogeneous, in other words, the fourth output vector must
   have a value of 1.0. The vector is assumed to be a column vector.
*/
void m4v4crossn (FLT a[][4], vec4 b, vec4 out)
{
   out[3] = a[3][0]*b[0] + a[3][1]*b[1]
          + a[3][2]*b[2] + a[3][3]*b[3];

   out[0] = (a[0][0]*b[0] + a[0][1]*b[1]
          +  a[0][2]*b[2] + a[0][3]*b[3]) / out[3];
   out[1] = (a[1][0]*b[0] + a[1][1]*b[1]
          +  a[1][2]*b[2] + a[1][3]*b[3]) / out[3];
   out[2] = (a[2][0]*b[0] + a[2][1]*b[1]
          +  a[2][2]*b[2] + a[2][3]*b[3]) / out[3];
   out[3] = (FLT)1.0;
}

/* FUNCTION m4v4cross
Parameters:
   vec4 a - IN - input 4-vector
   FLT b[][4] - IN - input matrix
   vec4 out - OUT - output 4-vector
Return Value:
   void
Local Variables:
   none
Description:
   This function performs the cross multiplication of the
   input 4-vector with the input matrix. This cross multiplication
   is homogeneous, in other words, the fourth output vector must
   have a value of 1.0. The vector is assumed to be a row vector.
*/
void v4m4crossn (vec4 a, FLT b[][4], vec4 out)
{
   out[3] = a[0]*b[0][3] + a[1]*b[1][3]
          + a[2]*b[2][3] + a[3]*b[3][3];

   out[0] = (a[0]*b[0][0] + a[1]*b[1][0]
          +  a[2]*b[2][0] + a[3]*b[3][0]) / out[3];
   out[1] = (a[0]*b[0][1] + a[1]*b[1][1]
          +  a[2]*b[2][1] + a[3]*b[3][1]) / out[3];
   out[2] = (a[0]*b[0][2] + a[1]*b[1][2]
          +  a[2]*b[2][2] + a[3]*b[3][2]) / out[3];
   out[3] = (FLT)1.0;
}

/* FUNCTION mat4smult
Parameters:
   FLT mat[][4] - IN - input matrix
   FLT val - IN - input scalar
   FLT out[][4] - OUT - output matrix
Return Value:
   void
Local Variables:
   none
Description:
   This function multiplies each element of the input matrix
   by the input scalar and assigns the result to the output
   matrix.
*/
void mat4smult (FLT mat[][4], FLT val, FLT out[][4])
{
   out[0][0] = mat[0][0] * val;
   out[0][1] = mat[0][1] * val;
   out[1][2] = mat[0][2] * val;
   out[0][3] = mat[0][3] * val;

   out[1][0] = mat[1][0] * val;
   out[1][1] = mat[1][1] * val;
   out[1][2] = mat[1][2] * val;
   out[1][3] = mat[1][3] * val;

   out[2][0] = mat[2][0] * val;
   out[2][1] = mat[2][1] * val;
   out[2][2] = mat[2][2] * val;
   out[2][3] = mat[2][3] * val;

   out[3][0] = mat[3][0] * val;
   out[3][1] = mat[3][1] * val;
   out[3][2] = mat[3][2] * val;
   out[3][3] = mat[3][3] * val;
}
