/***************************
slparser
Copyright (C) 1999  Peter Stuart

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*****************************/


/**********MODULE DATASTACK***************
Filename: datastack.c
Author: ophiuchus42@hotmail.com
Description: This module implements stack operations for data of
  types string, float, and vector. This is used in the execution
  of RenderMan shading language shaders compiled with BMRT.
Modifications: 
  12/23/98   commented code
  12/30/98   added code for free list
*************************/

#include "tokens.h"
#include "types.h"
#include "datastack.h"


/***** STATIC VARIABLES *****/
/* static PDATA_STACK freelist - list of free elements, used
   to avoid repetative memory allocation.
   static int num_stacks - number of stacks initialized. When
   last stack is deleted, freelist is also deleted.
*/

static PDATA_STACK freelist = NULL;
static int num_stacks = 0;

#ifdef __DEBUGDATASTACK__
   static int num_created;
   static int num_deleted;
#endif /* __DEBUGDATASTACK__ */


/***** STATIC FUNCTIONS ******/

/* create new element */
static int  datastack_create_element(PPDATA_STACK element);

/* expand freelist */
static int expand_freelist(void);

/***** "STATIC" MACROS ******/

/* get element from free list */
#define GET_FREE_ELEMENT(a) { if (freelist == NULL) { \
                  if (!expand_freelist()) return FALSE; } \
                  (*(a)) = freelist; freelist = (*(a))->next; };

/* free element (put back on free list) */
#define FREE_ELEMENT(a) { (a)->next = freelist; freelist = (a); };                 

/* FUNCTION datastack_init
Parameters:
   PPDATA_STACK stack - OUT - pointer to top of stack
Return Value:
   void
Local Variables:
   none
Description:
   This function initializes the datastack top of stack 
*/
void  datastack_init(PPDATA_STACK stack)
{
   int i = 0;
   PDATA_STACK current = NULL;

   #ifdef __DEBUGDATASTACK__
      num_created = 0;
      num_deleted = 0;
      dmessage("Initilizing datastack.\n");
   #endif /* __DEBUGDATASTACK__ */

   /* initialize top of stack to NULL */
   *stack = NULL; 

   /* increment number of stacks initialized */
   num_stacks++;

   /* create free list with FREE_START elements */
   if (freelist != NULL)
      return;

   if (!datastack_create_element(&freelist))
      return;

   current = freelist;

   for (i=0; i<FREE_START-1; i++)
   {
      if (!datastack_create_element(&current->next))
         return;
      current = current->next;
   }
}

/* FUNCTION datastack_popv
Parameters:
   PPDATA_STACK top - IN/OUT - top of stack
   pvec3 v - OUT - 3 element vector which becomes the element popped
      from the stack
Return Value:
   TRUE if pop operation succeeded, FALSE if it failed.
Local Variables:
   PDATA_STACK temp - used to store top element while the previous
      element's next pointer is updated
Description:
   This function implements the pop operation for vectors. The
   function returns false for failed pop. The conditions for 
   failure include NULL input parameters, or improper type for the
   top element of the stack.
*/
int datastack_popv(PPDATA_STACK top, pvec3 v)
{
   PDATA_STACK temp;

   /* fail if one of the input parameters is NULL */
   if ((*top == NULL) || (v == NULL))
   {
      warning_message("Datastack warning: improper input to pop\n");
      return FALSE;
   }

   /* fail if the type of the top element is not a vector */
   if ((*top)->data_id != VECTOR_TOKEN)
   {
      warning_message("Datastack warning: improper top type\n");
      return FALSE;
   }

   /* copy top element vector into input parameter v */
   vec3cpy(v, (*top)->vector_value);

   #ifdef __DEBUGDATASTACK__
      dmessage("Deleting vector element ");
      vec3print(v);
      dmessage("\n");
   #endif /* __DEBUGDATASTACK__ */

   /* update links in the stack */
   temp = *top;
   *top = (*top)->next;

   /* free top element of stack */
   FREE_ELEMENT(temp);

   return TRUE;
}

/* FUNCTION datastack_popf
Parameters:
   PPDATA_STACK top - IN/OUT - top of stack
   FLT* f - OUT - pointer to the float that will contain top element
Return Value:
   TRUE if pop operation succeeded, FALSE if it failed.
Local Variables:
   PDATA_STACK temp - used to store top element while the previous
      element's next pointer is updated
Description:
   This function implements the pop operation for floats. The
   function returns false for a failed pop. The conditions for 
   failure include NULL input parameters, or improper type for the
   top element of the stack.
*/
int datastack_popf(PPDATA_STACK top, FLT* f)
{
   PDATA_STACK temp;

   /* fail if one of the input parameters is NULL */
   if ((*top == NULL) || (f == NULL))
   {
      warning_message("Datastack warning: improper input to pop\n");
      return FALSE;
   }

   /* fail if the type of the top element is not a float */
   if ((*top)->data_id != FLOAT_TOKEN)
   {
      warning_message("Datastack warning: improper top type\n");
      return FALSE;
   }

   /* copy top element float into f */
   *f = (*top)->float_value;

   #ifdef __DEBUGDATASTACK__
      dmessage("Deleting float element %f\n", *f);
   #endif /* __DEBUGDATASTACK__ */

   /* update links in the stack */
   temp = *top;
   *top = (*top)->next;

   /* free top element of stack */
   FREE_ELEMENT(temp);

   return TRUE;
}

/* FUNCTION datastack_pops
Parameters:
   PPDATA_STACK top - IN/OUT - top of stack
   char** s - OUT - pointer to string
Return Value:
   TRUE if pop operation succeeded, FALSE if it failed.
Local Variables:
   PDATA_STACK temp - used to store top element while the previous
      element's next pointer is updated
Description:
   This function implements the pop operation for strings. The
   function returns false for a failed pop. The conditions for 
   failure include NULL input parameters, improper type for the
   top element of the stack, or inability to allocate memory for
   copied string.
*/
int datastack_pops(PPDATA_STACK top, char** s)
{
   PDATA_STACK temp;

   /* fail if one of the input parameters is NULL */
   if ((*top == NULL) || (s == NULL))
   {
      warning_message("Datastack warning: improper input to pop\n");
      return FALSE;
   }

   /* fail if the type of the top element is not a string */
   if ((*top)->data_id != STRING_TOKEN)
   {
      warning_message("Datastack warning: improper top type\n");
      return FALSE;
   }

   /* attempt to allocate memory for copied string */
   if (((*s) = (char*)calloc(strlen((*top)->string_value)+1, 
                              sizeof(char))) == NULL) 
   {
      error_message("Datastack error: Could not allocate memory");
      return FALSE;
   }
   /* copy top element string into s */
   strcpy(*s, (*top)->string_value);

   #ifdef __DEBUGDATASTACK__
      dmessage("Deleting string element %s\n", *s);
   #endif /* __DEBUGDATASTACK__ */

   /* update links in the stack */
   temp = *top;
   *top = (*top)->next;

   /* free element from stack */
   free(temp->string_value);
   temp->string_value = NULL;
   FREE_ELEMENT(temp);

   return TRUE;
}

/* FUNCTION datastack_popc
Parameters:
   PPDATA_STACK top - IN/OUT - top of stack
   PCOLOR c - OUT - pointer to the color
Return Value:
   TRUE if pop operation succeeded, FALSE if it failed.
Local Variables:
   PDATA_STACK temp - used to store top element while the previous
      element's next pointer is updated
Description:
   This function implements the pop operation for colors. The
   function returns false for a failed pop. The conditions for 
   failure include NULL input parameters, or improper type for the
   top element of the stack.
*/
int datastack_popc(PPDATA_STACK top, PCOLOR c)
{
   PDATA_STACK temp;

   /* fail if one of the input parameters is NULL */
   if ((*top == NULL) || (c == NULL))
   {
      warning_message("Datastack warning: improper input to pop\n");
      return FALSE;
   }

   /* fail if the type of the top element is not a color */
   if ((*top)->data_id != COLOR_TOKEN)
   {
      warning_message("Datastack warning: improper top type\n");
      return FALSE;
   }

   /* copy top element vector into input parameter c */
   colorcpy(c, (*top)->color_value);

   #ifdef __DEBUGDATASTACK__
      dmessage("Deleting color element ");
      colorprint(c);
      dmessage("\n");
   #endif /* __DEBUGDATASTACK__ */

   /* update links in the stack */
   temp = *top;
   *top = (*top)->next;

   /* free element from stack */
   FREE_ELEMENT(temp);

   return TRUE;
}

/* FUNCTION datastack_pushv
Parameters:
   PPDATA_STACK top - IN/OUT - pointer to top of stack
   pvec3 v - IN - 3 element vector to push
Return Value:
   TRUE if operation succeeds, FALSE if it fails
Local Variables:
   PDATA_STACK new_element - a pointer to the newly allocated element
   PDATA_STACK temp - used to keep track of top of stack while links
      are updataed
Description:
   This function implements push stack operation for vectors.
   The conditions for failure include NULL input vector, or
   failure to create a new stack element.
*/
int datastack_pushv(PPDATA_STACK top, pvec3 v)
{
   PDATA_STACK new_element = NULL;
   PDATA_STACK temp = NULL;

   /* fail if input vector is NULL */
   if (v == NULL)
   {
      error_message("Datastack warning: improper input to push\n");
      return FALSE;
   }

   GET_FREE_ELEMENT(&new_element);

   #ifdef __DEBUGDATASTACK__
      dmessage("Creating element for vector ");
      vec3print(v);
      dmessage("\n");
   #endif /* __DEBUGDATASTACK__ */

   /* assign type of new element */
   new_element->data_id = VECTOR_TOKEN;

   /* copy input vector into new element */
   vec3cpy(new_element->vector_value, v);

   /* adjust links in stack to add new element to top */
   temp = *top;
   *top = new_element;
   new_element->next = temp; 

   return TRUE;
}

/* FUNCTION datastack_pushf
Parameters:
   PPDATA_STACK top - IN/OUT - pointer to top of stack
   FLT* f - IN - pointer to input float
Return Value:
   TRUE if operation succeeds, FALSE if it fails
Local Variables:
   PDATA_STACK new_element - a pointer to the newly allocated element
   PDATA_STACK temp - used to keep track of top of stack while links
      are updataed
Description:
   This function implements push stack operation for floats.
   The conditions for failure include NULL input float, or
   failure to create a new stack element.
*/

int datastack_pushf(PPDATA_STACK top, FLT* f)
{
   PDATA_STACK new_element = NULL;
   PDATA_STACK temp = NULL;

   /* fail if input float pointer is NULL */
   if (f == NULL)
   {
      warning_message("Datastack warning: improper input to push\n");
      return FALSE;
   }

   GET_FREE_ELEMENT(&new_element);

   #ifdef __DEBUGDATASTACK__
      dmessage("Creating float element %f\n", (*f));
   #endif /* __DEBUGDATASTACK__ */

   /* assign type of new element */
   new_element->data_id = FLOAT_TOKEN;

   /* copy input float into new element */
   new_element->float_value = *f;

   /* adjust links in stack to add new element to top */
   temp = *top;
   *top = new_element;
   new_element->next = temp;

   return TRUE;
}

/* FUNCTION datastack_pushs
Parameters:
   PPDATA_STACK top - IN/OUT - pointer to top of stack
   char** s - IN - pointer to input string
Return Value:
   TRUE if operation succeeds, FALSE if it fails
Local Variables:
   PDATA_STACK new_element - a pointer to the newly allocated element
   PDATA_STACK temp - used to keep track of top of stack while links
      are updataed
Description:
   This function implements push stack operation for strings.
   The conditions for failure include NULL input string pointer, or
   failure to create a new stack element.
*/
int datastack_pushs(PPDATA_STACK top, char** s)
{
   PDATA_STACK new_element = NULL;
   PDATA_STACK temp = NULL;

   /* fail if input string pointer is NULL */
   if (s == NULL)
   {
      warning_message("Datastack warning: improper input to push\n");
      return FALSE;
   }

   GET_FREE_ELEMENT(&new_element);

   #ifdef __DEBUGDATASTACK__
      dmessage("Creating string element %s\n", *s);
   #endif /* __DEBUGDATASTACK__ */

   /* assign type of new element */
   new_element->data_id = STRING_TOKEN;

   /* allocate memory in new element for string */
   if ((new_element->string_value
                = (char*)calloc(strlen(*s)+1, sizeof(char))) == NULL)
   {
      error_message("Datastack error: Could not allocate memory");
      return FALSE;
   }
   /* copy input string into new element */
   strcpy(new_element->string_value, *s);

   /* adjust links in stack to add new element to top */
   temp = *top;
   *top = new_element;
   new_element->next = temp;

   return TRUE;
}

/* FUNCTION datastack_pushc
Parameters:
   PPDATA_STACK top - IN/OUT - pointer to top of stack
   PCOLOR c - IN - input color
Return Value:
   TRUE if operation succeeds, FALSE if it fails
Local Variables:
   PDATA_STACK new_element - a pointer to the newly allocated element
   PDATA_STACK temp - used to keep track of top of stack while links
      are updataed
Description:
   This function implements push stack operation for color.
   The conditions for failure include NULL input color, or
   failure to create a new stack element.
*/
int datastack_pushc(PPDATA_STACK top, PCOLOR c)
{
   PDATA_STACK new_element = NULL;
   PDATA_STACK temp = NULL;

   /* fail if input color is NULL */
   if (c == NULL)
   {
      warning_message("Datastack error: improper input to push\n");
      return FALSE;
   }

   GET_FREE_ELEMENT(&new_element);

   #ifdef __DEBUGDATASTACK__
      dmessage("Creating element for color ");
      colorprint(c);
      dmessage("\n");
   #endif /* __DEBUGDATASTACK__ */

   /* assign type of new element */
   new_element->data_id = COLOR_TOKEN;

   /* copy input color into new element */
   colorcpy(new_element->color_value, c);

   /* adjust links in stack to add new element to top */
   temp = *top;
   *top = new_element;
   new_element->next = temp;

   return TRUE;
}

/* FUNCTION datastack_create_element
Parameters:
   PPDATA_STACK element - OUT - pointer new stack element
Return Value:
   TRUE if creation succeeded, FALSE if it failed
Local Variables:
   PDATA_STACK new_element - pointer to newly allocated memory
Description:
   This function allocates memory for a new element, and assigns
   all pointers in element to NULL.
*/
int datastack_create_element(PPDATA_STACK element)
{
   PDATA_STACK new_element = NULL;

   /* allocate memory for stack element */
   if ((new_element = (PDATA_STACK)malloc(sizeof(DATA_STACK))) == NULL)
   {
      error_message("Datastack error: Could not allocate memory");
      return FALSE;
   }
   /* initialize structure fields */
   new_element->string_value = NULL;
   new_element->next = NULL;
   new_element->data_id = -1;
   new_element->float_value = (FLT)0.0;
   vec3init(new_element->vector_value);
   colorinit(new_element->color_value);

   /* copy newly allocated element into element*/
   *element = new_element;

   #ifdef __DEBUGDATASTACK__
      num_created++;
   #endif /* __DEBUGDATASTACK__ */

   return TRUE;
}

/* FUNCTION expand_freelist
Parameters:
   void
Return Value:
   Returns TRUE if successful, FALSE otherwise
Local Variables:
   int i - index variable
   PDATA_STACK current - current position within free list
Description:
   This function is called to expand the free list when it runs out
   of elements.
*/
int expand_freelist(void)
{
   int i;
   PDATA_STACK current = NULL;

   /* attempt to create new element at front of list */
   if (!datastack_create_element(&freelist))
      return FALSE;

   /* start at front of list */
   current = freelist;

   for (i=0; i<FREE_INCREMENT-1; i++)
   {
      /* create new element to add to list */
      if (!datastack_create_element(&current->next))
         return FALSE;
      current = current->next;
   }

   #ifdef __DEBUGDATASTACK__
      dmessage("Added %d elements to freelist\n", FREE_INCREMENT);
   #endif /* __DEBUGDATASTACK__ */

   return TRUE;
}

/* FUNCTION datastack_delete
Parameters:
   PPDATA_STACK top - IN/OUT - pointer to the top of stack
Return Value:
   void
Local Variables:
   PDATA_STACK temp - retains the next element while the current
      is freed
   PDATA_STACK current - pointer to the current element of stack
Description:
   This function deletes all leftover elements in the input stack.
*/
void  datastack_delete(PPDATA_STACK top)
{
   PDATA_STACK temp;
   PDATA_STACK current;

   /* assign the top of stack to current */
   current = *top;

   /* loop through each element left in stack */
   while(current != NULL)
   {
      #ifdef __DEBUGDATASTACK__
         num_deleted++;
      #endif /* __DEBUGDATASTACK__ */

      /* free memory allocated for string */
      free(current->string_value);

	   /* retain the next element */
      temp = current->next;

      /* free memory allocated for current element */
      free(current);

	   /* current element becomes next in stack */
      current = temp;
   }

   /* assign top of stack to NULL */
   *top = NULL;

   /* delete free list if last stack */
   num_stacks--;

   if (num_stacks == 0)
   {
      #ifdef __DEBUGDATASTACK__
         dmessage("Deleting free list\n");
      #endif /* __DEBUGDATASTACK__ */

      current = freelist;
      /* loop through each element left in free list */
      while(current != NULL)
      {
         #ifdef __DEBUGDATASTACK__
            num_deleted++;
         #endif /* __DEBUGDATASTACK__ */

         /* free memory allocated for string */
         free(current->string_value);

	      /* retain the next element */
         temp = current->next;

         /* free memory allocated for current element */
         free(current);

	      /* current element becomes next in stack */
         current = temp;
      }
      freelist = NULL;
   }

   #ifdef __DEBUGDATASTACK__
      dmessage("Number elements created: %d\n", num_created);
      dmessage("Number elements deleted: %d\n", num_deleted);
   #endif /* __DEBUGDATASTACK__ */
}
