/***************************
slparser
Copyright (C) 1999  Peter Stuart

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*****************************/


/**********MODULE SLPARSER***************
Filename: slparser.c
Author: ophiuchus42@hotmail.com
Description: This module implements code to create shaders based
   on a token list. It is used to create the data structures for
   shaders as well as execute the shader itself. Currently the
   code is pretty unoptimized. There are many things in here that
   can be (and will be) changed.
Modifications: 
  12/24/98   commented code
*************************/

#include "../include/matrix.h"
#include "../include/math3d.h"
#include "tokens.h"
#include "types.h"
#include "datastack.h"
#include "tokenlist.h"
#include "slparser.h"

/***** STATIC VARIABLES *****
int shader_num - used to assign a specific number to a shader
   (not in use yet)
*/
static int shader_num = 0;


/***** STATIC FUNCTIONS *****/

/* create name information of current token sequence */
static int slparser_name_info(PPTOKEN_LIST current, char** name);

/* create value information for type */
static int slparser_float_info(PPTOKEN_LIST current, PSHADER_DATA data, int flt_num);
static int slparser_color_info(PPTOKEN_LIST current, PSHADER_DATA data, int col_num);
static int slparser_string_info(PPTOKEN_LIST current, PSHADER_DATA data, int str_num);
static int slparser_vector_info(PPTOKEN_LIST current, PSHADER_DATA data, int vec_num);

/* total number of tokens found in sequence: id1 id2 */
static int slparser_num_pairs(PTOKEN_LIST list, TOKEN id1, TOKEN id2);

/* goto label number */
static int slparser_goto(PTOKEN_LIST list, PPTOKEN_LIST current, int number);


/***** "STATIC" MACROS ******/

/* macros to call the slparser_num_pairs function */
/* get number of local variables of specified type in a token list */
#define SLPARSER_NUM_LOCALS(a, b) (slparser_num_pairs((a), LOCAL_TOKEN, (b)))

/* get number of parameters of specified type in a token list */
#define SLPARSER_NUM_PARAMS(a, b) (slparser_num_pairs((a), PARAM_TOKEN, (b))) 

/* get total number of parameters and local variables of specified
   type in token list
*/
#define SLPARSER_NUM_VARS(a, b) (SLPARSER_NUM_LOCALS((a),(b)) \
                               + SLPARSER_NUM_PARAMS((a),(b)))

/* FUNCTION slparser_init
Parameters:
   PTOKEN_LIST list - IN - front of token list
   PSHADER_DATA data - OUT - data structure for shader parameters and
      local variables
Return Value:
   Returns TRUE if successful, FALSE if failed
Local Variables:
   int num - temporary variable to hold the number of parameters
      or local variables of a given type
   TOKEN id - temporary variable to hold the tokens while the token
      list is searched for the parameter and local variable values
   PTOKEN_LIST current - pointer to the current element of the token
      list. Used when searching for the parameter and local variable
      values.
   PTOKEN_INFO info - temporary element of token list used when
      searching for the parameter and local variable values.
   int flt_num - current float parameter or local variable
   int col_num - current color parameter or local variable
   int str_num - current string parameter and local variable
   int vec_num - current vector parameter and local variable
Description:
   This function creates the data structure for the shader parameters
   and local variables. It then fills in the initial values.
*/
int slparser_init(PTOKEN_LIST list, PSHADER_DATA data)
{
   int num = 0;
   int i = 0;
   TOKEN id;
   PTOKEN_LIST current = NULL;
   PTOKEN_INFO info;
   PDATA_STACK stack = NULL;
   int param_num = 0;
   int flt_num, col_num, str_num, vec_num;
   flt_num = col_num = str_num = vec_num = 0;

   /* list must not be empty */
   if (list == NULL)
   {
      error_message("Slparser error: Tokenlist is empty\n");
      return FALSE;
   }

   /* get shader number */
   data->shader_number = shader_num++;

   /* initialize shader_name */
   data->shader_name = NULL;

   /* initialize version info */
   data->generator_name = NULL;
   data->version_name = NULL;

   /* initialize parameter numbers */
   data->float_num = 0;
   data->color_num = 0;
   data->vector_num = 0;
   data->string_num = 0;

   /* get version info */
   current = list;
   if (tokenlist_seek(&current, VERSION_TOKEN))
   {
      if (tokenlist_advance(&current, &info) != ALPHA_TOKEN)
      {
         error_message("Slparser error: Generator not specified\n");
         return FALSE;
      }

      if (!slparser_calloc(strlen(info->alpha_value)+1,
           sizeof(char), (void**)&data->generator_name))
         return FALSE;

      strcpy(data->generator_name, info->alpha_value);

      if (tokenlist_advance(&current, &info) != ALPHA_TOKEN)
      {
         error_message("Slparser error: Version not specified\n");
         error_message("Number: %f\n", info->float_value);
         return FALSE;
      }

      if (!slparser_calloc(strlen(info->alpha_value)+1,
           sizeof(char), (void**)&data->version_name))
         return FALSE;

      strcpy(data->version_name, info->alpha_value);
   }

   /* get name of shader */
   current = list;
   if (!tokenlist_seek(&current, SURFACE_TOKEN))
   {
      error_message("Slparser error: Not a surface shader\n");
      return FALSE;
   }

   if(tokenlist_next(current, &info) != ALPHA_TOKEN)
   {
      error_message("Slparser error: Expecting name of surface\n");
      return FALSE;
   }

   if (!slparser_calloc(strlen(info->alpha_value)+1, 
        sizeof(char), (void**)&data->shader_name))
      return FALSE;

   strcpy(data->shader_name, info->alpha_value);


   /* allocate memory for data and name arrays */

   num = SLPARSER_NUM_VARS(list, FLOAT_TOKEN);
   if(!slparser_calloc(num, sizeof(FLT), (void**)&data->float_data))
      return FALSE;
   if(!slparser_calloc(num, sizeof(char*), (void**)&data->float_name))
      return FALSE;
   data->float_num = num;


   num = SLPARSER_NUM_VARS(list, COLOR_TOKEN);
   if(!slparser_calloc(num, sizeof(COLOR), (void**)&data->color_data))
      return FALSE;
   if(!slparser_calloc(num, sizeof(char*), (void**)&data->color_name))
      return FALSE;
   data->color_num = num;


   num = SLPARSER_NUM_VARS(list, STRING_TOKEN);
   if(!slparser_calloc(num, sizeof(char*), (void**)&data->string_data))
      return FALSE;
   if(!slparser_calloc(num, sizeof(char*), (void**)&data->string_name))
      return FALSE;
   data->string_num = num;


   num = SLPARSER_NUM_VARS(list, VECTOR_TOKEN); 
   if(!slparser_calloc(num, sizeof(vec3), (void**)&data->vector_data))
      return FALSE;
   if(!slparser_calloc(num, sizeof(char*), (void**)&data->vector_name))
      return FALSE;
   data->vector_num = num;

   /* set variable and name values for shader parameters */
   current = list;

   while(tokenlist_seek(&current, PARAM_TOKEN))
   {
      param_num++;
      id = tokenlist_next(current, NULL);
      switch(id)
      {
         case FLOAT_TOKEN:
            if (!slparser_name_info(&current, &data->float_name[flt_num]))
               return FALSE;
            if(!slparser_float_info(&current, data, flt_num))
               return FALSE;
            flt_num++;
            break;

         case COLOR_TOKEN:
            if (!slparser_name_info(&current, &data->color_name[col_num]))
               return FALSE;
            if(!slparser_color_info(&current, data, col_num))
               return FALSE;
            col_num++;
            break;

         case VECTOR_TOKEN:
            if (!slparser_name_info(&current, &data->vector_name[vec_num]))
               return FALSE;
            if(!slparser_vector_info(&current, data, vec_num))
               return FALSE;
            vec_num++;
            break;

         case STRING_TOKEN:
            if (!slparser_name_info(&current, &data->string_name[str_num]))
               return FALSE;
            if(!slparser_string_info(&current, data, str_num))
               return FALSE;
            str_num++;
            break;

         default:
            tokenlist_advance(&current, NULL);
            break;
      }
   }

   /* set parameter numbers */
   data->float_params = flt_num;
   data->color_params = col_num;
   data->vector_params = vec_num;
   data->string_params = str_num;

   /* set variable and name values for shader local variables */

   /* reset current element */
   current = list;

   while(tokenlist_seek(&current, LOCAL_TOKEN))
   {
      id = tokenlist_next(current, NULL);
      switch(id)
      {
         case FLOAT_TOKEN:
            if (!slparser_name_info(&current, &data->float_name[flt_num]))
               return FALSE;
            if (!slparser_float_info(&current, data, flt_num))
               return FALSE;
            flt_num++;
            break;

         case COLOR_TOKEN:
            if (!slparser_name_info(&current, &data->color_name[col_num]))
               return FALSE;
            if(!slparser_color_info(&current, data, col_num))
               return FALSE;
            col_num++;
            break;

         case STRING_TOKEN:
            if (!slparser_name_info(&current, &data->string_name[str_num]))
               return FALSE;
            if (!slparser_string_info(&current, data, str_num))
               return FALSE;

            str_num++;
            break;

         case VECTOR_TOKEN:
            if (!slparser_name_info(&current, &data->vector_name[vec_num]))
               return FALSE;
            if (!slparser_vector_info(&current, data, vec_num))
               return FALSE;
            vec_num++;
            break;

         default:
            tokenlist_advance(&current, NULL);
            break;
      }
   }

   /* initialize parameters */
   datastack_init(&stack);
   #ifdef __DEBUGSLPARSER__
      dmessage("slparser init: param_num = %d\n", param_num);
   #endif

   for (i=0; i<param_num ; i++)
   {
      current = list;
      while(tokenlist_seek(&current, INITCODE_TOKEN))
      {
         id = tokenlist_advance(&current, &info);
         if (id != NUMERIC_TOKEN)
         {
            error_message("Initialization error: initcode missing number\n");
            return FALSE;
         }

         if (i == (int)info->float_value)
         {
            #ifdef __DEBUGSLPARSER__
               dmessage("slparser init: executing initcode %d\n",
                         (int)info->float_value);
            #endif

            if (!slparser_execute(current, &stack, NULL, data))
            {
               error_message("Initialization error: parameter initialization execution failed\n");
               return FALSE;
            }
         }
      }
   }
   datastack_delete(&stack);

   return TRUE;
}

/* FUNCTION slparser_num_pairs
Parameters:
   PTOKEN_LIST list - IN - front of list
   TOKEN id1 - IN - token of first half of pair
   TOKEN id2 - IN - token of second half of pair
Return Value:
   Returns number of pairs found in token list
Local Variables:
   PTOKEN_LIST current - current position within token list
   int num - number of pairs found
Description:
   This function searches the token list for occurances of the given pair of
   tokens together. It sums these and returns the total number.
*/
int slparser_num_pairs(PTOKEN_LIST list, TOKEN id1, TOKEN id2)
{
   PTOKEN_LIST current = NULL;
   int num = 0;

   /* if list does not have any elements there are no pairs */
   if (list == NULL)
      return 0;

   /* start a beginning of list */
   current = list;

   /* loop until there are no more instances of the first token */
   while(tokenlist_seek(&current, id1))
   {
      /* if this is an instance of the pair increment num */
      if (tokenlist_next(current, NULL) == id2)
         num++;
      tokenlist_advance(&current, NULL);
   }
   return num;
}

/* FUNCTION slparser_malloc
Parameters:
   int size - IN - amount of memory to allocate
   void** p - OUT - pointer to allocated memory
Return Value:
   Returns TRUE if memory successfully allocated, FALSE if failed
Local Variables:
   none
Description:
   This function attempts to use malloc to allocate memory. It
   tests for success and returns FALSE if malloc failed. It also
   prints out an error message to stderr if failure occurs.
   (This will be changed to a macro later and put into a more generic
   module containing platform specific code. Instead of fprinting to
   stderr it will use a printerr defined function)
*/
int slparser_malloc(int size, void** p)
{
   /* attempt to allocate memory using malloc */
   if ((*p = malloc(size)) == NULL)
   {
      error_message("Slparser error: Could not allocate memory");
      return FALSE;
   }
   return TRUE;
}

/* FUNCTION slparser_calloc
Parameters:
   int num - IN - number of elements in array
   int size - IN - size of each element in array
   void** p - OUT - pointer to the allocated memory
Return Value:
   Returns true if memory successfully allocated, FALSE if it failed
Local Variables:
   none
Description:
   This function attempts to use calloc to allocate memory for an array. 
   It tests for success and returns FALSE if calloc failed. It also
   prints out an error message to stderr if failure occurs.
   (This will be changed to a macro later and put into a more generic
   module containing platform specific code. Instead of fprinting to
   stderr it will use a printerr defined function)
*/
int slparser_calloc(int num, int size, void** p)
{
   /* attempt to allocate memory */
   if ((*p = calloc(num, size)) == NULL)
   {
      error_message("Slparser error: Could not allocate memory");
      return FALSE;
   }
   return TRUE;
}

/* FUNCTION slparser_delete
Parameters:
   PTOKEN_LIST list - IN - front of token list
   PSHADER_DATA data - IN/OUT - shader information which must be freed
Return Value:
   void
Local Variables:
   int num - temporary variable to hold number of variables or parameters
   (This can be removed when number is added to shader info)
   int i - index variable
Description:
   This function is called to free the memory previously allocated
   to a shader data structure by the slparser_init function.
*/
void slparser_delete(PSHADER_DATA data)
{
   int num = 0;
   int i = 0;

   /* return if the list or the data structure have not been allocated */
   if(data == NULL)
   {
      #ifdef __DEBUGSLPARSER__
         dmessage("Slparser: data NULL\n");
      #endif
      return;
   }

   /* free the shader name string and set to NULL so that structure
      may be used again 
   */
   if (data->shader_name != NULL)
   {
      free(data->shader_name);
      data->shader_name = NULL;
   }

   /* free the version information */
   if (data->generator_name != NULL)
   {
      free(data->generator_name);
      data->generator_name = NULL;
   }

   if (data->version_name != NULL)
   {
      free(data->version_name);
      data->version_name = NULL;
   }

   /* free memory allocated for float variables */
   num = data->float_num;

   if (num > 0)
   {
      /* free each string within the name array */
      for (i=0; i<num; i++)
         free(data->float_name[i]);

      /* free data array */
      free(data->float_data);
      data->float_data = NULL;

      /* free name array */
      free(data->float_name);
      data->float_name = NULL;
   }

   /* free memory allocated for string variables */
   num = data->string_num;

   if (num > 0)
   {
      /* free each string within the name and data array */
      for (i=0; i<num; i++) 
      {
         free(data->string_data[i]);
         free(data->string_name[i]);
      }

      /* free data array */
      free(data->string_data);
      data->string_data = NULL;

      /* free name array */
      free(data->string_name);
      data->string_name = NULL;
   }

   /* free memory allocated for vector variables */
   num = data->vector_num;

   if (num > 0)
   {
      /* free each string within the name array */
      for (i=0; i<num; i++)
         free(data->vector_name[i]);

     /* free data array */
      free(data->vector_data);
      data->vector_data = NULL;

     /* free name array */
      free(data->vector_name);
      data->vector_name = NULL;
   }

   /* free memory allocated for color variables */
   num = data->color_num;

   if (num > 0)
   {
      /* free each string within the name array */
      for (i=0; i<num; i++)
         free(data->color_name[i]);

      /* free data array */
      free(data->color_data);
      data->color_data = NULL;

      /* free name array */
      free(data->color_name);
      data->color_name = NULL;
   }
}

/* FUNCTION slparser_name_info
Parameters:
   PPTOKEN_LIST current - IN/OUT - current position within token list
   char** name - OUT - name of parameter or local variable
Return Value:
   return TRUE if memory allocated, FALSE otherwise
Local Variables:
   PTOKEN_INFO info - token info structure used to hold string
   TOKEN id - temporary variable to hold next token value
Description:
   This function is used to name a shader local variable or parameter.
   It should be called when current is pointing to PARAM_TOKEN or LOCAL_TOKEN.
   It retrieves the string value of the next token, which should be
   ALPHA_TOKEN, and allocated memory for the name.
*/
int slparser_name_info(PPTOKEN_LIST current, char** name)
{
   PTOKEN_INFO info;
   TOKEN id;

   /* advance past the variable type */
   tokenlist_advance(current, NULL);

   /* get string value of name */
   id = tokenlist_advance(current, &info);
   if (id != ALPHA_TOKEN)
   {
      error_message("Slparser error: Need name of variable\n");
      return FALSE;
   }

   /* attempt to allocate memory for string */
   if (!slparser_calloc(strlen(info->alpha_value)+1,
         sizeof(char), (void**)name))
      return FALSE;

   /* copy the string value into the newly allocated array */
   strcpy(*name, info->alpha_value);

   #ifdef __DEBUGSLPARSER__
      dmessage("Allocating new name: ");
      dmessage(*name);
      dmessage("\n");
   #endif

   return TRUE;
}

/* FUNCTION slparser_float_info
Parameters:
   PPTOKEN_LIST current - IN/OUT - current position within the token list
   PSHADER_DATA data - OUT - shader data structure which shall be filled in
   int flt_num - IN - number of current float within array
Return Value:
   Returns TRUE if token list successfully parsed, FALSE otherwise
Local Variables:
   PTOKEN_INFO info - used to temporarily hold the info of a given token
   TOKEN id - temporary variable to hold token id
Description:
   This function fills in the variable and parameter information for
   the shader. The memory should have been previously allocated before
   this function is called. It only does one variable or parameter at a time.
   The position starts on the variable name and is advanced to the next token 
   beyond the initialization tokens. The function fails if a parse error occurs. 
   This is usually caused by tokens which are out of sequence.
*/
int slparser_float_info(PPTOKEN_LIST current, PSHADER_DATA data, int flt_num)
{
   PTOKEN_INFO info;
   TOKEN id;

   /* next token should be variable initialization NUMERIC_TOKEN */ 
   id = tokenlist_advance(current, &info);
   if (id != NUMERIC_TOKEN)
   {
      error_message("Slparser error: variable %s should be initialized\n",
                       data->float_name[flt_num]);
      return FALSE;
   }

   /* copy float value into shader data structure */
   data->float_data[flt_num] = info->float_value;
   return TRUE;
}

/* FUNCTION slparser_color_info
Parameters:
   PPTOKEN_LIST current - IN/OUT - current position within the token list
   PSHADER_DATA data - OUT - shader data structure which shall be filled in
   int col_num - IN - number of current color within array
Return Value:
   Returns TRUE if token list successfully parsed, FALSE otherwise
Local Variables:
   PTOKEN_INFO info - used to temporarily hold the info of a given token
   TOKEN id - temporary variable to hold token id
Description:
   This function fills in the variable and parameter information for
   the shader. The memory should have been previously allocated before
   this function is called. It only does one variable or parameter at a time.
   The position starts on the variable name and is advanced to the next token 
   beyond the initialization tokens. The function fails if a parse error occurs. 
   This is usually caused by tokens which are out of sequence.
*/
int slparser_color_info(PPTOKEN_LIST current, PSHADER_DATA data, int col_num)
{
   PTOKEN_INFO info;
   TOKEN id;

   /* next 3 tokens should be initialization tokens NUMERIC_TOKEN */
   id = tokenlist_advance(current, &info);
   if (id != NUMERIC_TOKEN)
   {
      error_message("Slparser error: variable %s should be initialized\n",
                       data->color_name[col_num]);
      return FALSE;
   }

   /* copy float value into first element of color array */
   data->color_data[col_num][0] = info->float_value;

   id = tokenlist_advance(current, &info);
   if (id != NUMERIC_TOKEN)
   {
      error_message("Slparser error: variable %s should be initialized\n",
                       data->color_name[col_num]);
      return FALSE;
   }

   /* copy float value into second element of color array */
   data->color_data[col_num][1] = info->float_value;

   id = tokenlist_advance(current, &info);
   if (id != NUMERIC_TOKEN)
   {
      error_message("Slparser error: variable %s should be initialized\n",
                       data->color_name[col_num]);
      return FALSE;
   }

   /* copy float value into third element of color array */
   data->color_data[col_num][2] = info->float_value;

   /*data->color_data[col_num][3] = 1.0;*/
   return TRUE;
}

/* FUNCTION slparser_string_info
Parameters:
   PPTOKEN_LIST current - IN/OUT - current position within the token list
   PSHADER_DATA data - OUT - shader data structure which shall be filled in
   int str_num - IN - number of current string within array
Return Value:
   Returns TRUE if token list successfully parsed, FALSE otherwise
Local Variables:
   PTOKEN_INFO info - used to temporarily hold the info of a given token
   TOKEN id - temporary variable to hold token id
Description:
   This function fills in the variable and parameter information for
   the shader. The memory should have been previously allocated before
   this function is called. It only does one variable or parameter at a time.
   The position starts on the variable name and is advanced to the next token 
   beyond the initialization tokens. The function fails if a parse error occurs or
   memory allocation failed. This is usually caused by tokens which are out of sequence.
*/
int slparser_string_info(PPTOKEN_LIST current, PSHADER_DATA data, int str_num)
{
   PTOKEN_INFO info;
   TOKEN id;

   /* next token should be initialization token ALPHA_TOKEN */
   id = tokenlist_advance(current, &info);
   if (id != ALPHA_TOKEN)
   {
      error_message("Slparser error: variable %s should be initialized\n",
                       data->string_name[str_num]);
      return FALSE;
   }

   /* allocate memory for string */
   if (!slparser_calloc(strlen(info->alpha_value)+1,
         sizeof(char), (void**)&data->string_data[str_num]))
      return FALSE;

   /* copy string into shader data structure */
   strcpy(data->string_data[str_num], info->alpha_value);

   return TRUE;
}

/* FUNCTION slparser_vector_info
Parameters:
   PPTOKEN_LIST current - IN/OUT - current position within the token list
   PSHADER_DATA data - OUT - shader data structure which shall be filled in
   int vec_num - IN - number of current vector within array
Return Value:
   Returns TRUE if token list successfully parsed, FALSE otherwise
Local Variables:
   PTOKEN_INFO info - used to temporarily hold the info of a given token
   TOKEN id - temporary variable to hold token id
Description:
   This function fills in the variable and parameter information for
   the shader. The memory should have been previously allocated before
   this function is called. It only does one variable or parameter at a time.
   The position starts on the variable name and is advanced to the next token 
   beyond the initialization tokens. The function fails if a parse error occurs. 
   This is usually caused by tokens which are out of sequence.
*/
int slparser_vector_info(PPTOKEN_LIST current, PSHADER_DATA data, int vec_num)
{
   PTOKEN_INFO info;
   TOKEN id;

   /* next 3 tokens should for initialization NUMERIC_TOKEN */
   id = tokenlist_advance(current, &info);
   if (id != NUMERIC_TOKEN)
   {
      error_message("Slparser error: variable %s should be initialized\n",
                       data->vector_name[vec_num]);
      return FALSE;
   }

   /* copy float value into first element of vector */
   data->vector_data[vec_num][0] = info->float_value;

   id = tokenlist_advance(current, &info);
   if (id != NUMERIC_TOKEN)
   {
      error_message("Slparser error: variable %s should be initialized\n",
                       data->vector_name[vec_num]);
      return FALSE;
   }

   /* copy float value into second element of vector */
   data->vector_data[vec_num][1] = info->float_value;

   id = tokenlist_advance(current, &info);
   if (id != NUMERIC_TOKEN)
   {
      error_message("Slparser error: variable %s should be initialized\n",
                       data->vector_name[vec_num]);
      return FALSE;
   }

   /* copy float value into third element of vector */
   data->vector_data[vec_num][2] = info->float_value;

   return TRUE;
}

/* FUNCTION slparser_get_float
Parameters:
   PSHADER_DATA data - IN - shader data structure containing necessary information
   char* var - IN - name of variable
Return Value:
   Returns a pointer to element within the shader data structure corresponding to
   the given name. It returns NULL if the name is not found. 
Local Variables:
   int i - index variable
Description:
   This function searches the shader data structure for the given 
   variable name. If it is found it returns a pointer to that element
   within the structure (this allows it to be modified) or NULL if
   it is not found.
*/
FLT* slparser_get_float(PSHADER_DATA data, char* var)
{
   int i = 0;

   /* loop through each element of the shader data structure */
   while(i<data->float_num)
   {
      /* if names match, return pointer to element */
      if (streq(var, data->float_name[i]))
         return &data->float_data[i];
      i++;
   }
   return NULL;
}

/* FUNCTION slparser_get_color
Parameters:
   PSHADER_DATA data - IN - shader data structure containing necessary information
   char* var - IN - name of variable
Return Value:
   Returns a pointer to element within the shader data structure corresponding to
   the given name. It returns NULL if the name is not found. 
Local Variables:
   int i - index variable
Description:
   This function searches the shader data structure for the given 
   variable name. If it is found it returns a pointer to that element
   within the structure (this allows it to be modified) or NULL if
   it is not found.
*/
PCOLOR slparser_get_color(PSHADER_DATA data, char* var)
{
   int i = 0;

   /* loop through each element in the shader data structure */
   while(i<data->color_num)
   {
      /* if names match, return pointer to element */
      if (streq(var, data->color_name[i]))
         return data->color_data[i];
      i++;
   }
   return NULL;
}

/* FUNCTION slparser_get_string
Parameters:
   PSHADER_DATA data - IN - shader data structure containing necessary information
   char* var - IN - name of variable
Return Value:
   Returns a pointer to element within the shader data structure corresponding to
   the given name. It returns NULL if the name is not found. 
Local Variables:
   int i - index variable
Description:
   This function searches the shader data structure for the given 
   variable name. If it is found it returns a pointer to that element
   within the structure (this allows it to be modified) or NULL if
   it is not found.
*/
char** slparser_get_string(PSHADER_DATA data, char* var)
{
   int i = 0;

   while(i<data->string_num)
   {
      if (streq(var, data->string_name[i]))
         return &data->string_data[i];
      i++;
   }
   return NULL;
}

/* FUNCTION slparser_get_vector
Parameters:
   PSHADER_DATA data - IN - shader data structure containing necessary information
   char* var - IN - name of variable
Return Value:
   Returns a pointer to element within the shader data structure corresponding to
   the given name. It returns NULL if the name is not found. 
Local Variables:
   int i - index variable
Description:
   This function searches the shader data structure for the given 
   variable name. If it is found it returns a pointer to that element
   within the structure (this allows it to be modified) or NULL if
   it is not found.
*/
pvec3 slparser_get_vector(PSHADER_DATA data, char* var)
{
   int i = 0;

   /* loop through each element in the shader data structure */
   while(i<data->vector_num)
   {
      /* return pointer to element if names match */
      if (streq(var, data->vector_name[i]))
         return data->vector_data[i];
      i++;
   }
   return NULL;
}

/* FUNCTION slparser_get_global
Parameters:
   PGLOBAL_VARIABLES globals - IN - pointer to global variables
   TOKEN t - IN - token representing global variable
Return Value:
   returns a pointer to the global variable within the
   globals data structure or NULL if the token is not found.
Local Variables:
   none
Description:
   This function returns a pointer to the global variable
   of the type given by the input token. The pointer is cast
   to void* in order to allow the function to return any of
   of global variables.
*/
void* slparser_get_global(PGLOBAL_VARIABLES globals, TOKEN t)
{
   switch(t)
   {
      case CS_GLOBAL_TOKEN:
         return (void*)&globals->Cs;
         break;
      case OS_GLOBAL_TOKEN:
         return (void*)&globals->Os;
         break;
      case P_GLOBAL_TOKEN:
         return (void*)&globals->P;
         break;
      case DPDU_GLOBAL_TOKEN:
         return (void*)&globals->dPdu;
         break;
      case DPDV_GLOBAL_TOKEN:
         return (void*)&globals->dPdv;
         break;
      case N_GLOBAL_TOKEN:
         return (void*)&globals->N;
         break;
      case NG_GLOBAL_TOKEN:
         return (void*)&globals->Ng;
         break;
      case U_GLOBAL_TOKEN:
         return (void*)&globals->u;
         break;
      case V_GLOBAL_TOKEN:
         return (void*)&globals->v;
         break;
      case DU_GLOBAL_TOKEN:
         return (void*)&globals->du;
         break;
      case DV_GLOBAL_TOKEN:
         return (void*)&globals->dv;
         break;
      case S_GLOBAL_TOKEN:
         return (void*)&globals->s;
         break;
      case T_GLOBAL_TOKEN:
         return (void*)&globals->t;
         break;
      case DSDU_GLOBAL_TOKEN:
         return (void*)&globals->dsdu;
         break;
      case DSDV_GLOBAL_TOKEN:
         return (void*)&globals->dsdv;
         break;
      case DTDU_GLOBAL_TOKEN:
         return (void*)&globals->dtdu;
         break; 
      case DTDV_GLOBAL_TOKEN:
         return (void*)&globals->dtdv;
         break;
      case L_GLOBAL_TOKEN:
         return (void*)&globals->L;
         break;
      case CL_GLOBAL_TOKEN:
         return (void*)&globals->Cl;
         break;
      case I_GLOBAL_TOKEN:
         return (void*)&globals->I;
         break;
      case CI_GLOBAL_TOKEN:
         return (void*)&globals->Ci;
         break;
      case OI_GLOBAL_TOKEN:
         return (void*)&globals->Oi;
         break;
      case E_GLOBAL_TOKEN:
         return (void*)&globals->E;
         break;
   }
   return NULL;
}

/* FUNCTION slparser_start
*/
void slparser_start(PPTOKEN_LIST list)
{
   PTOKEN_LIST current;
 
   current = *list;

   /* find the CODE_TOKEN and begin execution */
   while(tokenlist_seek(&current, CODE_TOKEN))
   {
      *list = current;
      tokenlist_advance(&current, NULL); 
   }
}


/* FUNCTION slparser_test
Parameters:
   PTOKEN_LIST list - front of list
Return Value:
   Returns TRUE if list parsed correctly, FALSE if error
Local Variables:
   PTOKEN_LIST current - current position within token list
   PTOKEN_INFO info - token data structure of current token
   TOKEN id - current token
Description:
   This function parses the token list and determines if
   there are any errors.
*/
int slparser_test(PTOKEN_LIST list)
{
   PTOKEN_LIST current = NULL;
   PTOKEN_INFO info = NULL;
   TOKEN id;


   #ifdef __DEBUGSLPARSER__
      dmessage("Entering slparser_test\n");
   #endif /* __DEBUGSLPARSER__ */

   /* fail if list is empty */
   if (list == NULL)
      return FALSE;

   /* start a beginning of list */
   current = list;

   id = tokenlist_advance(&current, NULL);

   /* find the CODE_TOKEN and begin execution */
/*
   while(tokenlist_seek(&current, CODE_TOKEN))
   {
      id = tokenlist_advance(&current, NULL);  
   }
*/
   /* loop until RETURN_TOKEN or end of file */
   while (id != EOF_TOKEN)
   {
      switch(id)
      {
         case CODE_TOKEN:
            error_message("Parse Error: Code token found\n");
            return FALSE;

         case INITCODE_TOKEN:
            error_message("Parse Error: Initcode\n");
            return FALSE;

         case JUMP_TOKEN:
            id = tokenlist_advance(&current, &info);
            if (id != NUMERIC_TOKEN)
            {
               error_message("Parse Error: Jump token needs label number\n");
               return FALSE;
            }
            break;

         case LOCAL_TOKEN:
            break;

         case PARAM_TOKEN:
            error_message("Parse Error: Param token\n");
            return FALSE;

         case RETURN_AREA_TOKEN:
            /* token has unknown use to me */
            break;

         case RETURN_TOKEN:
            return TRUE;
            break;

         case SURFACE_TOKEN:
            error_message("Parse Error: Surface token\n");
            return FALSE;

         case LABEL_TOKEN:
            id = tokenlist_advance(&current, NULL);
            if (id != NUMERIC_TOKEN)
            {
               error_message("Parse Error: Label without number\n");
               return FALSE;
            }
            break;

         case POPC_TOKEN:
            id = tokenlist_advance(&current, NULL);
            if ((id != ALPHA_TOKEN) && !IS_GLOBAL_COLOR(id))
            {
               error_message("Parse Error: No variable to pop color into\n");
               return FALSE;
            }
            break;

         case PUSHC_TOKEN:
            id = tokenlist_advance(&current, NULL);
            if ((id != ALPHA_TOKEN) && !IS_GLOBAL_COLOR(id))
            {
               error_message("Parse Error: No color variable to push\n");
               return FALSE;
            }
            break;

         case POPF_TOKEN:
            id = tokenlist_advance(&current, NULL);
            if ((id != ALPHA_TOKEN) && !IS_GLOBAL_FLOAT(id))
            {
               error_message("Parse Error: No variable to pop float into\n");
               return FALSE;
            }
            break;

         case PUSHF_TOKEN:
            id = tokenlist_advance(&current, NULL);
            if ((id != ALPHA_TOKEN) && !IS_GLOBAL_FLOAT(id))
            {
               error_message("Parse Error: No float variable to push\n");
               return FALSE;
            }
            break;

         case PUSHIF_TOKEN:
            id = tokenlist_advance(&current, NULL);
            if (id != NUMERIC_TOKEN)
            {
               error_message("Parse Error: No float value to push\n");
               return FALSE;
            }
            break;

         case PUSHIS_TOKEN:
            id = tokenlist_advance(&current, NULL);
            if (id != ALPHA_TOKEN)
            {
               error_message("Parse Error: No string value to push\n");
               return FALSE;
            }
            break;

         case PUSHIV_TOKEN:
            id = tokenlist_advance(&current, NULL);
            if (id != NUMERIC_TOKEN)
            {
               error_message("Parse Error: Missing x element of vector to push\n");
               return FALSE;
            }
            id = tokenlist_advance(&current, NULL);
            if (id != NUMERIC_TOKEN)
            {
               error_message("Parse Error: Missing y element of vector to push\n");
               return FALSE;
            }
            id = tokenlist_advance(&current, NULL);
            if (id != NUMERIC_TOKEN)
            {
               error_message("Parse Error: Missing z element of vector to push\n");
               return FALSE;
            }
            break;

         case PUSHS_TOKEN:
            id = tokenlist_advance(&current, NULL);
            if (id != ALPHA_TOKEN)
            {
               error_message("Parse Error: No string variable to push\n");
               return FALSE;
            }
            break;

         case POPV_TOKEN:
            id = tokenlist_advance(&current, NULL);
            if ((id != ALPHA_TOKEN) && !IS_GLOBAL_VECTOR(id))
            {
               error_message("Parse Error: No variable to pop vector into\n");
               return FALSE;
            }
            break;

         case PUSHV_TOKEN:
            id = tokenlist_advance(&current, NULL);
            if ((id != ALPHA_TOKEN) && !IS_GLOBAL_VECTOR(id))
            {
               error_message("Parse Error: No vector variable to push\n");
               return FALSE;
            }
            break;

         case ADDCC_TOKEN:
            break;

         case ADDCF_TOKEN:
            break;

         case ADDFF_TOKEN:
            break;

         case ADDFF_ININ_TOKEN:
            id = tokenlist_advance(&current, NULL);
            if ((id != ALPHA_TOKEN) && !IS_GLOBAL_FLOAT(id))
            {
               error_message("Parse Error: missing float \
                              variable to multiply\n");
               return FALSE;
            }
            id = tokenlist_advance(&current, NULL);
            if ((id != ALPHA_TOKEN) && !IS_GLOBAL_FLOAT(id))
            {
               error_message("Parse Error: missing second float \
                              variable to multiply\n");
               return FALSE;
            }
            break;

         case ADDVV_TOKEN:
            break;

         case ADDC_IN_TOKEN:
            id = tokenlist_advance(&current, NULL);
            if ((id != ALPHA_TOKEN) && !IS_GLOBAL_COLOR(id))
            {
               error_message("Parse Error: No color variable to add\n");
               return FALSE;
            }
            break;

         case ADDF_IN_TOKEN:
            id = tokenlist_advance(&current, NULL);
            if ((id != ALPHA_TOKEN) && !IS_GLOBAL_FLOAT(id))
            {
               error_message("Parse Error: No float variable to add\n");
               return FALSE;
            }
            break;

         case ADDV_IN_TOKEN:
            id = tokenlist_advance(&current, NULL);
            if ((id != ALPHA_TOKEN) && !IS_GLOBAL_VECTOR(id))
            {
               error_message("Parse Error: No vector variable to add\n");
               return FALSE;
            }
            break;

         case CPCC_IN_TOKEN:
            id = tokenlist_advance(&current, NULL);
            if ((id != ALPHA_TOKEN) && !IS_GLOBAL_COLOR(id))
            {
               error_message("Parse Error: No color variable to copy\n");
               return FALSE;
            }
            id = tokenlist_advance(&current, NULL);
            if ((id != ALPHA_TOKEN) && !IS_GLOBAL_COLOR(id))
            {
               error_message("Parse Error: No color variable to copy into\n");
               return FALSE;
            }
            break;

         case CPFF_IN_TOKEN:
            id = tokenlist_advance(&current, NULL);
            if ((id != ALPHA_TOKEN) && !IS_GLOBAL_FLOAT(id))
            {
               error_message("Parse Error: No float variable to copy\n");
               return FALSE;
            }
            id = tokenlist_advance(&current, NULL);
            if ((id != ALPHA_TOKEN) && !IS_GLOBAL_FLOAT(id))
            {
               error_message("Parse Error: No float variable to copy into\n");
               return FALSE;
            }
            break;

         case CROSS_TOKEN:
            break;

         case DIVFF_TOKEN:
            break;

         case DIVFF_ININ_TOKEN:
            id = tokenlist_advance(&current, NULL);
            if ((id != ALPHA_TOKEN) && !IS_GLOBAL_FLOAT(id))
            {
               error_message("Parse Error: missing float variable to divide\n");
               return FALSE;
            }
            id = tokenlist_advance(&current, NULL);
            if ((id != ALPHA_TOKEN) && !IS_GLOBAL_FLOAT(id))
            {
               error_message("Parse Error: missing float variable to divide\n");
               return FALSE;
            }
            break;

         case DIVVF_TOKEN:
            break;

         case DIVF_IN_TOKEN:
            id = tokenlist_advance(&current, NULL);
            if ((id != ALPHA_TOKEN) && !IS_GLOBAL_FLOAT(id))
            {
               error_message("Parse Error: No float variable to divide\n");
               return FALSE;
            }
            break;

         case FTOC_TOKEN:
            break;

         case FEQF_TOKEN:
            break;

         case FNEF_TOKEN:
            break;

         case GE_TOKEN:
            break;

         case GT_TOKEN:
            break;

         case IFZ_TOKEN:
            id = tokenlist_advance(&current, &info);
            if (id != NUMERIC_TOKEN)
            {
               error_message("Parse Error: No label given for conditional\n");
               return FALSE;
            }
            break;

         case LT_TOKEN:
            break;

         case MULC_IN_TOKEN:
            id = tokenlist_advance(&current, NULL);
            if ((id != ALPHA_TOKEN) && !IS_GLOBAL_COLOR(id))
            {
               error_message("Parse Error: No color variable to multiply\n");
               return FALSE;
            }
            break;

         case MULF_IN_TOKEN:
            id = tokenlist_advance(&current, NULL);
            if ((id != ALPHA_TOKEN) && (id != NUMERIC_TOKEN) 
               && !IS_GLOBAL_FLOAT(id))
            {
               error_message("Parse Error: No float variable or value to multiply\n");
               return FALSE;
            }
            break;

         case MULCC_TOKEN:
            break;

         case MULCC_ININ_TOKEN:
            id = tokenlist_advance(&current, NULL);
            if ((id != ALPHA_TOKEN) && !IS_GLOBAL_COLOR(id))
            {
               error_message("Parse Error: No color variables to multiply\n");
               return FALSE;
            }
            id = tokenlist_advance(&current, NULL);
            if ((id != ALPHA_TOKEN) && !IS_GLOBAL_COLOR(id))
            {
               error_message("Parse Error: No color variables to multiply\n");
               return FALSE;
            }
            break;

         case MULCF_TOKEN:
            break;

         case MULCF_ININ_TOKEN:
            id = tokenlist_advance(&current, NULL);
            if ((id != ALPHA_TOKEN) && !IS_GLOBAL_COLOR(id))
            {
               error_message("Parse Error: missing color \
                              variable to multiply\n");
               return FALSE;
            }
            id = tokenlist_advance(&current, NULL);
            if ((id != ALPHA_TOKEN) && !IS_GLOBAL_FLOAT(id))
            {
               error_message("Parse Error: missing float \
                              variable to multiply\n");
               return FALSE;
            }
            break;
 
         case MULFC_TOKEN:
            break;

         case MULFC_IN_TOKEN:
         {                                                                                  id = tokenlist_advance(&current, NULL);
            if ((id != ALPHA_TOKEN) && !IS_GLOBAL_FLOAT(id))
            {
               error_message("Parse Error: No float variable to multiply\n");
               return FALSE;
            }
            break;                                                                       }

         case MULFF_TOKEN:
            break;

         case MULFF_ININ_TOKEN:
            id = tokenlist_advance(&current, NULL);
            if ((id != ALPHA_TOKEN) && !IS_GLOBAL_FLOAT(id))
            {
               error_message("Parse Error: missing float \
                              variable to multiply\n");
               return FALSE;
            }
            id = tokenlist_advance(&current, NULL);
            if ((id != ALPHA_TOKEN) && !IS_GLOBAL_FLOAT(id))
            {
               error_message("Parse Error: missing float \
                              variable to multiply\n");
               return FALSE;
            }
            break;

         case MULFV_TOKEN:
            break;

         case MULVF_TOKEN:
            break;

         case MULFV_IN_TOKEN:
         {
            id = tokenlist_advance(&current, NULL);
            if ((id != ALPHA_TOKEN) && !IS_GLOBAL_FLOAT(id))
            {
               error_message("Parse Error: No float variable to multiply\n");
               return FALSE;
            }
            break;
         }
         case NEGF_TOKEN:
            break;

         case NEGV_TOKEN:
            break;

         case AND_TOKEN:
            break;

         case OR_TOKEN:
            break;

         case SNE_TOKEN:
            break;

         case SUBCC_TOKEN:
            break;

         case SUBCF_TOKEN:
            break;

         case SUBFF_TOKEN:
            break;

         case SUBFF_ININ_TOKEN:
            id = tokenlist_advance(&current, NULL);
            if ((id != ALPHA_TOKEN) && !IS_GLOBAL_FLOAT(id))
            {
               error_message("Parse Error: missing float \
                              variable to subtract\n");
               return FALSE;
            }
            id = tokenlist_advance(&current, NULL);
            if ((id != ALPHA_TOKEN) && !IS_GLOBAL_FLOAT(id))
            {
               error_message("Parse Error: missing float \
                              variable to subtract\n");
               return FALSE;
            }
            break;

         case SUBVV_TOKEN:
            break;

         case SUBF_IN_TOKEN:
            id = tokenlist_advance(&current, NULL);
            if ((id != ALPHA_TOKEN) && !IS_GLOBAL_FLOAT(id))
            {
               error_message("Parse Error: No float variable to subtract\n");
               return FALSE;
            }
            break;

         case VDOT_TOKEN:
            break;

         case AREA_TOKEN:
            break;

         case CMIX_TOKEN:
            break;

         case CELLNOISEF_TOKEN:
            break;

         case CELLNOISEFF_TOKEN:
            break;

         case COLORNOISEV_TOKEN:
            break;

         case COMP_TOKEN:
            break;

         case CSET_TOKEN:
            id = tokenlist_advance(&current, NULL);
            if ((id != ALPHA_TOKEN) && !IS_GLOBAL_COLOR(id))
            {
               error_message("Parse Error: No color variable to set\n");
               return FALSE;
            }
            break;

         case CALCULATENORMAL_TOKEN:
            break;

         case CTEXTURE_TOKEN:
            break;

         case FTEXTURE_TOKEN:
            break;

         case FACEFORWARD_TOKEN:
            break;

         case LENGTH_TOKEN:
            break;

         case NORMALIZE_TOKEN:
            break;

         case NTRANSFORMS_TOKEN:
            break;

         case VTRANSFORMS_TOKEN:
            break;

         case TRANSFORMS_TOKEN:
            break;

         case XCOMP_TOKEN:
            break;

         case YCOMP_TOKEN:
            break;

         case ZCOMP_TOKEN:
            break;

         case ABS_TOKEN:
            break;

         case DERIVV_TOKEN:
            break;

         case DVF_TOKEN:
            break;

         case DUF_TOKEN:
            break;

         case EXP_TOKEN:
            break;

         case FCLAMP_TOKEN:
            break;

         case FLOOR_TOKEN:
            break;

         case FMAX_TOKEN:
            break;

         case FMIN_TOKEN:
            break;

         case FRESNEL4_TOKEN:
            break;

         case MOD_TOKEN:
            break;

         case NOISEF_TOKEN:
            break;

         case NOISEFF_TOKEN:
            break;

         case NOISEV_TOKEN:
            break;

         case POINTNOISEFF_TOKEN:
            break;

         case POINTNOISEV_TOKEN:
            break;

         case POW_TOKEN:
            break;

         case RANDOM_TOKEN:
            break;

         case SMOOTHSTEP_TOKEN:
            break;

         case SQRT_TOKEN:
            break;

         case TRACE_TOKEN:
            break;

         case AMBIENT_TOKEN:
            break;

         case CENVIRONMENT_TOKEN:
            break;

         case DIFFUSE_TOKEN:
            break;

         case ILLUMINANCE3_TOKEN:
            id = tokenlist_advance(&current, NULL);
            if (id != NUMERIC_TOKEN)
            {
               error_message("Parse Error: Illuminance not given value\n");
               return FALSE;
            }
            break;

         case ILLUMINANCE_START_TOKEN:
            break;

         case LIGHTSOURCE_TOKEN:
            break;

         case REFLECT_TOKEN:
            break;

         case SPECULAR_TOKEN:
            break;

         case CS_GLOBAL_TOKEN:
         case OS_GLOBAL_TOKEN:
         case P_GLOBAL_TOKEN:
         case DPDU_GLOBAL_TOKEN:
         case DPDV_GLOBAL_TOKEN:
         case N_GLOBAL_TOKEN:
         case NG_GLOBAL_TOKEN:
         case U_GLOBAL_TOKEN:
         case V_GLOBAL_TOKEN:
         case DU_GLOBAL_TOKEN:
         case DV_GLOBAL_TOKEN:
         case S_GLOBAL_TOKEN:
         case T_GLOBAL_TOKEN:
         case DSDU_GLOBAL_TOKEN:
         case DSDV_GLOBAL_TOKEN:
         case DTDU_GLOBAL_TOKEN:
         case DTDV_GLOBAL_TOKEN:
         case L_GLOBAL_TOKEN:
         case CL_GLOBAL_TOKEN:
         case I_GLOBAL_TOKEN:
         case CI_GLOBAL_TOKEN:
         case OI_GLOBAL_TOKEN:
         case E_GLOBAL_TOKEN:
            error_message("Parse Error: Global Variable found\n");
            return FALSE;

         case ALPHA_TOKEN:
            error_message("Parse Error: Unexpected alpha token %s\n", 
               info->alpha_value);
            return FALSE;

         case NUMERIC_TOKEN:
            error_message("Parse Error: Unexpected numeric token %f\n",
               info->float_value);
            return FALSE;

         case EOF_TOKEN:
            error_message("End of file\n");
            return FALSE;

         default:
            error_message("Unknown token %d\n", id);
            return FALSE;
      } 
      id = tokenlist_advance(&current, &info);
   } 
   return TRUE;
}

/* FUNCTION slparser_execute
Parameters:
   PTOKEN_LIST list - IN - front of token list
   PPDATA_STACK stack - IN/OUT - top of data stack
   PGLOBAL_VARIABLES globals - IN/OUT - global variables structure
   PSHADER_DATA data - IN/OUT - shader data structure
Return Value:
   Returns TRUE if toke list successfully parsed and executed. FALSE
   otherwise.
Local Variables:
   PTOKEN_LIST current - current position within token list
   PTOKEN_INFO info - token data structure of current token
   TOKEN id - current token
Description:
   This function executes the input token list. The function begins at
   CODE_TOKEN and ends at RETURN_TOKEN. The input data stack is used to
   modify the data given in globals and data. Unfamiliar tokens or incorrectly
   placed tokens cause the function to fail. The stack will remain as it was
   if the function fails, so that it may be read to debug the failure. This is
   the backbone of the slparser code, and as such is the most important. Currently
   it only supports the tokens found in the checker.sl shader. More will be added.
*/
int slparser_execute(PTOKEN_LIST list, PPDATA_STACK stack, 
                     PGLOBAL_VARIABLES globals, PSHADER_DATA data)
{
   PTOKEN_LIST current = NULL;
   PTOKEN_INFO info;
   TOKEN id;

   /* fail if list is empty */
   if (list == NULL)
      return FALSE;

   /* start a beginning of list */
   current = list;
  
   id = tokenlist_advance(&current, &info);

   /* find the CODE_TOKEN and begin execution */
/*
   while(tokenlist_seek(&current, CODE_TOKEN))
   {
      id = tokenlist_advance(&current, &info);
   }
*/
   /* loop until RETURN_TOKEN or end of file */
   while ((id != RETURN_TOKEN) && (id != EOF_TOKEN))
   {
      switch(id)
      {
         case INITCODE_TOKEN:
         {
            /* token to initialize parameters */
            id = tokenlist_advance(&current, &info);
            if (id != NUMERIC_TOKEN)
            {
               error_message("Execution error: initcode missing number\n");
               return FALSE;
            } 
            #ifdef __DEBUGSLPARSER__
               dmessage("initialize parameter %d\n", (int)info->float_value);
            #endif
            break;
         }
         case JUMP_TOKEN:
         {
            /* token to jump to a specific label within the token list */
            id = tokenlist_advance(&current, &info);
            #ifdef __DEBUGSLPARSER__
               dmessage("Jump to Label %d\n", (int)info->float_value);
            #endif

            slparser_goto(list, &current, (int)info->float_value);

            break;
         }
         case RETURN_AREA_TOKEN:
            /* token has unknown use to me */
            #ifdef __DEBUGSLPARSER__
               dmessage("Return area\n");
            #endif
            break;

         case RETURN_TOKEN:
            #ifdef __DEBUGSLPARSER__
               dmessage("Return\n");
            #endif;
            return TRUE;

         case LABEL_TOKEN:
            /* token designates section within list */
            id = tokenlist_advance(&current, &info);

            #ifdef __DEBUGSLPARSER__
               dmessage("Label %d reached\n", (int)info->float_value);
            #endif /* __DEBUGSLPARSER__ */
            break;

         case POPC_TOKEN:
            /* pop color off of data stack and put into variable 
               vector variables can be popped off as well to hold
               compliance with BMRT 2.3.6 */

            id = tokenlist_advance(&current, &info);
            if (id == ALPHA_TOKEN)
            {
               #ifdef __DEBUGSLPARSER__
                  dmessage("Pop color into %s\n", info->alpha_value);
               #endif
               if (!datastack_popc(stack,
                       slparser_get_color(data, info->alpha_value)))
               {
                  if (!datastack_popv(stack,
                       slparser_get_color(data, info->alpha_value)))
                     return FALSE;
               }
            }
            else if (IS_GLOBAL_COLOR(id))
            {
               #ifdef __DEBUGSLPARSER__
                  dmessage("Pop color into global %d\n", id);
               #endif
               if (!datastack_popc(stack,
                       (PCOLOR)slparser_get_global(globals, id)))
               {
                  if (!datastack_popv(stack,
                       (PCOLOR)slparser_get_global(globals, id)))
                     return FALSE;
               }
            }
            break;

         case PUSHC_TOKEN:
            /* push color */
            id = tokenlist_advance(&current, &info);
            if (id == ALPHA_TOKEN)
            {
               #ifdef __DEBUGSLPARSER__
                  dmessage("Push color %s\n", info->alpha_value);
               #endif
               if (!datastack_pushc(stack,
                       slparser_get_color(data, info->alpha_value)))
                  return FALSE;
            }
            else if (IS_GLOBAL_COLOR(id))
            {
               #ifdef __DEBUGSLPARSER__
                 dmessage("Push global color %d\n", id);
               #endif
               if (!datastack_pushc(stack,
                       (PCOLOR)slparser_get_global(globals, id)))
                  return FALSE;
            }
            break;

         case POPF_TOKEN:
            /* pop float */
            id = tokenlist_advance(&current, &info);
            if (id == ALPHA_TOKEN)
            {
               #ifdef __DEBUGSLPARSER__
                  dmessage("Pop float into variable %s\n", info->alpha_value);
               #endif
               if (!datastack_popf(stack, 
                       slparser_get_float(data, info->alpha_value)))
                  return FALSE;
            }
            else if (IS_GLOBAL_FLOAT(id))
            {
               #ifdef __DEBUGSLPARSER__
                  dmessage ("Pop float into global %d\n", id);
               #endif
               if (!datastack_popf(stack,
                       (FLT*)slparser_get_global(globals, id)))
                  return FALSE;
            }
            break;

         case PUSHF_TOKEN:
            /* push float */
            id = tokenlist_advance(&current, &info);
            if (id == ALPHA_TOKEN)
            {
               #ifdef __DEBUGSLPARSER__
                  dmessage("Push float variable %s\n", info->alpha_value);
               #endif
               if (!datastack_pushf(stack,
                       slparser_get_float(data, info->alpha_value)))
                  return FALSE;
            }
            else if (IS_GLOBAL_FLOAT(id))
            {
               #ifdef __DEBUGSLPARSER__
                  dmessage("Push float variable %d\n", id);
               #endif
               if (!datastack_pushf(stack,
                       (FLT*)slparser_get_global(globals, id)))
                  return FALSE;
            }
            break;

         case PUSHIF_TOKEN:
            /* push float */
            id = tokenlist_advance(&current, &info);

            #ifdef __DEBUGSLPARSER__
               dmessage("Push float %f\n", info->float_value);
            #endif
            if (!datastack_pushf(stack, &info->float_value))
               return FALSE;
            break;

         case PUSHIS_TOKEN:
            /* push string */
            error_message("Execution error: PUSHIS not implemented\n");
            return FALSE;
            
         case PUSHIV_TOKEN:
         {
            /* push input vector */
            vec3 temp;

            #ifdef __DEBUGSLPARSER__
               dmessage("Push input vector\n");
            #endif

            id = tokenlist_advance(&current, &info);
            if (id != NUMERIC_TOKEN)
            {
               error_message("Execution error: improper input for PUSHIV\n");
               return FALSE;
            }
            temp[0] = info->float_value;

            id = tokenlist_advance(&current, &info);
            if (id != NUMERIC_TOKEN)
            {
               error_message("Execution error: improper input for PUSHIV\n");
               return FALSE;
            }
            temp[1] = info->float_value;

            id = tokenlist_advance(&current, &info);
            if (id != NUMERIC_TOKEN)
            {
               error_message("Execution error: improper input for PUSHIV\n");
               return FALSE;
            }
            temp[2] = info->float_value;

            if (!datastack_pushv(stack, temp))
               return FALSE;
 
            break;
         }
         case POPV_TOKEN:
         {
            /* pop vector */
            pvec3 temp;
            FLT val;

            id = tokenlist_advance(&current, &info);
            if (id == ALPHA_TOKEN)
            {
               #ifdef __DEBUGSLPARSER__
                  dmessage("Pop vector into %s\n", info->alpha_value);
               #endif
               temp = slparser_get_vector(data, info->alpha_value);
            }
            else if (IS_GLOBAL_VECTOR(id))
            {
               #ifdef __DEBUGSLPARSER__
                  dmessage("Pop vector into global %d\n", id);
               #endif
               temp = (pvec3)slparser_get_global(globals, id);
            }

            if (!datastack_popv(stack, temp))
            {
               if (!datastack_popf(stack, &val))
                  return FALSE;
               temp[0] = val;

               if (!datastack_popf(stack, &val))
                  return FALSE;
               temp[1] = val;

               if (!datastack_popf(stack, &val))
                  return FALSE;
               temp[2] = val;
            }
            break; 
         }
         case PUSHS_TOKEN:
            /* push string */

            id = tokenlist_advance(&current, &info);

            if (!datastack_pushs(stack, 
                     slparser_get_string(data, info->alpha_value)))
               return FALSE;

            #ifdef __DEBUGSLPARSER__
               dmessage("Push string variable %s\n", info->alpha_value);
            #endif /* __DEBUGSLPARSER__ */

            break;

         case PUSHV_TOKEN:
         {
            /* push vector */
            id = tokenlist_advance(&current, &info);
            if (id == ALPHA_TOKEN)
            {
               #ifdef __DEBUGSLPARSER__
                  dmessage("Push vector %s\n", info->alpha_value);
               #endif
               if (!datastack_pushv(stack,
                       slparser_get_vector(data, info->alpha_value)))
                  return FALSE;
            }
            else if (IS_GLOBAL_VECTOR(id))
            {
               #ifdef __DEBUGSLPARSER__
                  dmessage("Push vector global %d\n", id);
               #endif
               if (!datastack_pushv(stack,
                       (pvec3)slparser_get_global(globals, id)))
                  return FALSE;
            }
            break; 
         }
         case ADDC_IN_TOKEN:
         {
            COLOR temp1;
            PCOLOR val;

            id = tokenlist_advance(&current, &info);
            if (id == ALPHA_TOKEN)
            {
               #ifdef __DEBUGSLPARSER__
                  dmessage("add variable %s to color", info->alpha_value);
               #endif
               val = slparser_get_color(data, info->alpha_value);
               if (val == NULL)
                  return FALSE;
            }
            else if (IS_GLOBAL_COLOR(id))
            {
               #ifdef __DEBUGSLPARSER__
                  dmessage("add global %d to color", id); 
               #endif
               val = (PCOLOR)slparser_get_global(globals, id);
               if (val == NULL)
                  return FALSE;
            }

            coloradd(val, temp1, temp1);

            if (!datastack_pushc(stack, temp1))
               return FALSE;

            break;
         }
         case ADDCC_TOKEN:
         {
            /* add top two colors of stack */
            COLOR temp1;
            COLOR temp2;
            #ifdef __DEBUGSLPARSER__
               dmessage("Add two colors\n");
            #endif
            if (!datastack_popc(stack, temp1))
                  return FALSE;
            if (!datastack_popc(stack, temp2))
                  return FALSE;
            coloradd(temp1, temp2, temp1);
            if (!datastack_pushc(stack, temp1))
               return FALSE;
            break;
         }
         case ADDCF_TOKEN:
         {
            /* add float to color */
            COLOR temp1;
            FLT val;

            #ifdef __DEBUGSLPARSER__
               dmessage("Add float to each component of color\n");
            #endif /* __DEBUGSLPARSER__ */

            if (!datastack_popc(stack, temp1))
               return FALSE;
            if (!datastack_popf(stack, &val))
               return FALSE;

            /* add float to each component of color */
            temp1[0] += val;
            temp1[1] += val;
            temp1[2] += val;

            if (!datastack_pushc(stack, temp1))
               return FALSE;

            break;
         }
         case ADDF_IN_TOKEN:
         {
            /* add input float to float */
            FLT *temp1;
            FLT temp2, temp3;

            #ifdef __DEBUGSLPARSER__
               dmessage("Add input float to float\n");
            #endif

            id = tokenlist_advance(&current, &info);
            if (id == ALPHA_TOKEN)
            {
               temp1 = slparser_get_float(data, info->alpha_value);
               if (temp1 == NULL)
                  return FALSE;
            }
            else if (IS_GLOBAL_FLOAT(id))
            {
               temp1 = (FLT*)slparser_get_global(globals, id);
               if (temp1 == NULL)
                  return FALSE;
            }

            if (!datastack_popf(stack, &temp2))
               return FALSE;

            temp3 = *temp1 + temp2;

            if (!datastack_pushf(stack, &temp3))
               return FALSE;
            break;
         }
         case ADDFF_TOKEN:
         {
            /* add two floats */
            FLT val1, val2;

            #ifdef __DEBUGSLPARSER__
               dmessage("Add two floats\n");
            #endif /* __DEBUGSLPARSER__ */

            if (!datastack_popf(stack, &val1))
               return FALSE;
            if (!datastack_popf(stack, &val2))
               return FALSE;

            val1 += val2;

            if (!datastack_pushf(stack, &val1))
               return FALSE;

            break;
         }
         case ADDFF_ININ_TOKEN:
         {
            /* add two floats */
            FLT *val1, *val2;
            FLT val3;

            #ifdef __DEBUGSLPARSER__
               dmessage("Add input two floats\n");
            #endif /* __DEBUGSLPARSER__ */

            id = tokenlist_advance(&current, &info);
            if (id == ALPHA_TOKEN)
            {
               val1 = slparser_get_float(data, info->alpha_value);
               if (val1 == NULL)
                  return FALSE;
            }
            else if (IS_GLOBAL_FLOAT(id))
            {
               val1 = (FLT*)slparser_get_global(globals, id);
               if (val1 == NULL)
                  return FALSE;
            }

            id = tokenlist_advance(&current, &info);
            if (id == ALPHA_TOKEN)
            {
               val2 = slparser_get_float(data, info->alpha_value);
               if (val1 == NULL)
                  return FALSE;
            }
            else if (IS_GLOBAL_FLOAT(id))
            {
               val2 = (FLT*)slparser_get_global(globals, id);
               if (val1 == NULL)
                  return FALSE;
            }

            val3 = *val1 + *val2;

            if (!datastack_pushf(stack, &val3))
               return FALSE;

            break;
         }
         case ADDV_IN_TOKEN:
         {
            pvec3 temp1;
            vec3 temp2;

            #ifdef __DEBUGSLPARSER__
               dmessage("Add input vector to vector\n");
            #endif

            id = tokenlist_advance(&current, &info);
            if (id == ALPHA_TOKEN)
            {
               temp1 = slparser_get_vector(data, info->alpha_value);
               if (temp1 == NULL)
                  return FALSE;
            }
            else if (IS_GLOBAL_FLOAT(id))
            {
               temp1 = (pvec3)slparser_get_global(globals, id);
               if (temp1 == NULL)
                  return FALSE;
            }

            if (!datastack_popv(stack, temp2))
               return FALSE;

            #ifdef __DEBUGSLPARSER__
               dmessage("Adding ");
               vec3print(temp1);
               dmessage(" and ");
               vec3print(temp2);
               dmessage("\n");
            #endif 

            vec3add(temp1, temp2, temp2);

            if (!datastack_pushv(stack, temp2))
               return FALSE;
            break;
         }
         case CPCC_IN_TOKEN:
         {
            /* copy one color into another */
            PCOLOR c1;
            PCOLOR c2;

            id = tokenlist_advance(&current, &info);
            if (id == ALPHA_TOKEN)
            {
               #ifdef __DEBUGSLPARSER__
                  dmessage("Copy variable %s ", info->alpha_value);
               #endif
               c1 = slparser_get_color(data, info->alpha_value);
               if (c1 == NULL)
                  return FALSE;
            }
            else if (IS_GLOBAL_COLOR(id))
            {
               #ifdef __DEBUGSLPARSER__
                  dmessage("Copy global %d ", id); 
               #endif
               c1 = (PCOLOR)slparser_get_global(globals, id);
               if (c1 == NULL)
                  return FALSE;
               #ifdef __DEBUGSLPARSER__
                  colorprint(c1);
               #endif
            }

            id = tokenlist_advance(&current, &info);
            if (id == ALPHA_TOKEN)
            {
               #ifdef __DEBUGSLPARSER__
                  dmessage("into variable %s\n", info->alpha_value);
               #endif
               c2 = slparser_get_color(data, info->alpha_value);
               if (c2 == NULL)
                  return FALSE;
            }
            else if (IS_GLOBAL_COLOR(id))
            {
               #ifdef __DEBUGSLPARSER__
                  dmessage("into global %d\n", id);
               #endif
               c2 = (PCOLOR)slparser_get_global(globals, id);
               if (c2 == NULL)
                  return FALSE;
            }
            else
               return FALSE;

            c2[0] = c1[0];
            c2[1] = c1[1];
            c2[2] = c1[2];

            #ifdef __DEBUGSLPARSER__
               dmessage("%d is now: %f %f %f\n", id, 
                             c2[0], c2[1], c2[2]);
            #endif
            break;
         }
         case CPFF_IN_TOKEN:
         {
            /* copy one color into another */
            FLT *val1, *val2;

            id = tokenlist_advance(&current, &info);
            if (id == ALPHA_TOKEN)
            {
               #ifdef __DEBUGSLPARSER__
                  dmessage("Copy variable %s ", info->alpha_value);
               #endif
               val1 = slparser_get_float(data, info->alpha_value);
               if (val1 == NULL)
                  return FALSE;
            }
            else if (IS_GLOBAL_FLOAT(id))
            {
               #ifdef __DEBUGSLPARSER__
                  dmessage("Copy global %d ", id);
               #endif
               val1 = (FLT*)slparser_get_global(globals, id);
               if (val1 == NULL)
                  return FALSE;
               #ifdef __DEBUGSLPARSER__
                  dmessage(" %f ", *val1);
               #endif
            }

            id = tokenlist_advance(&current, &info);
            if (id == ALPHA_TOKEN)
            {
               #ifdef __DEBUGSLPARSER__
                  dmessage("into variable %s\n", info->alpha_value);
               #endif
               val2 = slparser_get_float(data, info->alpha_value);
               if (val2 == NULL)
                  return FALSE;
            }
            else if (IS_GLOBAL_FLOAT(id))
            {
               #ifdef __DEBUGSLPARSER__
                  dmessage("into global %d\n", id);
               #endif
               val2 = (FLT*)slparser_get_global(globals, id);
               if (val2 == NULL)
                  return FALSE;
            }
            else
               return FALSE;

            *val1 = *val2;

            break;
         }

         case CROSS_TOKEN:
         {
            /* cross product */
            vec3 temp1, temp2, temp3;

            #ifdef __DEBUGSLPARSER__
               dmessage("Cross product of two vectors\n");
            #endif /* __DEBUGSLPARSER__ */

            if (!datastack_popv(stack, temp1))
               return FALSE;
            if (!datastack_popv(stack, temp2))
               return FALSE;

            vec3cross(temp1, temp2, temp3);

            if (!datastack_pushv(stack, temp3))
               return FALSE;

            break;
         }
         case DIVF_IN_TOKEN:
         {
            /* divide top by input color */
            id = tokenlist_advance(&current, &info);
            if (id == ALPHA_TOKEN)
            {
               FLT temp1, temp2;
               FLT* val;
               #ifdef __DEBUGSLPARSER__
                  dmessage("Divide top by %s\n", info->alpha_value);
               #endif
               if ((val = slparser_get_float(data, 
                       info->alpha_value)) == NULL)
                  return FALSE;
               if (!datastack_popf(stack, &temp1))
                  return FALSE;

               /*temp1 = temp1 / (*val);*/
               temp2 = (*val) / temp1;

               if (!datastack_pushf(stack, &temp2))
                  return FALSE;
            }
            else if (IS_GLOBAL_COLOR(id))
            {
               FLT temp1, temp2;
               FLT* val;
               #ifdef __DEBUGSLPARSER__
                  dmessage("Divide top by global %d\n", id);
               #endif
               if ((val = (FLT*)slparser_get_global(globals, id)) == NULL)
                  return FALSE;
               if (!datastack_popf(stack, &temp1))
                  return FALSE;
               
               /* temp2 = temp1 / (*val); */
               temp2 = (*val) / temp1;

               if (!datastack_pushf(stack, &temp2))
                  return FALSE;
            }
            break;
         }
         case DIVFF_TOKEN:
         {
            /* divide two float */
            FLT val1, val2;

            #ifdef __DEBUGSLPARSER__
               dmessage("Divide two floats\n");
            #endif /* __DEBUGSLPARSER__ */

            if (!datastack_popf(stack, &val1))
               return FALSE;
            if (!datastack_popf(stack, &val2))
               return FALSE;

            val1 /= val2;

            if (!datastack_pushf(stack, &val1))
               return FALSE;
            
            break;
         }
         case DIVFF_ININ_TOKEN:
         {
            /* divide two floats */
            FLT *val1, *val2;
            FLT val3;

            #ifdef __DEBUGSLPARSER__
               dmessage("divide input two floats\n");
            #endif /* __DEBUGSLPARSER__ */

            id = tokenlist_advance(&current, &info);
            if (id == ALPHA_TOKEN)
            {
               val1 = slparser_get_float(data, info->alpha_value);
               if (val1 == NULL)
                  return FALSE;
            }
            else if (IS_GLOBAL_FLOAT(id))
            {
               val1 = (FLT*)slparser_get_global(globals, id);
               if (val1 == NULL)
                  return FALSE;
            }

            id = tokenlist_advance(&current, &info);
            if (id == ALPHA_TOKEN)
            {
               val2 = slparser_get_float(data, info->alpha_value);
               if (val1 == NULL)
                  return FALSE;
            }
            else if (IS_GLOBAL_FLOAT(id))
            {
               val2 = (FLT*)slparser_get_global(globals, id);
               if (val1 == NULL)
                  return FALSE;
            }

            val3 = *val1 / *val2;

            if (!datastack_pushf(stack, &val3))
               return FALSE;

            break;
         }
         case FNEF_TOKEN:
         {
            /* float not equal */
            FLT temp1, temp2;
            #ifdef __DEBUGSLPARSER__
               dmessage("float not equal\n");
            #endif

            if (!datastack_popf(stack, &temp1))
               return FALSE;
            if (!datastack_popf(stack, &temp2))
               return FALSE;
            temp1 = (FLT)(temp1 != temp2);
            if (!datastack_pushf(stack, &temp1))
               return FALSE;
            break;
         }
         case FTOC_TOKEN:
         {
            /* convert float to color */
            FLT val;
            COLOR temp1;

            #ifdef __DEBUGSLPARSER__
               dmessage("Convert float into color\n");
            #endif /* __DEBUGSLPARSER */

            if (!datastack_popf(stack, &val))
               return FALSE;

            temp1[0] = val;
            temp1[1] = val;
            temp1[2] = val;

            if (!datastack_pushc(stack, temp1))
               return FALSE;

            break;
         }
         case GE_TOKEN:
         {
            /* greater than or equal to */
            FLT temp1, temp2;
            #ifdef __DEBUGSLPARSER__
               dmessage("Push result of greater than\n");
            #endif

            if (!datastack_popf(stack, &temp1))
               return FALSE;
            if (!datastack_popf(stack, &temp2))
               return FALSE;
            temp1 = (FLT)(temp1 >= temp2);
            if (!datastack_pushf(stack, &temp1))
               return FALSE;
            break;
         }
         case GT_TOKEN:
         {
            /* greater than */
            FLT temp1, temp2;
            #ifdef __DEBUGSLPARSER__
               dmessage("Push result of greater than\n");
            #endif

            if (!datastack_popf(stack, &temp1))
               return FALSE;
            if (!datastack_popf(stack, &temp2))
               return FALSE;
            temp1 = (FLT)(temp1 > temp2);
            if (!datastack_pushf(stack, &temp1))
               return FALSE;
            break;
         }
         case IFZ_TOKEN:
         {
            /* conditional */
            FLT temp;
            id = tokenlist_advance(&current, &info);

            #ifdef __DEBUGSLPARSER__
               dmessage("if result of previous goto label %d\n", 
                              (int)info->float_value); 
            #endif
            if (!datastack_popf(stack, &temp))
               return FALSE; 
            if ((int)temp)
               slparser_goto(list, &current, (int)info->float_value);
            break;
         }
         case LT_TOKEN:
         {
            /* less than */
            FLT temp1, temp2;
            #ifdef __DEBUGSLPARSER__
               dmessage("Push result of Less than\n");
            #endif
            if (!datastack_popf(stack, &temp1))
               return FALSE;
            if (!datastack_popf(stack, &temp2))
               return FALSE;
            temp1 = (FLT)(temp1 < temp2);
            if (!datastack_pushf(stack, &temp1))
               return FALSE;
            break;
         }   
         case MULC_IN_TOKEN:
         {
            /* multiply color */
            id = tokenlist_advance(&current, &info);
            if (id == ALPHA_TOKEN)
            {
               COLOR temp;
               PCOLOR val;
               #ifdef __DEBUGSLPARSER__
                  dmessage("Multiply top by %s\n", info->alpha_value);
               #endif
               if ((val = slparser_get_color(data, 
                       info->alpha_value)) == NULL)
                  return FALSE;
               if (!datastack_popc(stack, temp))
                  return FALSE;

               colormult(val, temp, temp);

               if (!datastack_pushc(stack, temp))
                  return FALSE;
            }
            else if (IS_GLOBAL_COLOR(id))
            {
               COLOR temp;
               PCOLOR val;
               #ifdef __DEBUGSLPARSER__
                  dmessage("Multiply top by global %d\n", id);
               #endif
               if ((val = (PCOLOR)slparser_get_global(globals, id)) == NULL)
                  return FALSE;
               if (!datastack_popc(stack, temp))
                  return FALSE;
               colormult(val, temp, temp);
               if (!datastack_pushc(stack, temp))
                  return FALSE;
            }
            break;
         }
         case MULF_IN_TOKEN:
            /* multiply float */
            id = tokenlist_advance(&current, &info);
            if (id == NUMERIC_TOKEN)
            { 
               FLT temp;
               #ifdef __DEBUGSLPARSER__
                  dmessage("Multiply top by %f\n", info->float_value);
               #endif
               if (!datastack_popf(stack, &temp))
                  return FALSE;
               temp *= info->float_value;
               if (!datastack_pushf(stack, &temp))
                  return FALSE;
            }
            else if (IS_GLOBAL_FLOAT(id))
            {
               FLT temp;
               #ifdef __DEBUGSLPARSER__
                  dmessage("Multiply top by global %d\n", id);
               #endif
               if (!datastack_popf(stack, &temp))
                  return FALSE;
               temp *= *(FLT*)(slparser_get_global(globals, id));
               if (!datastack_pushf(stack, &temp))
                  return FALSE;
            }
            else if (id == ALPHA_TOKEN)
            {
               FLT temp;
               #ifdef __DEBUGSLPARSER__
                  dmessage("Multiply top by %s\n", info->alpha_value);
               #endif
               if (!datastack_popf(stack, &temp))
                  return FALSE;
               temp *= *(slparser_get_float(data, info->alpha_value));
               if (!datastack_pushf(stack, &temp))
                  return FALSE;  
            }
            break;

         case MULFF_TOKEN:
         {
            FLT val1, val2;

            /* multiply two floats */
            #ifdef __DEBUGSLPARSER__
               dmessage("Multiply two floats\n");
            #endif /* __DEBUGSLPARSER__ */

            if (!datastack_popf(stack, &val1))
               return FALSE;
            if (!datastack_popf(stack, &val2))
               return FALSE;

            val1 *= val2;

            if (!datastack_pushf(stack, &val1))
               return FALSE;

            break;
         }
         case MULFF_ININ_TOKEN:
         {
            /* multiply two floats */
            FLT *val1, *val2;
            FLT val3;

            #ifdef __DEBUGSLPARSER__
               dmessage("divide input two floats\n");
            #endif /* __DEBUGSLPARSER__ */

            id = tokenlist_advance(&current, &info);
            if (id == ALPHA_TOKEN)
            {
               val1 = slparser_get_float(data, info->alpha_value);
               if (val1 == NULL)
                  return FALSE;
            }
            else if (IS_GLOBAL_FLOAT(id))
            {
               val1 = (FLT*)slparser_get_global(globals, id);
               if (val1 == NULL)
                  return FALSE;
            }

            id = tokenlist_advance(&current, &info);
            if (id == ALPHA_TOKEN)
            {
               val2 = slparser_get_float(data, info->alpha_value);
               if (val1 == NULL)
                  return FALSE;
            }
            else if (IS_GLOBAL_FLOAT(id))
            {
               val2 = (FLT*)slparser_get_global(globals, id);
               if (val1 == NULL)
                  return FALSE;
            }

            val3 = *val1 * *val2;

            if (!datastack_pushf(stack, &val3))
               return FALSE;

            break;
         }
         case MULCC_TOKEN:
         {
            /* multiply two colors */
            COLOR temp1;
            COLOR temp2;
            #ifdef __DEBUGSLPARSER__
               dmessage("Multiply two colors\n");
            #endif
            if (!datastack_popc(stack, temp1))
               return FALSE;
            if (!datastack_popc(stack, temp2))
               return FALSE;

            colormult(temp1, temp2, temp1);

            if (!datastack_pushc(stack, temp1))
               return FALSE;
            break;
         }
         case MULCC_ININ_TOKEN:
         {
            PCOLOR temp1;
            PCOLOR temp2;
            COLOR  temp3;

            #ifdef __DEBUGSLPARSER__
               dmessage("Multiply two input colors\n");
            #endif

            id = tokenlist_advance(&current, &info);
            if (id == ALPHA_TOKEN)
            {
               if ((temp1 = (PCOLOR)slparser_get_color(data, 
                       info->alpha_value)) == NULL)
                  return FALSE;

               #ifdef __DEBUGSLPARSER__
                  dmessage("Multiply color %s: %f %f %f", info->alpha_value,
                            temp1[0], temp1[1], temp1[2]);
               #endif
            }
            else if (IS_GLOBAL_COLOR(id))
            {
               if ((temp1 = (PCOLOR)slparser_get_global(globals, id)) == NULL)
                  return FALSE;

               #ifdef __DEBUGSLPARSER__
                  dmessage("Multiply global color %d: %f %f %f", id,
                            temp1[0], temp1[1], temp1[2]);
               #endif
            }

            id = tokenlist_advance(&current, &info);
            if (id == ALPHA_TOKEN)
            {
               if ((temp2 = (PCOLOR)slparser_get_color(data,
                       info->alpha_value)) == NULL)
                  return FALSE;

               #ifdef __DEBUGSLPARSER__
                  dmessage("  by color %s: %f %f %f\n", info->alpha_value,
                            temp2[0], temp2[1], temp2[2]);
               #endif
            }
            else if (IS_GLOBAL_COLOR(id))
            {
               if ((temp2 = (PCOLOR)slparser_get_global(globals, id)) == NULL)
                  return FALSE;

               #ifdef __DEBUGSLPARSER__
                  dmessage("  by global color %d: %f %f %f\n", id,
                            temp2[0], temp2[1], temp2[2]);
               #endif
            }
            colormult(temp1, temp2, temp3);


            if (!datastack_pushc(stack, temp3))
               return FALSE;
            break;
         }
         case MULFC_TOKEN:
         {
            /* multipy color by float */
            COLOR temp1;
            FLT val;

            #ifdef __DEBUGSLPARSER__
               dmessage("Multiply each component of color by float\n");
            #endif /* __DEBUGSLPARSER__ */

            if (!datastack_popf(stack, &val))
               return FALSE;
            if (!datastack_popc(stack, temp1))
               return FALSE;

            /* multply each component by float */
            colorsmult(val, temp1, temp1);

            if (!datastack_pushc(stack, temp1))
               return FALSE;

            break;
         }

         case MULFC_IN_TOKEN:
         {
            /* multiply color by input float */
            id = tokenlist_advance(&current, &info);
            if (id == ALPHA_TOKEN)
            {
               COLOR tempc;
               FLT* val;
               #ifdef __DEBUGSLPARSER__
                  dmessage("Multiply color by %s\n", info->alpha_value);
               #endif
               if ((val = slparser_get_float(data,
                           info->alpha_value)) == NULL)
                  return FALSE;

               if (!datastack_popc(stack, tempc))
               {
                  return FALSE;
               }
               else
               {
                  tempc[0] *= *val;
                  tempc[1] *= *val;
                  tempc[2] *= *val;

                  if (!datastack_pushc(stack, tempc))
                     return FALSE;
               }
            }
            else if (IS_GLOBAL_FLOAT(id))
            {
               COLOR tempc;
               FLT* val;

               #ifdef __DEBUGSLPARSER__
                  dmessage("Multiply color by global %d\n", id);
               #endif

               if ((val = (FLT*)slparser_get_global(globals, id)) == NULL)
                  return FALSE;
               if (!datastack_popc(stack, tempc))
               {
                  return FALSE;
               }
               else
               {
                  tempc[0] *= *val;
                  tempc[1] *= *val;
                  tempc[2] *= *val;

                  if (!datastack_pushc(stack, tempc))
                     return FALSE;
               }
            }
            break;
         }
         case MULCF_TOKEN:
         {
            /* multiply color by float */
            COLOR temp1;
            FLT val;
            #ifdef __DEBUGSLPARSER__
               dmessage("Multiply color by float\n");
            #endif

            if (!datastack_popc(stack, temp1))
               return FALSE;
            if (!datastack_popf(stack, &val))
               return FALSE;

            temp1[0] *= val;
            temp1[1] *= val;
            temp1[2] *= val;

            if (!datastack_pushc(stack, temp1))
               return FALSE;

            break;
         }
         case MULCF_ININ_TOKEN:
         {
            PCOLOR temp1;
            FLT* temp2;
            COLOR  temp3;

            #ifdef __DEBUGSLPARSER__
               dmessage("Multiply input color by input float\n");
            #endif

            id = tokenlist_advance(&current, &info);
            if (id == ALPHA_TOKEN)
            {
               if ((temp1 = (PCOLOR)slparser_get_color(data,
                       info->alpha_value)) == NULL)
                  return FALSE;

               #ifdef __DEBUGSLPARSER__
                  dmessage("Multiply color %s: %f %f %f", info->alpha_value,
                            temp1[0], temp1[1], temp1[2]);
               #endif
            }
            else if (IS_GLOBAL_COLOR(id))
            {
               if ((temp1 = (PCOLOR)slparser_get_global(globals, id)) == NULL)
                  return FALSE;

               #ifdef __DEBUGSLPARSER__
                  dmessage("Multiply global color %d: %f %f %f", id,
                            temp1[0], temp1[1], temp1[2]);
               #endif
            }

            id = tokenlist_advance(&current, &info);
            if (id == ALPHA_TOKEN)
            {
               if ((temp2 = (PCOLOR)slparser_get_float(data,
                       info->alpha_value)) == NULL)
                  return FALSE;

               #ifdef __DEBUGSLPARSER__
                  dmessage("  by float %s: %f\n", info->alpha_value, *temp2);
               #endif
            }
            else if (IS_GLOBAL_COLOR(id))
            {
               if ((temp2 = (FLT*)slparser_get_global(globals, id)) == NULL)
                  return FALSE;

               #ifdef __DEBUGSLPARSER__
                  dmessage("  by global float %d: %f\n", id, *temp2);
               #endif
            }

            colorsmult(*temp2, temp1, temp3);

            if (!datastack_pushc(stack, temp3))
               return FALSE;
            break;
         }

         case MULFV_IN_TOKEN:
         {
            /* multiply vector by float */
            id = tokenlist_advance(&current, &info);
            if (id == ALPHA_TOKEN)
            {
               COLOR tempc;
               vec3 tempv;
               FLT* val;
               #ifdef __DEBUGSLPARSER__
                  dmessage("Multiply vector by %s\n", info->alpha_value);
               #endif
               if ((val = slparser_get_float(data, 
                           info->alpha_value)) == NULL)
                  return FALSE;

               if (!datastack_popv(stack, tempv))
               {
                  /* do this operation on color to maintain compatibility
                     with BMRT 2.3
                  */
                  if (!datastack_popc(stack, tempc))
                     return FALSE;

                  colorsmult(*val, tempc, tempc);

                  if (!datastack_pushc(stack, tempc))
                     return FALSE;
               }
               else
               {
                  vec3smult(*val, tempv, tempv);

                  if (!datastack_pushv(stack, tempv))
                     return FALSE;
               }
            }
            else if (IS_GLOBAL_FLOAT(id))
            {
               COLOR tempc;
               vec3 tempv;
               FLT* val;

               #ifdef __DEBUGSLPARSER__
                  dmessage("Multiply vector by global %d\n", id);
               #endif

               if ((val = (FLT*)slparser_get_global(globals, id)) == NULL)
                  return FALSE;

               if (!datastack_popv(stack, tempv))
               {
                  /* do this operation on color to maintain compatibility
                     with BMRT 2.3
                  */
                  if (!datastack_popc(stack, tempc))
                     return FALSE;

                  colorsmult(*val, tempc, tempc);

                  if (!datastack_pushc(stack, tempc))
                     return FALSE;
               }
               else
               {
                  vec3smult(*val, tempv, tempv);

                  if (!datastack_pushv(stack, tempv))
                     return FALSE;
               }
            }
            break;
         }
         case NEGF_TOKEN:
            /* negative float */
            error_message("Execution error: NEGF not implemented\n");
            return FALSE;

         case NEGV_TOKEN:
         {
            /* negative vector */
            vec3 temp1;
            
            #ifdef __DEBUGSLPARSER__
               dmessage("Negative vector\n");
            #endif /* __DEBUGSLPARSER__ */

            if (!datastack_popv(stack, temp1))
               return FALSE;

            vec3neg(temp1);

            if (!datastack_pushv(stack, temp1))
               return FALSE;

            break;
         }
         case OR_TOKEN:
         {
            /* boolean or */
            FLT temp1, temp2;
            #ifdef __DEBUGSLPARSER__
               dmessage("Push result of boolean or\n");
            #endif

            if (!datastack_popf(stack, &temp1))
               return FALSE;
            if (!datastack_popf(stack, &temp2))
               return FALSE;
            temp1 = (FLT)((int)temp1 | (int)temp2);
            if (!datastack_pushf(stack, &temp1))
               return FALSE;
            break;
         }
         case SUBCC_TOKEN:
         {
            /* subtract two colors */
            COLOR temp1, temp2;

            #ifdef __DEBUGSLPARSER__
               dmessage("Subtract two colors\n");
            #endif /* __DEBUGSLPARSER__ */

            if (!datastack_popc(stack, temp1))
               return FALSE;
            if (!datastack_popc(stack, temp2))
               return FALSE;

            colorsub(temp1, temp2, temp1);

            if (!datastack_pushc(stack, temp1))
               return FALSE;

            break;
         }
         case SUBCF_TOKEN:
         {
            /* subtract float from color */
            COLOR temp1;
            FLT val;

            #ifdef __DEBUGSLPARSER__
               dmessage("Subtract float from each component of color\n");
            #endif /* __DEBUGSLPARSER__ */

            if (!datastack_popc(stack, temp1))
               return FALSE;
            if (!datastack_popf(stack, &val))
               return FALSE;

            temp1[0] -= val;
            temp1[1] -= val;
            temp1[1] -= val;

            if (!datastack_pushc(stack, temp1))
               return FALSE;

            break;
         }
         case SUBF_IN_TOKEN:
         {
            /* subtract float from input float */
            FLT *temp1;
            FLT temp2, temp3;

            #ifdef __DEBUGSLPARSER__
               dmessage("Subtract float from input float\n");
            #endif

            id = tokenlist_advance(&current, &info);
            if (id == ALPHA_TOKEN)
            {
               if ((temp1 = slparser_get_float(data, info->alpha_value)) == NULL)
                  return FALSE;
            }
            else if (IS_GLOBAL_FLOAT(id))
            {
               if ((temp1 = (FLT*)slparser_get_global(globals, id)) == NULL)
                  return FALSE;
            }
            if (!datastack_popf(stack, &temp2))
               return FALSE; 

            temp3 = *temp1 - temp2;

            if (!datastack_pushf(stack, &temp3))
               return FALSE;

            break;
         }
         case SUBFF_TOKEN:
         {
            /* subtract two floats */
            FLT temp1, temp2;

            #ifdef __DEBUGSLPARSER__
               dmessage("Subtract two floats\n");
            #endif /* __DEBUGSLPARSER__ */

            if (!datastack_popf(stack, &temp1))
               return FALSE;
            if (!datastack_popf(stack, &temp2))
               return FALSE;

            temp1 -= temp2;

            if (!datastack_pushf(stack, &temp1))
               return FALSE;

            break;
         }
         case SUBFF_ININ_TOKEN:
         {
            /* subtract two floats */
            FLT *val1, *val2;
            FLT val3;

            #ifdef __DEBUGSLPARSER__
               dmessage("divide input two floats\n");
            #endif /* __DEBUGSLPARSER__ */

            id = tokenlist_advance(&current, &info);
            if (id == ALPHA_TOKEN)
            {
               val1 = slparser_get_float(data, info->alpha_value);
               if (val1 == NULL)
                  return FALSE;
            }
            else if (IS_GLOBAL_FLOAT(id))
            {
               val1 = (FLT*)slparser_get_global(globals, id);
               if (val1 == NULL)
                  return FALSE;
            }

            id = tokenlist_advance(&current, &info);
            if (id == ALPHA_TOKEN)
            {
               val2 = slparser_get_float(data, info->alpha_value);
               if (val1 == NULL)
                  return FALSE;
            }
            else if (IS_GLOBAL_FLOAT(id))
            {
               val2 = (FLT*)slparser_get_global(globals, id);
               if (val1 == NULL)
                  return FALSE;
            }

            val3 = *val1 - *val2;

            if (!datastack_pushf(stack, &val3))
               return FALSE;

            break;
         }
         case TRACE_TOKEN:
            /* trace new ray */
            error_message("Execution error: TRACE not implemented\n");
            return FALSE;

         case VDOT_TOKEN:
         {
            /* vector dot product */
            vec3 temp1;
            vec3 temp2;
            FLT  result;

            #ifdef __DEBUGSLPARSER__
               dmessage("Dot two top vectors\n");
            #endif /* __DEBUGSLPARSER__ */

            if (!datastack_popv(stack, temp1))
               return FALSE;
            if (!datastack_popv(stack, temp2))
               return FALSE;

            result = vec3dot(temp1, temp2);
            if (!datastack_pushf(stack, &result))
               return FALSE;

            break;
         }
         case CMIX_TOKEN:
         {
            /* color mix */
            COLOR col1, col2, col3;
            FLT val;

            #ifdef __DEBUGSLPARSER__
               dmessage("Mix two colors using float\n");
            #endif

            if (!datastack_popc(stack, col1))
               return FALSE;
            if (!datastack_popc(stack, col2))
               return FALSE;
            if (!datastack_popf(stack, &val))
               return FALSE;

            cmix(col1, col2, val, col3);

            if (!datastack_pushc(stack, col3))
               return FALSE;
            break;
         }
         case COLORNOISEV_TOKEN:
         {
            /* color noise vector */
            vec3 temp1;

            #ifdef __DEBUGSLPARSER__
               dmessage("Color Noise\n");
            #endif /* __DEBUGSLPARSER__`*/

            warning_message("Execution warning: Color Noise not finished\n");

            if (!datastack_popv(stack, temp1))
               return FALSE;

            /* perform noise on vector */

            if (!datastack_pushc(stack, (PCOLOR)temp1))
               return FALSE;

            break;
         }
         case COMP_TOKEN:
         {
            /* component of vector, color, or point */
            COLOR temp1;
            FLT comp;

            #ifdef __DEBUGSLPARSER__
               dmessage("Get component of color\n");
            #endif /* __DEBUGSLPARSER__ */

            if (!datastack_popc(stack, temp1))
               return FALSE;
            if (!datastack_popf(stack, &comp))
               return FALSE;

            /* get component of color */
            comp = temp1[(int)comp];

            if (!datastack_pushf(stack, &comp))
               return FALSE;

            break;
         }
         case AREA_TOKEN:
         {
            /* calculate the surface element area */
            FLT val;

            #ifdef __DEBUGSLPARSER__
               dmessage("Calculate area\n");
            #endif

            val = globals->du * globals->dv;

            if (!datastack_pushf(stack, &val))
               return FALSE;
         }
         case CALCULATENORMAL_TOKEN:
         {
            /* calculate normal at point */
            vec3 temp1;

            #ifdef __DEBUGSLPARSER__
               dmessage("Calculate Normal\n");
            #endif
            warning_message("Execution warning: normal calculation not finished\n");
            if (!datastack_popv(stack, temp1))
               return FALSE;

            /* do normal calculation here */

            if (!datastack_pushv(stack, globals->N))
               return FALSE;
            break;
         }
         case CSET_TOKEN:
         {
            FLT val = 0.0;
            PCOLOR pc = NULL;

            /* set each component of color to float */
            id = tokenlist_advance(&current, &info);
            if (!datastack_popf(stack, &val))
               return FALSE;

            if (id == ALPHA_TOKEN)
            {
               #ifdef __DEBUGSLPARSER__
                  dmessage("Set components of %s to %f\n", info->alpha_value, val);
               #endif /* __DEBUGSLPARSER__ */
               pc = slparser_get_color(data, info->alpha_value);
            }
            else if (IS_GLOBAL_COLOR(id))
            {
               #ifdef __DEBUGSLPARSER__
                  dmessage("Set components of global %d to %f\n", id, val);
               #endif /* __DEBUGSLPARSER__ */
               pc = (PCOLOR)slparser_get_global(globals, id);
            }

            if (pc == NULL)
               error_message("Execution error: Unknown variable\n");

            pc[0] = val;
            pc[1] = val;
            pc[2] = val;

            break;
         }
         case FACEFORWARD_TOKEN:
         {
            /* face forward vector */
            vec3 temp1;
            vec3 temp2;
            #ifdef __DEBUGSLPARSER__
               dmessage("Faceforward using two vectors on stack\n");
            #endif
            if (!datastack_popv(stack, temp1))
               return FALSE;
            if (!datastack_popv(stack, temp2))
               return FALSE;
            faceforward(temp1, temp2, temp1);
            if (!datastack_pushv(stack, temp1))
               return FALSE;
            break;
         }
         case FCLAMP_TOKEN:
         {
            /* clamp float */
            FLT temp1, temp2, temp3, temp4;

            #ifdef __DEBUGSLPARSER__
               dmessage("Clamp float\n");
            #endif

            if (!datastack_popf(stack, &temp1))
               return FALSE;
            if (!datastack_popf(stack, &temp2))
               return FALSE;
            if (!datastack_popf(stack, &temp3))
               return FALSE;

            temp4 = CLAMP(temp1, temp2, temp3);

            if (!datastack_pushf(stack, &temp4))
               return FALSE;

            break;
         }
         case FLOOR_TOKEN:
         {
            /* floor of float */
            FLT temp1, temp2;
            #ifdef __DEBUGSLPARSER__
               dmessage("Floor of float\n");
            #endif 
            if (!datastack_popf(stack, &temp1))
               return FALSE;

            temp2 = floor(temp1);

            if (!datastack_pushf(stack, &temp2))
               return FALSE;
            break;
         }
         case FMAX_TOKEN:
         {
            /* maximum of two floats */
            FLT temp1, temp2, temp3;
            #ifdef __DEBUGSLPARSER__
               dmessage("Maximum of two floats\n");
            #endif
            if (!datastack_popf(stack, &temp1))
               return FALSE;
            if (!datastack_popf(stack, &temp2))
               return FALSE;

            temp3 = MAX(temp1, temp2);

            if (!datastack_pushf(stack, &temp3))
               return FALSE;

            break;
         }
         case FMIN_TOKEN:
         {
            /* minimum of two floats */
            FLT temp1, temp2, temp3;
            #ifdef __DEBUGSLPARSER__
               dmessage("Minimum of two floats\n");
            #endif
            if (!datastack_popf(stack, &temp1))
               return FALSE;
            if (!datastack_popf(stack, &temp2))
               return FALSE;

            temp3 = MIN(temp1, temp2);

            if (!datastack_pushf(stack, &temp3))
               return FALSE;

            break;
         }
         case FRESNEL4_TOKEN:
         {
            vec3 I;
            vec3 N;
            FLT  eta;

            FLT Kr;
            FLT Kt;
            vec3 R;
            vec3 T;
 
            warning_message("Execution warning: Fresnel4 not finished\n");

            #ifdef __DEBUGSLPARSER__
               dmessage("Fresnel called with 4 outputs\n");
            #endif /* __DEBUGSLPARSER */

            if (!datastack_popv(stack, I))
               return FALSE;
            if (!datastack_popv(stack, N))
               return FALSE;
            if (!datastack_popf(stack, &eta))
               return FALSE;

            /* do Fresnel stuff here */

            Kr = 0.25;
            Kt = 0.25;
            vec3init(R);
            vec3init(T);

            if (!datastack_pushv(stack, T))
               return FALSE;
            if (!datastack_pushv(stack, R))
               return FALSE;
            if (!datastack_pushf(stack, &Kt))
               return FALSE;
            if (!datastack_pushf(stack, &Kr))
               return FALSE;

            break;
         }
         case LENGTH_TOKEN:
         {
            /* magnitude of vector */
            vec3 temp1;
            FLT len;

            #ifdef __DEBUGSLPARSER__
               dmessage("Length of vector on stack\n");
            #endif /* __DEBUGSLPARSER__ */

            if (!datastack_popv(stack, temp1))
               return FALSE;

            len = vec3mag(temp1);

            if (!datastack_pushf(stack, &len))
               return FALSE;
            
            break;
         }
         case NORMALIZE_TOKEN:
         {
            /* normalize vector */
            vec3 temp;
            FLT f1, f2, f3;
            #ifdef __DEBUGSLPARSER__
               dmessage("Normalize vector on stack\n");
            #endif
            if (!datastack_popv(stack, temp))
            {
               if (!datastack_popf(stack, &f3))
                  return FALSE;
               if (!datastack_popf(stack, &f2))
                  return FALSE;
               if (!datastack_popf(stack, &f1))
                  return FALSE;

               temp[0] = f1;
               temp[1] = f2;
               temp[2] = f3;
            }

            vec3norm(temp)
            if (!datastack_pushv(stack, temp))
               return FALSE;
            break;
         }
         case NOISEF_TOKEN:
         {
            /* noise function for floats */
            FLT temp1;

            if (!datastack_popf(stack, &temp1))
               return FALSE;

            temp1 = noise1(temp1);

            if (!datastack_pushf(stack, &temp1))
               return FALSE;
            break;
         }
         case NOISEFF_TOKEN:
         {
            /* noise function for two floats */
            float v[2];
            FLT temp1, temp2;

            warning_message("Execution Warning: noisef not finished\n");

            if (!datastack_popf(stack, &temp1))
               return FALSE;
            if (!datastack_popf(stack, &temp2))
               return FALSE;

            v[0] = temp1;
            v[1] = temp2;
            temp1 = noise2(v);

            if (!datastack_pushf(stack, &temp1))
               return FALSE;
            break;
         }
         case NOISEV_TOKEN:
         {
            /* vector noise */
            vec3 temp1;
            float v[3];
            FLT val;

            #ifdef __DEBUGSLPARSER__
               dmessage("Vector Noise\n");
            #endif /* __DEBUGSLPARSER__`*/

            warning_message("Execution warning: Vector Noise not finished\n");

            if (!datastack_popv(stack, temp1))
               return FALSE;

            vec3cpy(v, temp1);

            /* perform noise on vector */
            val = noise3(v);

            if (!datastack_pushf(stack, &val))
               return FALSE;

            break;
         }
         case POW_TOKEN:
         {
            /* power function */
            FLT temp1, temp2, temp3;

            #ifdef __DEBUGSLPARSER__
               dmessage("Power function\n");
            #endif

            if (!datastack_popf(stack, &temp1))
               return FALSE;
            if (!datastack_popf(stack, &temp2))
               return FALSE;

            temp3 = pow(temp1, temp2);

            if (!datastack_pushf(stack, &temp3))
               return FALSE;
            break;
         }
         case SMOOTHSTEP_TOKEN:
         {
            /* smooth step function */
            FLT temp1, temp2, temp3;

            #ifdef __DEBUGSLPARSER__
               dmessage("Smooth step function\n");
            #endif

            if (!datastack_popf(stack, &temp1))
               return FALSE;
            if (!datastack_popf(stack, &temp2))
               return FALSE;
            if (!datastack_popf(stack, &temp3))
               return FALSE;

            temp3 = SMOOTHSTEP(temp1, temp2, temp3);

            if (!datastack_pushf(stack, &temp3))
               return FALSE;

            break;
         }
         case TRANSFORMS_TOKEN:
         {
            /* transform vector, or point */
            char* space = NULL;
            vec3 pt;
            
            if (!datastack_pops(stack, &space))
               return FALSE;
            if (!datastack_popv(stack, pt))
               return FALSE;

            #ifdef __DEBUGSLPARSER__
               dmessage("Transform vector into %s space\n", space);
            #endif /* __DEBUGSLPARSER__ */

            if (!datastack_pushv(stack, pt))
               return FALSE;

            break;
         }
         case XCOMP_TOKEN:
         {
            /* x component of vector */
            vec3 temp1;
            FLT comp;

            #ifdef __DEBUGSLPARSER__
               dmessage("Get x component of vector\n");
            #endif /* __DEBUGSLPARSER__ */

            if (!datastack_popv(stack, temp1))
               return FALSE;

            comp = temp1[0];

            if (!datastack_pushf(stack, &comp))
               return FALSE;

            break;
         }
         case YCOMP_TOKEN:
         {
            /* y component of vector */
            vec3 temp1;
            FLT comp;

            #ifdef __DEBUGSLPARSER__
               dmessage("Get y component of vector\n");
            #endif /* __DEBUGSLPARSER__ */

            if (!datastack_popv(stack, temp1))
               return FALSE;

            comp = temp1[1];

            if (!datastack_pushf(stack, &comp))
               return FALSE;

            break;
         }
         case ZCOMP_TOKEN:
         {
            /* z component of vector */
            vec3 temp1;
            FLT comp;

            #ifdef __DEBUGSLPARSER__
               dmessage("Get z component of vector\n");
            #endif /* __DEBUGSLPARSER__ */

            if (!datastack_popv(stack, temp1))
               return FALSE;

            comp = temp1[2];

            if (!datastack_pushf(stack, &comp))
               return FALSE;

            break;
         }
         case ABS_TOKEN:
         {
            /* absolute value */
            FLT val1;

            #ifdef __DEBUGSLPARSER__
               dmessage("Absolute value of float\n");
            #endif /* __DEBUGSLAPRSER__ */

            if (!datastack_popf(stack, &val1))
               return FALSE;

            val1 = ABS(val1);

            if (!datastack_pushf(stack, &val1))
               return FALSE;

            break;
         }
         case EXP_TOKEN:
            /* exponentiation */
            error_message("Execution error: EXP not implemented\n");
            return FALSE;

         case MOD_TOKEN:
         {
            /* mod function */
            FLT temp1, temp2;
            #ifdef __DEBUGSLPARSER__
               dmessage("Mod function\n");
            #endif
            if (!datastack_popf(stack, &temp1))
               return FALSE;
            if (!datastack_popf(stack, &temp2))
               return FALSE;
            temp1 = mod(temp1, temp2);
            if (!datastack_pushf(stack, &temp1))
               return FALSE;
            break;
         }
         case AMBIENT_TOKEN:
         {
            /* ambient illumination */
            COLOR amb;
            #ifdef __DEBUGSLPARSER__
               dmessage("Ambient illumination\n");
            #endif

            warning_message("Execution warning: Ambient not finished\n");

            amb[0] = 1.0;
            amb[1] = 1.0;
            amb[2] = 1.0; 
            if (!datastack_pushc(stack, amb))
               return FALSE;
            break;
         }
         case DIFFUSE_TOKEN:
         {
            /* diffuse illumination */
            vec3 temp1;
            COLOR temp2;
            #ifdef __DEBUGSLPARSER__
               dmessage("Diffuse function using vector\n");
            #endif

            warning_message("Execution warning: Diffuse not finished\n");

            colorinit(temp2);
            temp2[0] = 1.0;
            temp2[1] = 1.0;
            temp2[2] = 1.0;

            if (!datastack_popv(stack, temp1))
               return FALSE;

            colorsmult(vec3dot(globals->L, temp1), temp2, temp2);
            
            if (!datastack_pushc(stack, temp2))
               return FALSE;

            break;
         }
         case SPECULAR_TOKEN:
         {
            /* specular illumination */
            vec3 N;
            vec3 V;
            FLT  roughness;
            FLT  flt1;

            COLOR spec;

            #ifdef __DEBUGSLPARSER__
               dmessage("specular illumination\n");
            #endif /* __DEBUGSLPARSER__ */
            warning_message("Execution warning: Specular not finished\n");

            if (!datastack_popv(stack, N))
               return FALSE;
            if (!datastack_popv(stack, V))
               return FALSE;
            if (!datastack_popf(stack, &roughness))
               return FALSE;

            /* specular stuff here */
            colorcpy(spec, globals->Cl);

            flt1 = 2.0*vec3dot(N, globals->L);
            vec3smult(flt1, N, N);
            vec3sub(N, globals->L, N);
            colorsmult(pow(vec3dot(N, V), (int)(1.0/roughness)), spec, spec);

            if (!datastack_pushc(stack, spec))
               return FALSE;

            break;
         }
         case CS_GLOBAL_TOKEN:
         case OS_GLOBAL_TOKEN:
         case P_GLOBAL_TOKEN:
         case DPDU_GLOBAL_TOKEN:
         case DPDV_GLOBAL_TOKEN:
         case N_GLOBAL_TOKEN:
         case NG_GLOBAL_TOKEN:
         case U_GLOBAL_TOKEN:
         case V_GLOBAL_TOKEN:
         case DU_GLOBAL_TOKEN:
         case DV_GLOBAL_TOKEN:
         case S_GLOBAL_TOKEN:
         case T_GLOBAL_TOKEN:
         case DSDU_GLOBAL_TOKEN:
         case DSDV_GLOBAL_TOKEN:
         case DTDU_GLOBAL_TOKEN:
         case DTDV_GLOBAL_TOKEN:
         case L_GLOBAL_TOKEN:
         case CL_GLOBAL_TOKEN:
         case I_GLOBAL_TOKEN:
         case CI_GLOBAL_TOKEN:
         case OI_GLOBAL_TOKEN:
         case E_GLOBAL_TOKEN:
            error_message("Execution error: unexpected global: %d\n", id);
            return FALSE;

         case ALPHA_TOKEN:
            error_message("Execution error: unexpected string: %s\n",
               info->alpha_value);
            return FALSE;

         case NUMERIC_TOKEN:
            error_message("Execution error: unexpected number: %f\n",
               info->float_value);
            return FALSE;

         case EOF_TOKEN:
            #ifdef __DEBUGSLPARSER__
               dmessage("End of file\n");
            #endif
            return FALSE;

         default:
            error_message("Execution error: unknown token %d\n", id);
            return FALSE;
      } 
      id = tokenlist_advance(&current, &info);
   } 
   return TRUE;
}

/* FUNCTION slparser_goto
Parameters:
   PTOKEN_LIST list - IN - front of list
   PPTOKEN_LIST current - IN/OUT - current position within list
   int number - IN - label number to go to
Return Value:
   Returns TRUE if label was found, FALSE otherwise
Local Variables:
   TOKEN id - temporary variable to hold token at current position
   PTOKEN_INFO info - token data of current token
Description:
   This function moves the current position to the first occurrance
   of the LABEL_TOKEN followed by the input number. If that label
   is not found, it returns FALSE
*/
int slparser_goto(PTOKEN_LIST list, PPTOKEN_LIST current, int number)
{
   TOKEN id = CODE_TOKEN;
   PTOKEN_INFO info;

   /* begin at beginning of list */
   *current = list;

   /* loop until end of file */
   while(id != EOF_TOKEN)
   {
      /* find next LABEL_TOKEN */
      if(!tokenlist_seek(current, LABEL_TOKEN))
         return FALSE;

      /* if the next token is the number return */
      id = tokenlist_advance(current, &info);
      if (id != NUMERIC_TOKEN)
         return FALSE;

      if (info->float_value == number)
         return TRUE;
   }
   return FALSE;
}
