/***************************
slparser
Copyright (C) 1999  Peter Stuart

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*****************************/


/**********MODULE TOKENIZER***************
Filename: tokenizer.c
Author: ophiuchus42@hotmail.com
Description: This module is used to tokenize a file. Currently it 
   will only tokenize RenderMan shading language shaders compiled
	with BMRT, but later it will be modified to accept a table of
	reserved words, so that it may be used with any file format.
Modifications: 
  12/24/98   commented code
*************************/

#include "tokens.h"
#include "types.h"
#include "tokenizer.h"
#include "tokenlist.h"

/******* STATIC VARIABLES ******/

/* table of reserved words for RenderMan shading language */
static RESERVED_WORDS reserved[MAXTOKENS] = {
   {ABS_TOKEN, "abs"},
   {ADDC_IN_TOKEN, "addC"},
   {ADDCC_TOKEN, "addcc"},
   {ADDCF_TOKEN, "addcf"},
   {ADDF_IN_TOKEN, "addF"},
   {ADDFF_TOKEN, "addff"},
   {ADDFF_ININ_TOKEN, "addFF"},
   {ADDV_IN_TOKEN, "addV"},
   {ADDVV_TOKEN, "addvv"},
   {AMBIENT_TOKEN, "ambient"},
   {AND_TOKEN, "and"},
   {AREA_TOKEN, "area"},
   {CALCULATENORMAL_TOKEN, "calculatenormal"},
   {CELLNOISEF_TOKEN, "cellnoisef"},
   {CELLNOISEFF_TOKEN, "cellnoiseff"},
   {CENVIRONMENT_TOKEN, "cenvironment"},
   {CI_GLOBAL_TOKEN, "Ci"},
   {CL_GLOBAL_TOKEN, "Cl"},
   {CMIX_TOKEN, "cmix"},
   {CODE_TOKEN, "code"},
   {COLOR_TOKEN, "color"},
   {COLORNOISEV_TOKEN, "colornoisev"},
   {COMP_TOKEN, "comp"},
   {CPCC_IN_TOKEN, "cpCC"},
   {CPFF_IN_TOKEN, "cpFF"},
   {CROSS_TOKEN, "cross"},
   {CS_GLOBAL_TOKEN, "Cs"},
   {CSET_TOKEN, "cset"},
   {CTEXTURE_TOKEN, "ctexture"},
   {DERIVV_TOKEN, "Derivv"},
   {DIFFUSE_TOKEN, "diffuse"},
   {DIVF_IN_TOKEN, "divF"},
   {DIVFF_TOKEN, "divff"},
   {DIVFF_ININ_TOKEN, "divFF"},
   {DIVVF_TOKEN, "divvf"},
   {DPDU_GLOBAL_TOKEN, "dPdU"},
   {DPDV_GLOBAL_TOKEN, "dPdV"},
   {DSDU_GLOBAL_TOKEN, "dsdu"},
   {DSDV_GLOBAL_TOKEN, "dsdv"},
   {DTDU_GLOBAL_TOKEN, "dtdu"},
   {DTDV_GLOBAL_TOKEN, "dtdv"},
   {DU_GLOBAL_TOKEN, "du"},
   {DUF_TOKEN, "Duf"},
   {DV_GLOBAL_TOKEN, "dv"},
   {DVF_TOKEN, "Dvf"},
   {E_GLOBAL_TOKEN, "E"},
   {EXP_TOKEN, "exp"},
   {FACEFORWARD_TOKEN, "faceforward"},
   {FCLAMP_TOKEN, "fclamp"},
   {FEQF_TOKEN, "feqf"},
   {FLOAT_TOKEN, "float"},
   {FLOOR_TOKEN, "floor"},
   {FMAX_TOKEN, "fmax"},
   {FMIN_TOKEN, "fmin"},
   {FNEF_TOKEN, "fnef"},
   {FRESNEL4_TOKEN, "fresnel4"},
   {FTOC_TOKEN, "ftoc"},
   {FTEXTURE_TOKEN, "ftexture"},
   {GE_TOKEN, "ge"},
   {GT_TOKEN, "gt"},
   {I_GLOBAL_TOKEN, "I"},
   {IFZ_TOKEN, "ifz"},
   {ILLUMINANCE_START_TOKEN, "illuminance_start"},
   {ILLUMINANCE3_TOKEN, "illuminance3"},
   {INITCODE_TOKEN, "initcode"},
   {JUMP_TOKEN, "jump"},
   {L_GLOBAL_TOKEN, "L"},
   {LABEL_TOKEN, "label"},
   {LENGTH_TOKEN, "length"},
   {LIGHTSOURCE_TOKEN, "lightsource"},
   {LOCAL_TOKEN, "local"},
   {LT_TOKEN, "lt"},
   {MOD_TOKEN, "mod"},
   {MULC_IN_TOKEN, "mulC"},
   {MULCC_TOKEN, "mulcc"},
   {MULCC_ININ_TOKEN, "mulCC"},
   {MULCF_TOKEN, "mulcf"},
   {MULCF_ININ_TOKEN, "mulCF"},
   {MULF_IN_TOKEN, "mulF"},
   {MULFC_TOKEN, "mulfc"},
   {MULFC_IN_TOKEN, "mulFc"},
   {MULFF_TOKEN, "mulff"},
   {MULFF_ININ_TOKEN, "mulFF"},
   {MULFV_IN_TOKEN, "mulFv"},
   {MULFV_TOKEN, "mulfv"},
   {MULVF_TOKEN, "mulvf"},
   {N_GLOBAL_TOKEN, "N"},
   {NEGF_TOKEN, "negf"},
   {NEGV_TOKEN, "negv"},
   {NG_GLOBAL_TOKEN, "Ng"},
   {NOISEF_TOKEN, "noisef"},
   {NOISEFF_TOKEN, "noiseff"},
   {NOISEV_TOKEN, "noisev"},
   {NORMALIZE_TOKEN, "normalize"},
   {NTRANSFORMS_TOKEN, "ntransforms"},
   {OI_GLOBAL_TOKEN, "Oi"},
   {OR_TOKEN, "or"},
   {OS_GLOBAL_TOKEN, "Os"},
   {P_GLOBAL_TOKEN, "P"},
   {PARAM_TOKEN, "param"},
   {POINTNOISEFF_TOKEN, "pointnoiseff"},
   {POINTNOISEV_TOKEN, "pointnoisev"},
   {POPC_TOKEN, "popc"},
   {POPF_TOKEN, "popf"},
   {POPV_TOKEN, "popv"},
   {POW_TOKEN, "pow"},
   {PUSHC_TOKEN, "pushc"},
   {PUSHF_TOKEN, "pushf"},
   {PUSHIF_TOKEN, "pushif"},
   {PUSHIS_TOKEN, "pushis"},
   {PUSHIV_TOKEN, "pushiv"},
   {PUSHS_TOKEN, "pushs"},
   {PUSHV_TOKEN, "pushv"},
   {RANDOM_TOKEN, "random"},
   {REFLECT_TOKEN, "reflect"},
   {RETURN_AREA_TOKEN, "return_area"},
   {RETURN_TOKEN, "return"},
   {S_GLOBAL_TOKEN, "s"},
   {SMOOTHSTEP_TOKEN, "smoothstep"},
   {SNE_TOKEN, "sne"},
   {SPECULAR_TOKEN, "specular"},
   {SQRT_TOKEN, "sqrt"},
   {STRING_TOKEN, "string"},
   {SUBCC_TOKEN, "subcc"},
   {SUBCF_TOKEN, "subcf"},
   {SUBF_IN_TOKEN, "subF"},
   {SUBFF_TOKEN, "subff"},
   {SUBFF_ININ_TOKEN, "subFF"},
   {SUBVV_TOKEN, "subvv"},
   {SURFACE_TOKEN, "surface"},
   {T_GLOBAL_TOKEN, "t"},
   {TRACE_TOKEN, "trace"},
   {TRANSFORMS_TOKEN, "transforms"},
   {U_GLOBAL_TOKEN, "u"},
   {V_GLOBAL_TOKEN, "v"},
   {VDOT_TOKEN, "vdot"},
   {VECTOR_TOKEN, "normal"},
   {VECTOR_TOKEN, "point"},
   {VECTOR_TOKEN, "vector"},
   {VERSION_TOKEN, "version"},
   {VTRANSFORMS_TOKEN, "vtransforms"},
   {XCOMP_TOKEN, "xcomp"},
   {YCOMP_TOKEN, "ycomp"},
   {ZCOMP_TOKEN, "zcomp"}
};


/******* STATIC FUNCTIONS *********/

/* initialize tokenizer */
static void tokenizer_getfile(char* filename, FILE** fpp);

/* read next token from file stream */
static int  tokenizer_read_symbol(PPTOKEN_LIST list, FILE* fp);

/* check to see if current word is reserved word */
static int  tokenizer_find_reserved(char* word);



/* FUNCTION tokenizer_getfile
Parameters:
   char* filename - IN - name of file to tokenize
   FILE** fpp - OUT - pointer to file handle
Return Value:
   void
Local Variables:
   none
Description:
   This function is called by tokenizer_get_tokens in order
   to initialize the tokenizer and open the input file. The 
   handle to the opened file stream is returned in fpp.
*/
void tokenizer_getfile(char* filename, FILE** fpp)
{
   /* open file filename as read-only and copy handle into fpp */
   if ((*fpp = fopen(filename, "r")) == NULL)
   {
      error_message("Tokenizer error: Could not open file %s\n", filename);
      return;
   }

   #ifdef __DEBUGTOKENIZER__
      printf("Opened file %s\n", filename);
   #endif
}

/* FUNCTION tokenizer_terminate
Parameters:
   void
Return Value:
   void
Local Variables:
   none
Description:
   Used to terminate the tokenizer. Currently unused.
*/
void tokenizer_terminate(void)
{
}

/* FUNCTION tokenizer_get_tokens
Parameters:
   char* filename - IN - name of file to tokenize
   PPTOKEN_LIST list - OUT - pointer to the front of the newly created
      token list.
Return Value:
   Return TRUE if function succeeds, FALSE if it fails.
Local Variables:
   char c - holds the present character read from input file
   FILE* fp - handle to the file stream
Description:
   This function is used to tokenize the input file. The when each
   word is tokenized, it is added to the input token list. 
   Non-reserved words are tokenized as "ALPHA" or "NUMERIC", and
   the value is added into the appropriate fields of the token list
   element structure. The function fails if the file is unable to
   be opened, if characters cannot be ungetted (probably caused 
   by a file I/O error), or if memory cannot be allocated.
*/
int tokenizer_get_tokens(PPTOKEN_LIST list, char* filename)
{
   char c = (char)0;
   FILE* fp = NULL;

   /* initialize tokenizer (open file) */
   tokenizer_getfile(filename, &fp);

   /* fail if the file cannot be opened */
   if (fp == NULL)
      return FALSE;

   /* loop until the end of file is reached */
   while(c != EOF)
   {
      /* retrieve present character from the file */
      c = getc(fp);

      /* exit out of loop if the end of file is reached */
      if (c == EOF)
         break;

      /* ignore whitespace */
      if ((c != ' ') && (c != '\t') && (c != '\n'))
      {
         /* fail if unable to put character back into file stream */
         if (ungetc(c, fp) == EOF)
         {
            error_message("Tokenizer error: Could not unget %c\n", c);
            return FALSE;
         }

         /* convert current word to appropriate token */
         if (!tokenizer_read_symbol(list, fp))
            return FALSE;
      }
   }

   /* close file if it had been previously opened successfully */
   if (fp != NULL)
   {
      fclose(fp);

      #ifdef __DEBUGTOKENIZER__
         printf("Closed file.\n");
      #endif
   }
   return TRUE;
}

/* FUNCTION tokenizer_read_symbol
Parameters:
   PPTOKEN_LIST list - OUT - pointer to front of token list
   FILE* fp - IN - file handle
Return Value:
   Returns TRUE if memory allocation succeeded, FALSE otherwise
Local Variables:
   char word[MAXCHARACTERS] - contains the string representation 
      of the current word
   char c - contains the current character
   TOKEN id - token representation of current word
   int pos - position of current character within current word
Description:
   This function converts the current word into a token. It
   returns when a whitespace character is reached.
*/
int tokenizer_read_symbol(PPTOKEN_LIST list, FILE* fp)
{
   char word[MAXCHARACTERS] = "";
   char c;
   TOKEN id;
   int pos = 0;

   /* get first character of current word */
   c = getc(fp);

   /* loop until whitespace is reached or maximum number of
      characters is reached
   */
   while ((c != ' ') && (c != '\t') && (c != '\n') && (pos < MAXCHARACTERS))
   {
      /* set character at current position to current character */
      word[pos++] = c;

      /* get next character */
      c = getc(fp);
   }

   /* convert word into token */
   id = tokenizer_find_reserved(word);

   /* add token to token list */
   if (!tokenlist_add(list, id, word))
      return FALSE;

   #ifdef __DEBUGTOKENIZER__
      printf("Read in symbol %s\n\n", word);
   #endif

   return TRUE;
}

/* FUNCTION tokenizer_find_reserved
Parameters:
   char* word - IN - string representation of current word
Return Value:
   Returns TOKEN if found, otherwise it returns -1
Local Variables:
   int pos - position within reserved words table
Description:
   This function attempts to convert a string word into a token.
   It loops through the reserved word table and compares each
   string to the word.
*/
int tokenizer_find_reserved(char* word)
{
   int pos = 0;

   /* loop until end of reserved words table is reached */
   while(pos < MAXTOKENS)
   {
      /* compare word to current token in table */
      if (strcmp(word, reserved[pos].name) == 0)
         return reserved[pos].id;

      /* try next token in table */
      pos++;
   }
   #ifdef __DEBUGTOKENIZER__
      printf("finding reserved word...\n");
   #endif

   return -1;
}
