/***************************
slparser
Copyright (C) 1999  Peter Stuart

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*****************************/


/**********MODULE TOKENLIST***************
Filename: tokenlist.c
Author: ophiuchus42@hotmail.com
Description: This module implements a singly-linked list
   used for holding tokens. The token list element structure
	includes fields for holding string and numeric information
	as well as the token information.
Modifications: 
  12/24/98   commented code
*************************/

#include "tokens.h"
#include "types.h"
#include "tokenlist.h"
#include "tokenizer.h"

#ifdef __DEBUGTOKENLIST__
   static int reserved_count;
   static int numeric_count;
   static int string_count;
   static int token_count;
#endif

/* FUNCTION tokenlist_init
Parameters:
   PPTOKEN_LIST list - OUT - pointer to the first element of list
Return Value:
   void
Local Variables:
   none
Description:
   Initializes token list. All this really does is initialize
   the list pointer to NULL.
*/
void tokenlist_init(PPTOKEN_LIST list)
{
   /* initialize front to NULL */
   *list = NULL;

   #ifdef __DEBUGTOKENLIST__
      dmessage("Initialized Tokenlist\n");
      reserved_count = 0;
      numeric_count = 0;
      string_count = 0;
      token_count = 0;
   #endif
}

/* FUNCTION tokenlist_delete
Parameters:
   PPTOKEN_LIST list - IN/OUT - pointer to front of list
Return Value:
   void
Local Variables:
   PTOKEN_LIST temp - used to hold next element while current is freed
   PTOKEN_LIST current - the current element in token list
Description:
   This function deletes all elements in the input token list.
*/
void tokenlist_delete(PPTOKEN_LIST list)
{
   PTOKEN_LIST temp = NULL;
   PTOKEN_LIST current = NULL;

   #ifdef __DEBUGTOKENLIST__
      int num = 0;
      int num_reserved = 0;
      int num_numeric = 0;
      int num_string = 0;

      dmessage("Created %d tokens.\n", token_count);
      dmessage("Created %d reserved.\n", reserved_count);
      dmessage("Created %d numeric.\n", numeric_count);
      dmessage("Created %d string.\n", string_count);

      if (list == NULL)
         dmessage("List ptr is NULL!\n");
      if (*list == NULL)
         dmessage("List is NULL!!\n");
   #endif

   /* start at beginning of list */
   current = *list;

   /* loop until end of list */
   while(current != NULL)
   {
      #ifdef __DEBUGTOKENLIST__
         if (current->info.id == ALPHA_TOKEN)
            num_string++;
         else if (current->info.id == NUMERIC_TOKEN)
            num_numeric++;
         else
            num_reserved++;
         num++;
      #endif

      /* free memory allocated to for string if token is "ALPHA" */
      if (current->info.id == ALPHA_TOKEN)
         free(current->info.alpha_value);

      /* hold next element */
      temp = current->next;

      /* free memory allocated for current element */
      free(current);

      /* set current element to next */
      current = temp;
   }

   /* assign NULL to front of list so that it may be used again */
   *list = NULL;

   #ifdef __DEBUGTOKENLIST__
      dmessage("Deleted %d tokens.\n", num);
      dmessage("Deleted %d reserved.\n", num_reserved);
      dmessage("Deleted %d numeric.\n", num_numeric);
      dmessage("Deleted %d string.\n", num_string);
      dmessage("Deleted Tokenlist\n");
   #endif
}

/* FUNCTION tokenlist_add
Parameters:
   PPTOKEN_LIST list - IN/OUT - pointer to front of list
   TOKEN id - IN - token to add onto end of list
   char* word - IN - string representing current token. Used only if the
      token is "ALPHA" or "NUMERIC."
Return Value:
   Returns TRUE if memory allocated, FALSE otherwise
Local Variables:
   FLT flt_value - value of word if token is "NUMERIC"
   PTOKEN_LIST temp - contains current token element 
      (should be changed to current)
   PTOKEN_LIST new_token - pointer to newly created element
Description:
   This function adds a new element to the end of the list. The
   element has a token of type id. If the the token is "ALPHA"
   or "NUMERIC", the extra fields within the element structure
   are used to hold the value of the token.
*/
int tokenlist_add(PPTOKEN_LIST list, TOKEN id, char* word)
{
   FLT flt_value = (FLT)0.0;
   char* extra_str = NULL;
   PTOKEN_LIST temp = NULL;
   PTOKEN_LIST new_token = NULL;

   /* attempt to allocate memory for new element */
   if ((new_token = (PTOKEN_LIST)malloc(sizeof(TOKEN_LIST))) == NULL)
   {
      fprintf(stderr, "Could not allocate memory");
      return FALSE;
   }

   /* assign next pointer of new element to NULL */
   new_token->next = NULL;

   /* if the list does not have an elements yet assign front to
      newly allocated element
   */
   if (*list == NULL)
   {
      #ifdef __DEBUGTOKENLIST__
         dmessage("Adding onto front of list.");
      #endif

      *list = new_token;
   }
   else
   {
      /* start at front of list */
      temp = *list;

      /* find the end of the list */
      while(temp->next != NULL)
         temp = temp->next;

      /* add new element to end of list */
      temp->next = new_token;
   }
   
   #ifdef __DEBUGTOKENLIST__
      dmessage("Allocated new token in list...\n");
   #endif

   /* if token is not reserved word use other fields in element */
   if (id == -1)
   {
      /* if word is a string assign "ALPHA" token, and create string */
      flt_value = (FLT)strtod(word, &extra_str);
      
      if (((flt_value == 0.0) && strnneq(word, "0", 1)) 
          || (strlen(extra_str) > 0))
      {
         #ifdef __DEBUGTOKENLIST__
            dmessage("Found string %s\n", word);
            string_count++;
            token_count++;
         #endif

         new_token->info.id = ALPHA_TOKEN;

         /* attempt to allocate string */
         if ((new_token->info.alpha_value
                   = (char*)calloc(strlen(word)+1, sizeof(char))) == NULL)
         {
            error_message("Tokenlist error: Could not allocate memory");
            return FALSE;
         }

         /* copy input string into the alpha_value of the element */
         strcpy(new_token->info.alpha_value, word);
      }
      else
      {
         /* token is numeric */
         #ifdef __DEBUGTOKENLIST__
            dmessage("Found numeric value %f\n", (float)flt_value);
            numeric_count++;
            token_count++;
         #endif

         /* assign numeric token */
         new_token->info.id = NUMERIC_TOKEN;

         /* copy numeric value into float_value field of element */
         new_token->info.float_value = flt_value;

         /* assign string value to NULL */
         new_token->info.alpha_value = NULL;
      }
   }
   else
   {
      /* reserved word is found so assign token */
      #ifdef __DEBUGTOKENLIST__
         dmessage("Found reserved word %d\n", id);
         reserved_count++;
         token_count++;
      #endif

      new_token->info.id = id;
   }

   return TRUE;
}

/* FUNCTION tokenlist_next
Parameters:
   PTOKEN_LIST position - IN - pointer to current element
   PPTOKEN_INFO info - OUT - info field of next element
Return Value:
   Returns token of next element
Local Variables:
   TOKEN ret_token - holds token of next element
Description:
   This function returns the token of the next element in the list.
   The info variable is filled with the info information of the
   next element. If info is NULL, the function does not copy
   the information.
*/
TOKEN tokenlist_next(PTOKEN_LIST position, PPTOKEN_INFO info)
{
   TOKEN ret_token;

   /* check if end of token list has been reached */
   if (position->next == NULL)
      return EOF_TOKEN;

   /* set ret_token to be the token of the next element */
   ret_token = position->next->info.id;

   /* fill in info of next element if input is not NULL */
   if (info != NULL)
      *info = &position->next->info;

   return ret_token;
}

/* FUNCTION tokenlist_advance
Parameters:
   PPTOKEN_LIST position - IN/OUT - current position in list
   PPTOKEN_INFO info - OUT - info field of next element
Return Value:
   Returns token of next element, EOF_TOKEN if end of list
Local Variables:
   TOKEN ret_token - holds token of next element
Description:
   This function returns the token of the next element in the list.
   The info variable is filled with the info information of the
   next element. If info is NULL, the function does not copy
   the information. Once this is done position is advanced to point
   at the next element.
*/
TOKEN tokenlist_advance(PPTOKEN_LIST position, PPTOKEN_INFO info)
{
   TOKEN ret_token;

   /* check if end of token list has been reached */
   if ((*position)->next == NULL)
      return EOF_TOKEN;

   /* set ret_token to be the token of the next element */
   ret_token = (*position)->next->info.id;

   /* fill in info from next element if info is not NULL */
   if (info != NULL)
      *info = &(*position)->next->info;

   /* advance pointer to next element */
   *position = (*position)->next;
   return ret_token;
}

/* FUNCTION tokenlist_seek
Parameters:
   PPTOKEN_LIST position - IN/OUT - current position within token list
   TOKEN token - IN - token to find
Return Value:
   Returns TRUE if the token was found, FALSE otherwise
Local Variables:
   PTOKEN_LIST current - used to hold the current element
   TOKEN id - token of current element
Description:
   This function seeks the first occurrance of a specific 
   token within the list beginningwith the position given. 
   When the token is found position is	updated to point 
   to the occurance of that token.
*/
int tokenlist_seek(PPTOKEN_LIST position, TOKEN token)
{
   PTOKEN_LIST current;
   TOKEN id;

   /* start with front of list */
   current = *position;

   /* token of current element */
   id = (*position)->info.id;

   /* loop until element with input token is found or until
      the list ends
   */
   while((id != token) && (id != EOF_TOKEN))
   {
      /* retrieve next id and advance pointer (this can be done
         with one call to tokenlist_advance
      */
      id = tokenlist_next(current, NULL);
      tokenlist_advance(&current, NULL);
   }

   /* update position if token is found */
   if (id == token)   
      *position = current;

   return (id == token);
}
