/***************************
slparser
Copyright (C) 1999  Peter Stuart

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*****************************/

#ifndef __SLPARSER_TYPES_H__
#define __SLPARSER_TYPES_H__

typedef int TOKEN;

/* forward type definitions for all structures */
typedef struct shader_data SHADER_DATA;
typedef SHADER_DATA *PSHADER_DATA;
typedef struct global_variables GLOBAL_VARIABLES;
typedef GLOBAL_VARIABLES *PGLOBAL_VARIABLES;
typedef struct reserved_words RESERVED_WORDS;
typedef struct token_info TOKEN_INFO;
typedef TOKEN_INFO *PTOKEN_INFO;
typedef TOKEN_INFO **PPTOKEN_INFO;
typedef struct token_list TOKEN_LIST;
typedef TOKEN_LIST *PTOKEN_LIST;
typedef TOKEN_LIST **PPTOKEN_LIST;
typedef struct data_stack DATA_STACK;
typedef DATA_STACK *PDATA_STACK;
typedef DATA_STACK **PPDATA_STACK;

/****** MODULE SLPARSER *******/

/* STRUCTURE shader_data
Fields:
   int shader_number - a number given to a specific shader 
	   (not currently in use)
   char* shader_name - name of the specified shader
   char** float_name - array of names of the float parameters and
	   variables of the specified shader
	FLT* float_data - array of the values of the float parameters
	   and variables of the shader
   inf float_params - number of float parameters
   int float_num - number of float variables
	char** vector_name - array of names of the vector variables and
	   parameters
	vec3* - array of values of the vector variables and parameters
   int vector_params - number of vector params
   int vector_num - number of vector variables and parameters
	char** string_name - array of names of the string variables and
	   parameters
	char** string_data - array of the string values of the string
	   variables and parameters
   int string_num - number of string variables and parameters
	char** color_name - array of names of the color variables and
	   parameters
	PCOLOR color_data - array of color values for the variables and
	   parameters
   int color_params - number of color parameters
   int color_num - number of color variables and parameters
Description:
   This structure is used to hold the names and values of all of
	the shader's parameters and local variables. (This should also
	store the number of each type as well)
*/
struct shader_data
{
   int    shader_number;
   char*  shader_name;
   char*  generator_name;
   char*  version_name;
   char** float_name;
   FLT*   float_data;
   int    float_params;
   int    float_num;
   char** vector_name;
   vec3*  vector_data;
   int    vector_params;
   int    vector_num;
   char** string_name;
   char** string_data;
   int    string_params;
   int    string_num;
   char** color_name;
   COLOR* color_data;
   int    color_params;
   int    color_num;
};

/* STRUCTURE global_variables
Fields:
   COLOR  Cs - surface color (input)
   COLOR  Os - surface opacity (input)
   point  P - surface position
   vec3   dPdu - change in position with u
   vec3   dPdv -  change in position with v
   normal N - surface shading normal
   normal Ng - surface geometric normal
   FLT    u - surface u parameter
   FLT    v - surface v parameter
   FLT    du - change in u across element
   FLT    dv - change in v across element
   FLT    s - texture s coordinate
   FLT    t - texture t coordinate
   FLT    dsdu - change in s with u
   FLT    dsdv - change in s with v
   FLT    dtdu - change in t with u
   FLT    dtdv - change in t with v
   vec3   L - direction to light
   COLOR  Cl - light color
   vec3   I - direction of ray
   COLOR  Ci - color from surface (output)
   COLOR  Oi - opacity from surface (output)
   point  E  - position of camera
Description:
   This structure is used to hold the global variable information
	for the shading language global variables.
*/
struct global_variables
{
   COLOR  Cs;
   COLOR  Os;
   vec3   P;
   vec3   dPdu;
   vec3   dPdv;
   vec3   N;
   vec3   Ng;
   FLT    u;
   FLT    v;
   FLT    du;
   FLT    dv;
   FLT    s;
   FLT    t;
   FLT    dsdu;
   FLT    dsdv;
   FLT    dtdu;
   FLT    dtdv;
   vec3   L;
   COLOR  Cl;
   vec3   I;
   COLOR  Ci;
   COLOR  Oi;
   vec3   E;
};

/****** MODULE TOKENIZER ******/

/* STRUCTURE reserved_words
Fields:
   TOKEN id - token of present word
	char* name - string representation of word
Description:
   This structure contains information necessary to convert
	reserved words into tokens. The string representation of
	the current word is compared to a string to determine if
	the string is the reserved word.
*/
struct reserved_words
{
   TOKEN id;
   char* name;
};

/* MODULE TOKEN LIST */

/* STRUCTURE token_info
Fields:
   TOKEN id - token of present element
	char* alpha_value - string value if token is ALPHA_TOKEN
	FLT float_value - float value if token is NUMERIC_TOKEN
Description:
   This structure contains the data for an element in the token
	list. If the token is not a reserved word, this structure also
	contains other information.
*/
struct token_info
{
   TOKEN  id;
   char*  alpha_value;
   FLT    float_value;
};

/* STRUCTURE token_list
Fields:
   token_info info - info structure containing the element data
	token_list* next - pointer to the next element in the list
Description:
   This is the structure which represents an element in the token
	list.
*/
struct token_list
{
   TOKEN_INFO info;
   PTOKEN_LIST next;
};

/****** MODULE DATA STACK ******/

/* STRUCTURE data_stack
Fields:
   TOKEN data_id - type of data in element
	FLT float_value - holds information for FLOAT_TOKEN data
	char* string_value - holds information for STRING_TOKEN data
	COLOR color_value - holds information for COLOR_TOKEN data
	vec3 vector_value - holds information for VECTOR_TOKEN data
	data_stack* next - pointer to the next element in the stack
Description:
   This is the structure which represents an element in the data
	stack.
*/
struct data_stack
{
   TOKEN  data_id;
   FLT    float_value;
   char*  string_value;
   COLOR  color_value;
   vec3   vector_value;

   PDATA_STACK next;
};


#endif
